import { ChangeDetectionStrategy, Component, ContentChild, ContentChildren, Directive, ElementRef, EventEmitter, HostBinding, HostListener, Inject, Injectable, Input, KeyValueDiffers, NgModule, NgZone, Output, Renderer, TemplateRef, ViewChild, ViewContainerRef, ViewEncapsulation } from '@angular/core';
import { CommonModule } from '@angular/common';
import 'rxjs/add/observable/fromEvent';
import { DOCUMENT } from '@angular/platform-browser';
import { Observable as Observable$1 } from 'rxjs/Observable';
import 'rxjs/add/operator/takeUntil';
/**
 * Creates a unique object id.
 * http://stackoverflow.com/questions/6248666/how-to-generate-short-uid-like-ax4j9z-in-js
 * @return {?}
 */
function id() {
    return ('0000' + (Math.random() * Math.pow(36, 4) << 0).toString(36)).slice(-4);
}
/**
 * Returns the columns by pin.
 * @param {?} cols
 * @return {?}
 */
function columnsByPin(cols) {
    var /** @type {?} */ ret = {
        left: [],
        center: [],
        right: []
    };
    if (cols) {
        for (var _i = 0, cols_1 = cols; _i < cols_1.length; _i++) {
            var col = cols_1[_i];
            if (col.frozenLeft) {
                ret.left.push(col);
            }
            else if (col.frozenRight) {
                ret.right.push(col);
            }
            else {
                ret.center.push(col);
            }
        }
    }
    return ret;
}
/**
 * Returns the widths of all group sets of a column
 * @param {?} groups
 * @param {?} all
 * @return {?}
 */
function columnGroupWidths(groups, all) {
    return {
        left: columnTotalWidth(groups.left),
        center: columnTotalWidth(groups.center),
        right: columnTotalWidth(groups.right),
        total: columnTotalWidth(all)
    };
}
/**
 * Calculates the total width of all columns and their groups
 * @param {?} columns
 * @param {?=} prop
 * @return {?}
 */
function columnTotalWidth(columns, prop) {
    var /** @type {?} */ totalWidth = 0;
    if (columns) {
        for (var _i = 0, columns_1 = columns; _i < columns_1.length; _i++) {
            var c = columns_1[_i];
            var /** @type {?} */ has = prop && c[prop];
            var /** @type {?} */ width = has ? c[prop] : c.width;
            totalWidth = totalWidth + parseFloat(width);
        }
    }
    return totalWidth;
}
/**
 * Calculates the total width of all columns and their groups
 * @param {?} columns
 * @param {?=} prop
 * @return {?}
 */
function columnsTotalWidth(columns, prop) {
    var /** @type {?} */ totalWidth = 0;
    for (var _i = 0, columns_2 = columns; _i < columns_2.length; _i++) {
        var column = columns_2[_i];
        var /** @type {?} */ has = prop && column[prop];
        totalWidth = totalWidth + (has ? column[prop] : column.width);
    }
    return totalWidth;
}
/**
 * @param {?} val
 * @return {?}
 */
function columnsByPinArr(val) {
    var /** @type {?} */ colsByPinArr = [];
    var /** @type {?} */ colsByPin = columnsByPin(val);
    colsByPinArr.push({ type: 'left', columns: colsByPin['left'] });
    colsByPinArr.push({ type: 'center', columns: colsByPin['center'] });
    colsByPinArr.push({ type: 'right', columns: colsByPin['right'] });
    return colsByPinArr;
}
// maybe rename this file to prop-getters.ts
/**
 * Always returns the empty string ''
 * @return {?}
 */
function emptyStringGetter() {
    return '';
}
/**
 * Returns the appropriate getter function for this kind of prop.
 * If prop == null, returns the emptyStringGetter.
 * @param {?} prop
 * @return {?}
 */
function getterForProp(prop) {
    if (prop == null)
        return emptyStringGetter;
    if (typeof prop === 'number') {
        return numericIndexGetter;
    }
    else {
        // deep or simple
        if (prop.indexOf('.') !== -1) {
            return deepValueGetter;
        }
        else {
            return shallowValueGetter;
        }
    }
}
/**
 * Returns the value at this numeric index.
 * @param {?} row array of values
 * @param {?} index numeric index
 * @return {?}
 */
function numericIndexGetter(row, index) {
    // mimic behavior of deepValueGetter
    if (!row || index == null)
        return row;
    var /** @type {?} */ value = row[index];
    if (value == null)
        return '';
    return value;
}
/**
 * Returns the value of a field.
 * (more efficient than deepValueGetter)
 * @param {?} obj object containing the field
 * @param {?} fieldName field name string
 * @return {?}
 */
function shallowValueGetter(obj, fieldName) {
    if (!obj || !fieldName)
        return obj;
    var /** @type {?} */ value = obj[fieldName];
    if (value == null)
        return '';
    return value;
}
/**
 * Returns a deep object given a string. zoo['animal.type']
 * @param {?} obj
 * @param {?} path
 * @return {?}
 */
function deepValueGetter(obj, path) {
    if (!obj || !path)
        return obj;
    // check if path matches a root-level field
    // { "a.b.c": 123 }
    var /** @type {?} */ current = obj[path];
    if (current !== undefined)
        return current;
    current = obj;
    var /** @type {?} */ split = path.split('.');
    if (split.length) {
        for (var /** @type {?} */ i = 0; i < split.length; i++) {
            current = current[split[i]];
            // if found undefined, return empty string
            if (current === undefined || current === null)
                return '';
        }
    }
    return current;
}
/**
 * Converts strings from something to camel case
 * http://stackoverflow.com/questions/10425287/convert-dash-separated-string-to-camelcase
 * @param {?} str
 * @return {?}
 */
function camelCase(str) {
    // Replace special characters with a space
    str = str.replace(/[^a-zA-Z0-9 ]/g, ' ');
    // put a space before an uppercase letter
    str = str.replace(/([a-z](?=[A-Z]))/g, '$1 ');
    // Lower case first character and some other stuff
    str = str.replace(/([^a-zA-Z0-9 ])|^[0-9]+/g, '').trim().toLowerCase();
    // uppercase characters preceded by a space or number
    str = str.replace(/([ 0-9]+)([a-zA-Z])/g, function (a, b, c) {
        return b.trim() + c.toUpperCase();
    });
    return str;
}
/**
 * Converts strings from camel case to words
 * http://stackoverflow.com/questions/7225407/convert-camelcasetext-to-camel-case-text
 *
 * @export
 * @param {?} str
 * @return {?} string
 */
function deCamelCase(str) {
    return str
        .replace(/([A-Z])/g, function (match) { return " " + match; })
        .replace(/^./, function (match) { return match.toUpperCase(); });
}
var Keys = {};
Keys.up = 38;
Keys.down = 40;
Keys.return = 13;
Keys.escape = 27;
Keys.left = 37;
Keys.right = 39;
Keys[Keys.up] = "up";
Keys[Keys.down] = "down";
Keys[Keys.return] = "return";
Keys[Keys.escape] = "escape";
Keys[Keys.left] = "left";
Keys[Keys.right] = "right";
/**
 * Calculates the Total Flex Grow
 * @param {?} columns
 * @return {?}
 */
function getTotalFlexGrow(columns) {
    var /** @type {?} */ totalFlexGrow = 0;
    for (var _i = 0, columns_3 = columns; _i < columns_3.length; _i++) {
        var c = columns_3[_i];
        totalFlexGrow += c.flexGrow || 0;
    }
    return totalFlexGrow;
}
/**
 * Adjusts the column widths.
 * Inspired by: https://github.com/facebook/fixed-data-table/blob/master/src/FixedDataTableWidthHelper.js
 * @param {?} allColumns
 * @param {?} expectedWidth
 * @return {?}
 */
function adjustColumnWidths(allColumns, expectedWidth) {
    var /** @type {?} */ columnsWidth = columnsTotalWidth(allColumns);
    var /** @type {?} */ totalFlexGrow = getTotalFlexGrow(allColumns);
    var /** @type {?} */ colsByGroup = columnsByPin(allColumns);
    if (columnsWidth !== expectedWidth) {
        scaleColumns(colsByGroup, expectedWidth, totalFlexGrow);
    }
}
/**
 * Resizes columns based on the flexGrow property, while respecting manually set widths
 * @param {?} colsByGroup
 * @param {?} maxWidth
 * @param {?} totalFlexGrow
 * @return {?}
 */
function scaleColumns(colsByGroup, maxWidth, totalFlexGrow) {
    // calculate total width and flexgrow points for coulumns that can be resized
    for (var /** @type {?} */ attr in colsByGroup) {
        for (var _i = 0, _a = colsByGroup[attr]; _i < _a.length; _i++) {
            var column = _a[_i];
            if (!column.canAutoResize) {
                maxWidth -= column.width;
                totalFlexGrow -= column.flexGrow;
            }
            else {
                column.width = 0;
            }
        }
    }
    var /** @type {?} */ hasMinWidth = {};
    var /** @type {?} */ remainingWidth = maxWidth;
    // resize columns until no width is left to be distributed
    do {
        var /** @type {?} */ widthPerFlexPoint = remainingWidth / totalFlexGrow;
        remainingWidth = 0;
        for (var /** @type {?} */ attr in colsByGroup) {
            for (var _b = 0, _c = colsByGroup[attr]; _b < _c.length; _b++) {
                var column = _c[_b];
                // if the column can be resize and it hasn't reached its minimum width yet
                if (column.canAutoResize && !hasMinWidth[column.prop]) {
                    var /** @type {?} */ newWidth = column.width + column.flexGrow * widthPerFlexPoint;
                    if (column.minWidth !== undefined && newWidth < column.minWidth) {
                        remainingWidth += newWidth - column.minWidth;
                        column.width = column.minWidth;
                        hasMinWidth[column.prop] = true;
                    }
                    else {
                        column.width = newWidth;
                    }
                }
            }
        }
    } while (remainingWidth !== 0);
}
/**
 * Forces the width of the columns to
 * distribute equally but overflowing when necessary
 *
 * Rules:
 *
 *  - If combined withs are less than the total width of the grid,
 *    proportion the widths given the min / max / normal widths to fill the width.
 *
 *  - If the combined widths, exceed the total width of the grid,
 *    use the standard widths.
 *
 *  - If a column is resized, it should always use that width
 *
 *  - The proportional widths should never fall below min size if specified.
 *
 *  - If the grid starts off small but then becomes greater than the size ( + / - )
 *    the width should use the original width; not the newly proportioned widths.
 *
 * @param {?} allColumns
 * @param {?} expectedWidth
 * @param {?} startIdx
 * @param {?} allowBleed
 * @param {?=} defaultColWidth
 * @return {?}
 */
function forceFillColumnWidths(allColumns, expectedWidth, startIdx, allowBleed, defaultColWidth) {
    if (defaultColWidth === void 0) { defaultColWidth = 300; }
    var /** @type {?} */ columnsToResize = allColumns
        .slice(startIdx + 1, allColumns.length)
        .filter(function (c) {
        return c.canAutoResize !== false;
    });
    for (var _i = 0, columnsToResize_1 = columnsToResize; _i < columnsToResize_1.length; _i++) {
        var column = columnsToResize_1[_i];
        if (!column.$$oldWidth) {
            column.$$oldWidth = column.width;
        }
    }
    var /** @type {?} */ additionWidthPerColumn = 0;
    var /** @type {?} */ exceedsWindow = false;
    var /** @type {?} */ contentWidth = getContentWidth(allColumns, defaultColWidth);
    var /** @type {?} */ remainingWidth = expectedWidth - contentWidth;
    var /** @type {?} */ columnsProcessed = [];
    // This loop takes care of the
    do {
        additionWidthPerColumn = remainingWidth / columnsToResize.length;
        exceedsWindow = contentWidth >= expectedWidth;
        for (var _a = 0, columnsToResize_2 = columnsToResize; _a < columnsToResize_2.length; _a++) {
            var column = columnsToResize_2[_a];
            if (exceedsWindow && allowBleed) {
                column.width = column.$$oldWidth || column.width || defaultColWidth;
            }
            else {
                var /** @type {?} */ newSize = (column.width || defaultColWidth) + additionWidthPerColumn;
                if (column.minWidth && newSize < column.minWidth) {
                    column.width = column.minWidth;
                    columnsProcessed.push(column);
                }
                else if (column.maxWidth && newSize > column.maxWidth) {
                    column.width = column.maxWidth;
                    columnsProcessed.push(column);
                }
                else {
                    column.width = newSize;
                }
            }
            column.width = Math.max(0, column.width);
        }
        contentWidth = getContentWidth(allColumns);
        remainingWidth = expectedWidth - contentWidth;
        removeProcessedColumns(columnsToResize, columnsProcessed);
    } while (remainingWidth > 0 && columnsToResize.length !== 0);
}
/**
 * Remove the processed columns from the current active columns.
 *
 * @param {?} columnsToResize  Array containing the columns that need to be resized.
 * @param {?} columnsProcessed Array containing the columns that have already been processed.
 * @return {?}
 */
function removeProcessedColumns(columnsToResize, columnsProcessed) {
    for (var _i = 0, columnsProcessed_1 = columnsProcessed; _i < columnsProcessed_1.length; _i++) {
        var column = columnsProcessed_1[_i];
        var /** @type {?} */ index = columnsToResize.indexOf(column);
        columnsToResize.splice(index, 1);
    }
}
/**
 * Gets the width of the columns
 *
 * @param {?} allColumns
 * @param {?=} defaultColWidth
 * @return {?}
 */
function getContentWidth(allColumns, defaultColWidth) {
    if (defaultColWidth === void 0) { defaultColWidth = 300; }
    var /** @type {?} */ contentWidth = 0;
    for (var _i = 0, allColumns_1 = allColumns; _i < allColumns_1.length; _i++) {
        var column = allColumns_1[_i];
        contentWidth += (column.width || defaultColWidth);
    }
    return contentWidth;
}
var cache = {};
var testStyle = document.createElement('div').style;
// Get Prefix
// http://davidwalsh.name/vendor-prefix
var prefix = (function () {
    var /** @type {?} */ styles = window.getComputedStyle(document.documentElement, '');
    var /** @type {?} */ pre = (Array.prototype.slice.call(styles).join('').match(/-(moz|webkit|ms)-/))[1];
    var /** @type {?} */ dom = ('WebKit|Moz|MS|O').match(new RegExp('(' + pre + ')', 'i'))[1];
    return {
        dom: dom,
        lowercase: pre,
        css: "-" + pre + "-",
        js: pre[0].toUpperCase() + pre.substr(1)
    };
})();
/**
 * @param {?} property
 * @return {?}
 */
function getVendorPrefixedName(property) {
    var /** @type {?} */ name = camelCase(property);
    if (!cache[name]) {
        if (testStyle[prefix.css + property] !== undefined) {
            cache[name] = prefix.css + property;
        }
        else if (testStyle[property] !== undefined) {
            cache[name] = property;
        }
    }
    return cache[name];
}
/**
 * @param {?} selected
 * @param {?} row
 * @param {?} comparefn
 * @return {?}
 */
function selectRows(selected, row, comparefn) {
    var /** @type {?} */ selectedIndex = comparefn(row, selected);
    if (selectedIndex > -1) {
        selected.splice(selectedIndex, 1);
    }
    else {
        selected.push(row);
    }
    return selected;
}
/**
 * @param {?} selected
 * @param {?} rows
 * @param {?} index
 * @param {?} prevIndex
 * @param {?} comparefn
 * @return {?}
 */
function selectRowsBetween(selected, rows, index, prevIndex, comparefn) {
    var /** @type {?} */ reverse = index < prevIndex;
    for (var /** @type {?} */ i = 0; i < rows.length; i++) {
        var /** @type {?} */ row = rows[i];
        var /** @type {?} */ greater = i >= prevIndex && i <= index;
        var /** @type {?} */ lesser = i <= prevIndex && i >= index;
        var /** @type {?} */ range = { start: 0, end: 0 };
        if (reverse) {
            range = {
                start: index,
                end: (prevIndex - index)
            };
        }
        else {
            range = {
                start: prevIndex,
                end: index + 1
            };
        }
        if ((reverse && lesser) || (!reverse && greater)) {
            // if in the positive range to be added to `selected`, and
            // not already in the selected array, add it
            if (i >= range.start && i <= range.end) {
                selected.push(row);
            }
        }
    }
    return selected;
}
// browser detection and prefixing tools
var transform = getVendorPrefixedName('transform');
var backfaceVisibility = getVendorPrefixedName('backfaceVisibility');
var hasCSSTransforms = !!getVendorPrefixedName('transform');
var hasCSS3DTransforms = !!getVendorPrefixedName('perspective');
var ua = window.navigator.userAgent;
var isSafari = (/Safari\//).test(ua) && !(/Chrome\//).test(ua);
/**
 * @param {?} styles
 * @param {?} x
 * @param {?} y
 * @return {?}
 */
function translateXY(styles, x, y) {
    if (hasCSSTransforms) {
        if (!isSafari && hasCSS3DTransforms) {
            styles[transform] = "translate3d(" + x + "px, " + y + "px, 0)";
            styles[backfaceVisibility] = 'hidden';
        }
        else {
            styles[camelCase(transform)] = "translate(" + x + "px, " + y + "px)";
        }
    }
    else {
        styles.top = y + "px";
        styles.left = x + "px";
    }
}
/**
 * Throttle a function
 *
 * @export
 * @param {?} func
 * @param {?} wait
 * @param {?=} options
 * @return {?}
 */
function throttle(func, wait, options) {
    options = options || {};
    var /** @type {?} */ context;
    var /** @type {?} */ args;
    var /** @type {?} */ result;
    var /** @type {?} */ timeout = null;
    var /** @type {?} */ previous = 0;
    /**
     * @return {?}
     */
    function later() {
        previous = options.leading === false ? 0 : +new Date();
        timeout = null;
        result = func.apply(context, args);
    }
    return function () {
        var /** @type {?} */ now = +new Date();
        if (!previous && options.leading === false) {
            previous = now;
        }
        var /** @type {?} */ remaining = wait - (now - previous);
        context = this;
        args = arguments;
        if (remaining <= 0) {
            clearTimeout(timeout);
            timeout = null;
            previous = now;
            result = func.apply(context, args);
        }
        else if (!timeout && options.trailing !== false) {
            timeout = setTimeout(later, remaining);
        }
        return result;
    };
}
/**
 * Throttle decorator
 *
 *  class MyClass {
 *    throttleable(10)
 *    myFn() { ... }
 *  }
 *
 * @export
 * @param {?} duration
 * @param {?=} options
 * @return {?}
 */
function throttleable(duration, options) {
    return function innerDecorator(target, key, descriptor) {
        return {
            configurable: true,
            enumerable: descriptor.enumerable,
            get: function getter() {
                Object.defineProperty(this, key, {
                    configurable: true,
                    enumerable: descriptor.enumerable,
                    value: throttle(descriptor.value, duration, options)
                });
                return this[key];
            }
        };
    };
}
var ColumnMode = {};
ColumnMode.standard = ('standard');
ColumnMode.flex = ('flex');
ColumnMode.force = ('force');
ColumnMode[ColumnMode.standard] = "standard";
ColumnMode[ColumnMode.flex] = "flex";
ColumnMode[ColumnMode.force] = "force";
var SortType = {};
SortType.single = ('single');
SortType.multi = ('multi');
SortType[SortType.single] = "single";
SortType[SortType.multi] = "multi";
var SortDirection = {};
SortDirection.asc = ('asc');
SortDirection.desc = ('desc');
SortDirection[SortDirection.asc] = "asc";
SortDirection[SortDirection.desc] = "desc";
var SelectionType = {};
SelectionType.single = ('single');
SelectionType.multi = ('multi');
SelectionType.multiClick = ('multiClick');
SelectionType.cell = ('cell');
SelectionType.checkbox = ('checkbox');
SelectionType[SelectionType.single] = "single";
SelectionType[SelectionType.multi] = "multi";
SelectionType[SelectionType.multiClick] = "multiClick";
SelectionType[SelectionType.cell] = "cell";
SelectionType[SelectionType.checkbox] = "checkbox";
var ClickType = {};
ClickType.single = ('single');
ClickType.double = ('double');
ClickType[ClickType.single] = "single";
ClickType[ClickType.double] = "double";
/**
 * Gets the next sort direction
 * @param {?} sortType
 * @param {?} current
 * @return {?}
 */
function nextSortDir(sortType, current) {
    if (sortType === SortType.single) {
        if (current === SortDirection.asc) {
            return SortDirection.desc;
        }
        else {
            return SortDirection.asc;
        }
    }
    else {
        if (!current) {
            return SortDirection.asc;
        }
        else if (current === SortDirection.asc) {
            return SortDirection.desc;
        }
        else if (current === SortDirection.desc) {
            return undefined;
        }
    }
}
/**
 * Adapted from fueld-ui on 6/216
 * https://github.com/FuelInteractive/fuel-ui/tree/master/src/pipes/OrderBy
 * @param {?} a
 * @param {?} b
 * @return {?}
 */
function orderByComparator(a, b) {
    if (a === null || typeof a === 'undefined')
        a = 0;
    if (b === null || typeof b === 'undefined')
        b = 0;
    if ((isNaN(parseFloat(a)) || !isFinite(a)) || (isNaN(parseFloat(b)) || !isFinite(b))) {
        // Convert to string in case of a=0 or b=0
        a = String(a);
        b = String(b);
        // Isn't a number so lowercase the string to properly compare
        if (a.toLowerCase() < b.toLowerCase())
            return -1;
        if (a.toLowerCase() > b.toLowerCase())
            return 1;
    }
    else {
        // Parse strings as numbers to compare properly
        if (parseFloat(a) < parseFloat(b))
            return -1;
        if (parseFloat(a) > parseFloat(b))
            return 1;
    }
    // equal each other
    return 0;
}
/**
 * Sorts the rows
 *
 * @export
 * @param {?} rows
 * @param {?} columns
 * @param {?} dirs
 * @return {?}
 */
function sortRows(rows, columns, dirs) {
    if (!rows || !dirs || !dirs.length || !columns)
        return rows;
    var /** @type {?} */ temp = rows.slice();
    var /** @type {?} */ cols = columns.reduce(function (obj, col) {
        if (col.comparator && typeof col.comparator === 'function') {
            obj[col.prop] = col.comparator;
        }
        return obj;
    }, {});
    // cache valueGetter and compareFn so that they
    // do not need to be looked-up in the sort function body
    var /** @type {?} */ cachedDirs = dirs.map(function (dir) {
        var /** @type {?} */ prop = dir.prop;
        return {
            prop: prop,
            dir: dir.dir,
            valueGetter: getterForProp(prop),
            compareFn: cols[prop] || orderByComparator
        };
    });
    return temp.sort(function (a, b) {
        for (var _i = 0, cachedDirs_1 = cachedDirs; _i < cachedDirs_1.length; _i++) {
            var cachedDir = cachedDirs_1[_i];
            var prop = cachedDir.prop, valueGetter = cachedDir.valueGetter;
            var /** @type {?} */ propA = valueGetter(a, prop);
            var /** @type {?} */ propB = valueGetter(b, prop);
            var /** @type {?} */ comparison = cachedDir.dir !== SortDirection.desc ?
                cachedDir.compareFn(propA, propB) :
                -cachedDir.compareFn(propA, propB);
            // Don't return 0 yet in case of needing to sort by next property
            if (comparison !== 0)
                return comparison;
        }
        // equal each other
        return 0;
    });
}
/**
 * This object contains the cache of the various row heights that are present inside
 * the data table.   Its based on Fenwick tree data structure that helps with
 * querying sums that have time complexity of log n.
 *
 * Fenwick Tree Credits: http://petr-mitrichev.blogspot.com/2013/05/fenwick-tree-range-updates.html
 * https://github.com/mikolalysenko/fenwick-tree
 *
 */
var RowHeightCache = /** @class */ (function () {
    function RowHeightCache() {
        /**
         * Tree Array stores the cumulative information of the row heights to perform efficient
         * range queries and updates.  Currently the tree is initialized to the base row
         * height instead of the detail row height.
         */
        this.treeArray = [];
    }
    /**
     * Clear the Tree array.
     * @return {?}
     */
    RowHeightCache.prototype.clearCache = function () {
        this.treeArray = [];
    };
    /**
     * Initialize the Fenwick tree with row Heights.
     *
     * @param {?} rows The array of rows which contain the expanded status.
     * @param {?} rowHeight The row height.
     * @param {?} detailRowHeight The detail row height.
     * @return {?}
     */
    RowHeightCache.prototype.initCache = function (rows, rowHeight, detailRowHeight) {
        var /** @type {?} */ isFn = typeof rowHeight === 'function';
        var /** @type {?} */ isDetailFn = typeof detailRowHeight === 'function';
        if (!isFn && isNaN(rowHeight)) {
            throw new Error("Row Height cache initialization failed. Please ensure that 'rowHeight' is a\n        valid number or function value: (" + rowHeight + ") when 'scrollbarV' is enabled.");
        }
        // Add this additional guard in case detailRowHeight is set to 'auto' as it wont work.
        if (!isDetailFn && isNaN(detailRowHeight)) {
            throw new Error("Row Height cache initialization failed. Please ensure that 'detailRowHeight' is a\n        valid number or function value: (" + detailRowHeight + ") when 'scrollbarV' is enabled.");
        }
        var /** @type {?} */ n = rows.length;
        this.treeArray = new Array(n);
        for (var /** @type {?} */ i = 0; i < n; ++i) {
            this.treeArray[i] = 0;
        }
        for (var /** @type {?} */ i = 0; i < n; ++i) {
            var /** @type {?} */ row = rows[i];
            var /** @type {?} */ currentRowHeight = rowHeight;
            if (isFn) {
                currentRowHeight = rowHeight(row);
            }
            // Add the detail row height to the already expanded rows.
            // This is useful for the table that goes through a filter or sort.
            if (row && row.$$expanded === 1) {
                if (isDetailFn) {
                    currentRowHeight += detailRowHeight(row, row.$$index);
                }
                else {
                    currentRowHeight += detailRowHeight;
                }
            }
            this.update(i, currentRowHeight);
        }
    };
    /**
     * Given the ScrollY position i.e. sum, provide the rowIndex
     * that is present in the current view port.  Below handles edge cases.
     *
     * @param {?} scrollY - The scrollY position.
     * @return {?}
     */
    RowHeightCache.prototype.getRowIndex = function (scrollY) {
        if (scrollY === 0)
            return 0;
        return this.calcRowIndex(scrollY);
    };
    /**
     * When a row is expanded or rowHeight is changed, update the height.  This can
     * be utilized in future when Angular Data table supports dynamic row heights.
     *
     *
     * @param {?} atRowIndex Update the data at this index row in the grid.
     * @param {?} byRowHeight Update by the rowHeight provided.
     * @return {?}
     */
    RowHeightCache.prototype.update = function (atRowIndex, byRowHeight) {
        if (!this.treeArray.length) {
            throw new Error("Update at index " + atRowIndex + " with value " + byRowHeight + " failed:\n        Row Height cache not initialized.");
        }
        var /** @type {?} */ n = this.treeArray.length;
        atRowIndex |= 0;
        while (atRowIndex < n) {
            this.treeArray[atRowIndex] += byRowHeight;
            atRowIndex |= (atRowIndex + 1);
        }
    };
    /**
     * Range Sum query from 1 to the rowIndex
     *
     * @param {?} atIndex The row index until which the total height needs to be obtained.
     * @return {?}
     */
    RowHeightCache.prototype.query = function (atIndex) {
        if (!this.treeArray.length) {
            throw new Error("query at index " + atIndex + " failed: Fenwick tree array not initialized.");
        }
        var /** @type {?} */ sum = 0;
        atIndex |= 0;
        while (atIndex >= 0) {
            sum += this.treeArray[atIndex];
            atIndex = (atIndex & (atIndex + 1)) - 1;
        }
        return sum;
    };
    /**
     * Find the total height between 2 row indexes
     * @param {?} atIndexA The row index from
     * @param {?} atIndexB The row index to
     * @return {?}
     */
    RowHeightCache.prototype.queryBetween = function (atIndexA, atIndexB) {
        return this.query(atIndexB) - this.query(atIndexA - 1);
    };
    /**
     * Given the ScrollY position i.e. sum, provide the rowIndex
     * that is present in the current view port.
     *
     * @param {?} sum - The scrollY position.
     * @return {?}
     */
    RowHeightCache.prototype.calcRowIndex = function (sum) {
        if (!this.treeArray.length)
            return 0;
        var /** @type {?} */ pos = -1;
        var /** @type {?} */ dataLength = this.treeArray.length;
        // Get the highest bit for the block size.
        var /** @type {?} */ highestBit = Math.pow(2, dataLength.toString(2).length - 1);
        for (var /** @type {?} */ blockSize = highestBit; blockSize !== 0; blockSize >>= 1) {
            var /** @type {?} */ nextPos = pos + blockSize;
            if (nextPos < dataLength && sum >= this.treeArray[nextPos]) {
                sum -= this.treeArray[nextPos];
                pos = nextPos;
            }
        }
        return pos + 1;
    };
    return RowHeightCache;
}());
/**
 * Sets the column defaults
 *
 * @export
 * @param {?} columns
 * @return {?}
 */
function setColumnDefaults(columns) {
    if (!columns)
        return;
    for (var _i = 0, columns_4 = columns; _i < columns_4.length; _i++) {
        var column = columns_4[_i];
        if (!column.$$id) {
            column.$$id = id();
        }
        // prop can be numeric; zero is valid not a missing prop
        // translate name => prop
        if (column.prop == null && column.name) {
            column.prop = camelCase(column.name);
        }
        if (!column.$$valueGetter) {
            column.$$valueGetter = getterForProp(column.prop);
        }
        // format props if no name passed
        if (column.prop != null && !column.name) {
            column.name = deCamelCase(String(column.prop));
        }
        if (!column.hasOwnProperty('resizeable')) {
            column.resizeable = true;
        }
        if (!column.hasOwnProperty('sortable')) {
            column.sortable = true;
        }
        if (!column.hasOwnProperty('draggable')) {
            column.draggable = true;
        }
        if (!column.hasOwnProperty('canAutoResize')) {
            column.canAutoResize = true;
        }
        if (!column.hasOwnProperty('width')) {
            column.width = 150;
        }
    }
}
/**
 * Translates templates definitions to objects
 *
 * @export
 * @param {?} templates
 * @return {?}
 */
function translateTemplates(templates) {
    var /** @type {?} */ result = [];
    for (var _i = 0, templates_1 = templates; _i < templates_1.length; _i++) {
        var temp = templates_1[_i];
        var /** @type {?} */ col = {};
        var /** @type {?} */ props = Object.getOwnPropertyNames(temp);
        for (var _a = 0, props_1 = props; _a < props_1.length; _a++) {
            var prop = props_1[_a];
            col[prop] = temp[prop];
        }
        if (temp.headerTemplate) {
            col.headerTemplate = temp.headerTemplate;
        }
        if (temp.cellTemplate) {
            col.cellTemplate = temp.cellTemplate;
        }
        result.push(col);
    }
    return result;
}
if (!document.elementsFromPoint) {
    document.elementsFromPoint = elementsFromPoint;
}
/**
 * Polyfill for `elementsFromPoint`
 *
 * https://developer.mozilla.org/en-US/docs/Web/API/Document/elementsFromPoint
 * https://gist.github.com/iddan/54d5d9e58311b0495a91bf06de661380
 * https://gist.github.com/oslego/7265412
 *
 * @export
 * @param {?} x
 * @param {?} y
 * @return {?}
 */
function elementsFromPoint(x, y) {
    var /** @type {?} */ elements = [];
    var /** @type {?} */ previousPointerEvents = [];
    var /** @type {?} */ current; // TODO: window.getComputedStyle should be used with inferred type (Element)
    var /** @type {?} */ i;
    var /** @type {?} */ d;
    // get all elements via elementFromPoint, and remove them from hit-testing in order
    while ((current = document.elementFromPoint(x, y)) && elements.indexOf(current) === -1 && current != null) {
        // push the element and its current style
        elements.push(current);
        previousPointerEvents.push({
            value: current.style.getPropertyValue('pointer-events'),
            priority: current.style.getPropertyPriority('pointer-events')
        });
        // add "pointer-events: none", to get to the underlying element
        current.style.setProperty('pointer-events', 'none', 'important');
    }
    // restore the previous pointer-events values
    for (i = previousPointerEvents.length; d = previousPointerEvents[--i];) {
        elements[i].style.setProperty('pointer-events', d.value ? d.value : '', d.priority);
    }
    // return our results
    return elements;
}
/*tslint:enable*/
/**
 * Gets the width of the scrollbar.  Nesc for windows
 * http://stackoverflow.com/a/13382873/888165
 *
 * @export
 * \@class ScrollbarHelper
 */
var ScrollbarHelper = /** @class */ (function () {
    /**
     * @param {?} document
     */
    function ScrollbarHelper(document) {
        this.document = document;
        this.width = this.getWidth();
    }
    /**
     * @return {?}
     */
    ScrollbarHelper.prototype.getWidth = function () {
        var /** @type {?} */ outer = this.document.createElement('div');
        outer.style.visibility = 'hidden';
        outer.style.width = '100px';
        outer.style.msOverflowStyle = 'scrollbar';
        this.document.body.appendChild(outer);
        var /** @type {?} */ widthNoScroll = outer.offsetWidth;
        outer.style.overflow = 'scroll';
        var /** @type {?} */ inner = this.document.createElement('div');
        inner.style.width = '100%';
        outer.appendChild(inner);
        var /** @type {?} */ widthWithScroll = inner.offsetWidth;
        outer.parentNode.removeChild(outer);
        return widthNoScroll - widthWithScroll;
    };
    return ScrollbarHelper;
}());
ScrollbarHelper.decorators = [
    { type: Injectable },
];
/**
 * @nocollapse
 */
ScrollbarHelper.ctorParameters = function () { return [
    { type: undefined, decorators: [{ type: Inject, args: [DOCUMENT,] },] },
]; };
var ScrollerComponent = /** @class */ (function () {
    /**
     * @param {?} element
     * @param {?} renderer
     */
    function ScrollerComponent(element, renderer) {
        this.renderer = renderer;
        this.scrollbarV = false;
        this.scrollbarH = false;
        this.scroll = new EventEmitter();
        this.scrollYPos = 0;
        this.scrollXPos = 0;
        this.prevScrollYPos = 0;
        this.prevScrollXPos = 0;
        this.element = element.nativeElement;
    }
    /**
     * @return {?}
     */
    ScrollerComponent.prototype.ngOnInit = function () {
        // manual bind so we don't always listen
        if (this.scrollbarV || this.scrollbarH) {
            this.parentElement = this.element.parentElement.parentElement;
            this.onScrollListener = this.renderer.listen(this.parentElement, 'scroll', this.onScrolled.bind(this));
        }
    };
    /**
     * @return {?}
     */
    ScrollerComponent.prototype.ngOnDestroy = function () {
        if (this.scrollbarV || this.scrollbarH) {
            this.onScrollListener();
        }
    };
    /**
     * @param {?} offsetY
     * @return {?}
     */
    ScrollerComponent.prototype.setOffset = function (offsetY) {
        if (this.parentElement) {
            this.parentElement.scrollTop = offsetY;
        }
    };
    /**
     * @param {?} event
     * @return {?}
     */
    ScrollerComponent.prototype.onScrolled = function (event) {
        var /** @type {?} */ dom = (event.currentTarget);
        this.scrollYPos = dom.scrollTop;
        this.scrollXPos = dom.scrollLeft;
        requestAnimationFrame(this.updateOffset.bind(this));
    };
    /**
     * @return {?}
     */
    ScrollerComponent.prototype.updateOffset = function () {
        var /** @type {?} */ direction;
        if (this.scrollYPos < this.prevScrollYPos) {
            direction = 'down';
        }
        else if (this.scrollYPos > this.prevScrollYPos) {
            direction = 'up';
        }
        this.scroll.emit({
            direction: direction,
            scrollYPos: this.scrollYPos,
            scrollXPos: this.scrollXPos
        });
        this.prevScrollYPos = this.scrollYPos;
        this.prevScrollXPos = this.scrollXPos;
    };
    return ScrollerComponent;
}());
ScrollerComponent.decorators = [
    { type: Component, args: [{
                selector: 'datatable-scroller',
                template: "\n    <ng-content></ng-content>\n  ",
                host: {
                    class: 'datatable-scroll'
                }
            },] },
];
/**
 * @nocollapse
 */
ScrollerComponent.ctorParameters = function () { return [
    { type: ElementRef, },
    { type: Renderer, },
]; };
ScrollerComponent.propDecorators = {
    'scrollbarV': [{ type: Input },],
    'scrollbarH': [{ type: Input },],
    'scrollHeight': [{ type: HostBinding, args: ['style.height.px',] }, { type: Input },],
    'scrollWidth': [{ type: HostBinding, args: ['style.width.px',] }, { type: Input },],
    'scroll': [{ type: Output },],
};
var DataTableBodyComponent = /** @class */ (function () {
    /**
     * Creates an instance of DataTableBodyComponent.
     *
     * \@memberOf DataTableBodyComponent
     */
    function DataTableBodyComponent() {
        var _this = this;
        this.selected = [];
        this.scroll = new EventEmitter();
        this.page = new EventEmitter();
        this.activate = new EventEmitter();
        this.select = new EventEmitter();
        this.detailToggle = new EventEmitter();
        this.rowContextmenu = new EventEmitter(false);
        this.rowHeightsCache = new RowHeightCache();
        this.temp = [];
        this.offsetY = 0;
        this.indexes = {};
        /**
         * Get the height of the detail row.
         *
         *
         * \@memberOf DataTableBodyComponent
         */
        this.getDetailRowHeight = function (row, index) {
            if (!_this.rowDetail)
                return 0;
            var rowHeight = _this.rowDetail.rowHeight;
            return typeof rowHeight === 'function' ? rowHeight(row, index) : rowHeight;
        };
        // declare fn here so we can get access to the `this` property
        this.rowTrackingFn = function (index, row) {
            if (this.trackByProp) {
                return row.$$index + "-" + this.trackByProp;
            }
            else {
                return row.$$index;
            }
        }.bind(this);
    }
    Object.defineProperty(DataTableBodyComponent.prototype, "pageSize", {
        /**
         * @return {?}
         */
        get: function () {
            return this._pageSize;
        },
        /**
         * @param {?} val
         * @return {?}
         */
        set: function (val) {
            this._pageSize = val;
            this.recalcLayout();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DataTableBodyComponent.prototype, "rows", {
        /**
         * @return {?}
         */
        get: function () {
            return this._rows;
        },
        /**
         * @param {?} val
         * @return {?}
         */
        set: function (val) {
            this._rows = val;
            this.recalcLayout();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DataTableBodyComponent.prototype, "columns", {
        /**
         * @return {?}
         */
        get: function () {
            return this._columns;
        },
        /**
         * @param {?} val
         * @return {?}
         */
        set: function (val) {
            this._columns = val;
            var /** @type {?} */ colsByPin = columnsByPin(val);
            this.columnGroupWidths = columnGroupWidths(colsByPin, val);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DataTableBodyComponent.prototype, "offset", {
        /**
         * @return {?}
         */
        get: function () {
            return this._offset;
        },
        /**
         * @param {?} val
         * @return {?}
         */
        set: function (val) {
            this._offset = val;
            this.recalcLayout();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DataTableBodyComponent.prototype, "rowCount", {
        /**
         * @return {?}
         */
        get: function () {
            return this._rowCount;
        },
        /**
         * @param {?} val
         * @return {?}
         */
        set: function (val) {
            this._rowCount = val;
            this.recalcLayout();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DataTableBodyComponent.prototype, "bodyWidth", {
        /**
         * @return {?}
         */
        get: function () {
            if (this.scrollbarH) {
                return this.innerWidth + 'px';
            }
            else {
                return '100%';
            }
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DataTableBodyComponent.prototype, "bodyHeight", {
        /**
         * @return {?}
         */
        get: function () {
            return this._bodyHeight;
        },
        /**
         * @param {?} val
         * @return {?}
         */
        set: function (val) {
            if (this.scrollbarV) {
                this._bodyHeight = val + 'px';
            }
            else {
                this._bodyHeight = 'auto';
            }
            this.recalcLayout();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DataTableBodyComponent.prototype, "selectEnabled", {
        /**
         * Returns if selection is enabled.
         *
         * \@readonly
         * \@memberOf DataTableBodyComponent
         * @return {?}
         */
        get: function () {
            return !!this.selectionType;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DataTableBodyComponent.prototype, "scrollHeight", {
        /**
         * Property that would calculate the height of scroll bar
         * based on the row heights cache for virtual scroll. Other scenarios
         * calculate scroll height automatically (as height will be undefined).
         *
         * \@readonly
         * \@memberOf DataTableBodyComponent
         * @return {?}
         */
        get: function () {
            if (this.scrollbarV) {
                return this.rowHeightsCache.query(this.rowCount - 1);
            }
        },
        enumerable: true,
        configurable: true
    });
    /**
     * Called after the constructor, initializing input properties
     *
     * \@memberOf DataTableBodyComponent
     * @return {?}
     */
    DataTableBodyComponent.prototype.ngOnInit = function () {
        var _this = this;
        if (this.rowDetail) {
            this.listener = this.rowDetail.toggle
                .subscribe(function (_a) {
                var type = _a.type, value = _a.value;
                if (type === 'row')
                    _this.toggleRowExpansion(value);
                if (type === 'all')
                    _this.toggleAllRows(value);
            });
        }
    };
    /**
     * Called once, before the instance is destroyed.
     *
     * \@memberOf DataTableBodyComponent
     * @return {?}
     */
    DataTableBodyComponent.prototype.ngOnDestroy = function () {
        if (this.rowDetail)
            this.listener.unsubscribe();
    };
    /**
     * Updates the Y offset given a new offset.
     *
     *
     * \@memberOf DataTableBodyComponent
     * @param {?=} offset
     * @return {?}
     */
    DataTableBodyComponent.prototype.updateOffsetY = function (offset) {
        // scroller is missing on empty table
        if (!this.scroller)
            return;
        if (this.scrollbarV && offset) {
            // First get the row Index that we need to move to.
            var /** @type {?} */ rowIndex = this.pageSize * offset;
            offset = this.rowHeightsCache.query(rowIndex - 1);
        }
        this.scroller.setOffset(offset || 0);
    };
    /**
     * Body was scrolled, this is mainly useful for
     * when a user is server-side pagination via virtual scroll.
     *
     *
     * \@memberOf DataTableBodyComponent
     * @param {?} event
     * @return {?}
     */
    DataTableBodyComponent.prototype.onBodyScroll = function (event) {
        var /** @type {?} */ scrollYPos = event.scrollYPos;
        var /** @type {?} */ scrollXPos = event.scrollXPos;
        // if scroll change, trigger update
        // this is mainly used for header cell positions
        if (this.offsetY !== scrollYPos || this.offsetX !== scrollXPos) {
            this.scroll.emit({
                offsetY: scrollYPos,
                offsetX: scrollXPos
            });
        }
        this.offsetY = scrollYPos;
        this.offsetX = scrollXPos;
        this.updateIndexes();
        this.updatePage(event.direction);
        this.updateRows();
    };
    /**
     * Updates the page given a direction.
     *
     *
     * \@memberOf DataTableBodyComponent
     * @param {?} direction
     * @return {?}
     */
    DataTableBodyComponent.prototype.updatePage = function (direction) {
        var /** @type {?} */ offset = this.indexes.first / this.pageSize;
        if (direction === 'up') {
            offset = Math.ceil(offset);
        }
        else if (direction === 'down') {
            offset = Math.ceil(offset);
        }
        if (direction !== undefined && !isNaN(offset)) {
            this.page.emit({ offset: offset });
        }
    };
    /**
     * Updates the rows in the view port
     *
     * \@memberOf DataTableBodyComponent
     * @return {?}
     */
    DataTableBodyComponent.prototype.updateRows = function () {
        var _a = this.indexes, first = _a.first, last = _a.last;
        var /** @type {?} */ rowIndex = first;
        var /** @type {?} */ idx = 0;
        var /** @type {?} */ temp = [];
        while (rowIndex < last && rowIndex < this.rowCount) {
            var /** @type {?} */ row = this.rows[rowIndex];
            if (row) {
                row.$$index = rowIndex;
                temp[idx] = row;
            }
            idx++;
            rowIndex++;
        }
        this.temp = temp;
    };
    /**
     * Get the row height
     *
     *
     * \@memberOf DataTableBodyComponent
     * @param {?} row
     * @return {?}
     */
    DataTableBodyComponent.prototype.getRowHeight = function (row) {
        var /** @type {?} */ rowHeight = this.rowHeight;
        // if its a function return it
        if (typeof this.rowHeight === 'function') {
            rowHeight = this.rowHeight(row);
        }
        return rowHeight;
    };
    /**
     * Calculate row height based on the expanded state of the row.
     *
     *
     * \@memberOf DataTableBodyComponent
     * @param {?} row
     * @return {?}
     */
    DataTableBodyComponent.prototype.getRowAndDetailHeight = function (row) {
        var /** @type {?} */ rowHeight = this.getRowHeight(row);
        // Adding detail row height if its expanded.
        if (row.$$expanded === 1) {
            rowHeight += this.getDetailRowHeight(row);
        }
        return rowHeight;
    };
    /**
     * Calculates the styles for the row so that the rows can be moved in 2D space
     * during virtual scroll inside the DOM.   In the below case the Y position is
     * manipulated.   As an example, if the height of row 0 is 30 px and row 1 is
     * 100 px then following styles are generated:
     *
     * transform: translate3d(0px, 0px, 0px);    ->  row0
     * transform: translate3d(0px, 30px, 0px);   ->  row1
     * transform: translate3d(0px, 130px, 0px);  ->  row2
     *
     * Row heights have to be calculated based on the row heights cache as we wont
     * be able to determine which row is of what height before hand.  In the above
     * case the positionY of the translate3d for row2 would be the sum of all the
     * heights of the rows before it (i.e. row0 and row1).
     *
     *
     * \@memberOf DataTableBodyComponent
     * @param {?} row
     * @return {?}
     */
    DataTableBodyComponent.prototype.getRowsStyles = function (row) {
        var /** @type {?} */ rowHeight = this.getRowAndDetailHeight(row);
        var /** @type {?} */ styles = {
            height: rowHeight + 'px'
        };
        if (this.scrollbarV) {
            var /** @type {?} */ idx = row ? row.$$index : 0;
            // const pos = idx * rowHeight;
            // The position of this row would be the sum of all row heights
            // until the previous row position.
            var /** @type {?} */ pos = this.rowHeightsCache.query(idx - 1);
            translateXY(styles, 0, pos);
        }
        return styles;
    };
    /**
     * Hides the loading indicator
     *
     *
     * \@memberOf DataTableBodyComponent
     * @return {?}
     */
    DataTableBodyComponent.prototype.hideIndicator = function () {
        var _this = this;
        setTimeout(function () { return _this.loadingIndicator = false; }, 500);
    };
    /**
     * Updates the index of the rows in the viewport
     *
     * \@memberOf DataTableBodyComponent
     * @return {?}
     */
    DataTableBodyComponent.prototype.updateIndexes = function () {
        var /** @type {?} */ first = 0;
        var /** @type {?} */ last = 0;
        if (this.scrollbarV) {
            // Calculation of the first and last indexes will be based on where the
            // scrollY position would be at.  The last index would be the one
            // that shows up inside the view port the last.
            var /** @type {?} */ height = parseInt(this.bodyHeight, 0);
            first = this.rowHeightsCache.getRowIndex(this.offsetY);
            last = this.rowHeightsCache.getRowIndex(height + this.offsetY) + 1;
        }
        else {
            first = Math.max(this.offset * this.pageSize, 0);
            last = Math.min((first + this.pageSize), this.rowCount);
        }
        this.indexes = { first: first, last: last };
    };
    /**
     * Refreshes the full Row Height cache.  Should be used
     * when the entire row array state has changed.
     *
     *
     * \@memberOf DataTableBodyComponent
     * @return {?}
     */
    DataTableBodyComponent.prototype.refreshRowHeightCache = function () {
        if (!this.scrollbarV)
            return;
        // clear the previous row height cache if already present.
        // this is useful during sorts, filters where the state of the
        // rows array is changed.
        this.rowHeightsCache.clearCache();
        // Initialize the tree only if there are rows inside the tree.
        if (this.rows && this.rows.length) {
            this.rowHeightsCache.initCache(this.rows, this.rowHeight, this.getDetailRowHeight);
        }
    };
    /**
     * Gets the index for the view port
     *
     *
     * \@memberOf DataTableBodyComponent
     * @return {?}
     */
    DataTableBodyComponent.prototype.getAdjustedViewPortIndex = function () {
        // Capture the row index of the first row that is visible on the viewport.
        // If the scroll bar is just below the row which is highlighted then make that as the
        // first index.
        var /** @type {?} */ viewPortFirstRowIndex = this.indexes.first;
        if (this.scrollbarV) {
            var /** @type {?} */ offsetScroll = this.rowHeightsCache.query(viewPortFirstRowIndex - 1);
            return offsetScroll <= this.offsetY ? viewPortFirstRowIndex - 1 : viewPortFirstRowIndex;
        }
        return viewPortFirstRowIndex;
    };
    /**
     * Toggle the Expansion of the row i.e. if the row is expanded then it will
     * collapse and vice versa.   Note that the expanded status is stored as
     * a part of the row object itself as we have to preserve the expanded row
     * status in case of sorting and filtering of the row set.
     *
     *
     * \@memberOf DataTableBodyComponent
     * @param {?} row
     * @return {?}
     */
    DataTableBodyComponent.prototype.toggleRowExpansion = function (row) {
        // Capture the row index of the first row that is visible on the viewport.
        var /** @type {?} */ viewPortFirstRowIndex = this.getAdjustedViewPortIndex();
        // If the detailRowHeight is auto --> only in case of non-virtualized scroll
        if (this.scrollbarV) {
            var /** @type {?} */ detailRowHeight = this.getDetailRowHeight(row) * (row.$$expanded ? -1 : 1);
            this.rowHeightsCache.update(row.$$index, detailRowHeight);
        }
        // Update the toggled row and update the heights in the cache.
        row.$$expanded ^= 1;
        this.detailToggle.emit({
            rows: [row],
            currentIndex: viewPortFirstRowIndex
        });
    };
    /**
     * Expand/Collapse all the rows no matter what their state is.
     *
     *
     * \@memberOf DataTableBodyComponent
     * @param {?} expanded
     * @return {?}
     */
    DataTableBodyComponent.prototype.toggleAllRows = function (expanded) {
        var /** @type {?} */ rowExpanded = expanded ? 1 : 0;
        // Capture the row index of the first row that is visible on the viewport.
        var /** @type {?} */ viewPortFirstRowIndex = this.getAdjustedViewPortIndex();
        for (var _i = 0, _a = this.rows; _i < _a.length; _i++) {
            var row = _a[_i];
            row.$$expanded = rowExpanded;
        }
        if (this.scrollbarV) {
            // Refresh the full row heights cache since every row was affected.
            this.recalcLayout();
        }
        // Emit all rows that have been expanded.
        this.detailToggle.emit({
            rows: this.rows,
            currentIndex: viewPortFirstRowIndex
        });
    };
    /**
     * Recalculates the table
     *
     * \@memberOf DataTableBodyComponent
     * @return {?}
     */
    DataTableBodyComponent.prototype.recalcLayout = function () {
        this.refreshRowHeightCache();
        this.updateIndexes();
        this.updateRows();
    };
    return DataTableBodyComponent;
}());
DataTableBodyComponent.decorators = [
    { type: Component, args: [{
                selector: 'datatable-body',
                template: "\n    <datatable-selection\n      #selector\n      [selected]=\"selected\"\n      [rows]=\"temp\"\n      [selectCheck]=\"selectCheck\"\n      [selectEnabled]=\"selectEnabled\"\n      [selectionType]=\"selectionType\"\n      [rowIdentity]=\"rowIdentity\"\n      (select)=\"select.emit($event)\"\n      (activate)=\"activate.emit($event)\">\n      <datatable-progress\n        *ngIf=\"loadingIndicator\">\n      </datatable-progress>\n      <datatable-scroller\n        *ngIf=\"rows?.length\"\n        [scrollbarV]=\"scrollbarV\"\n        [scrollbarH]=\"scrollbarH\"\n        [scrollHeight]=\"scrollHeight\"\n        [scrollWidth]=\"columnGroupWidths.total\"\n        (scroll)=\"onBodyScroll($event)\">\n        <datatable-row-wrapper\n          *ngFor=\"let row of temp; let i = index; trackBy: rowTrackingFn;\"\n          [ngStyle]=\"getRowsStyles(row)\"\n          [rowDetail]=\"rowDetail\"\n          [detailRowHeight]=\"getDetailRowHeight(row,i)\"\n          [row]=\"row\"\n          [expanded]=\"row.$$expanded === 1\"\n          (rowContextmenu)=\"rowContextmenu.emit($event)\">\n          <datatable-body-row\n            tabindex=\"-1\"\n            [isSelected]=\"selector.getRowSelected(row)\"\n            [innerWidth]=\"innerWidth\"\n            [offsetX]=\"offsetX\"\n            [columns]=\"columns\"\n            [rowHeight]=\"getRowHeight(row)\"\n            [row]=\"row\"\n            [rowClass]=\"rowClass\"\n            (activate)=\"selector.onActivate($event, i)\">\n          </datatable-body-row>\n        </datatable-row-wrapper>\n      </datatable-scroller>\n      <div\n        class=\"empty-row\"\n        *ngIf=\"!rows?.length\"\n        [innerHTML]=\"emptyMessage\">\n      </div>\n    </datatable-selection>\n  ",
                host: {
                    class: 'datatable-body'
                }
            },] },
];
/**
 * @nocollapse
 */
DataTableBodyComponent.ctorParameters = function () { return []; };
DataTableBodyComponent.propDecorators = {
    'scrollbarV': [{ type: Input },],
    'scrollbarH': [{ type: Input },],
    'loadingIndicator': [{ type: Input },],
    'rowHeight': [{ type: Input },],
    'offsetX': [{ type: Input },],
    'emptyMessage': [{ type: Input },],
    'selectionType': [{ type: Input },],
    'selected': [{ type: Input },],
    'rowIdentity': [{ type: Input },],
    'rowDetail': [{ type: Input },],
    'selectCheck': [{ type: Input },],
    'trackByProp': [{ type: Input },],
    'rowClass': [{ type: Input },],
    'pageSize': [{ type: Input },],
    'rows': [{ type: Input },],
    'columns': [{ type: Input },],
    'offset': [{ type: Input },],
    'rowCount': [{ type: Input },],
    'innerWidth': [{ type: Input },],
    'bodyWidth': [{ type: HostBinding, args: ['style.width',] },],
    'bodyHeight': [{ type: Input }, { type: HostBinding, args: ['style.height',] },],
    'scroll': [{ type: Output },],
    'page': [{ type: Output },],
    'activate': [{ type: Output },],
    'select': [{ type: Output },],
    'detailToggle': [{ type: Output },],
    'rowContextmenu': [{ type: Output },],
    'scroller': [{ type: ViewChild, args: [ScrollerComponent,] },],
};
var DataTableBodyCellComponent = /** @class */ (function () {
    /**
     * @param {?} element
     */
    function DataTableBodyCellComponent(element) {
        this.activate = new EventEmitter();
        this.isFocused = false;
        this.element = element.nativeElement;
    }
    Object.defineProperty(DataTableBodyCellComponent.prototype, "sorts", {
        /**
         * @return {?}
         */
        get: function () {
            return this._sorts;
        },
        /**
         * @param {?} val
         * @return {?}
         */
        set: function (val) {
            this._sorts = val;
            this.calcSortDir = this.calcSortDir(val);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DataTableBodyCellComponent.prototype, "columnCssClasses", {
        /**
         * @return {?}
         */
        get: function () {
            var /** @type {?} */ cls = 'datatable-body-cell';
            if (this.column.cellClass) {
                if (typeof this.column.cellClass === 'string') {
                    cls += ' ' + this.column.cellClass;
                }
                else if (typeof this.column.cellClass === 'function') {
                    var /** @type {?} */ res = this.column.cellClass({
                        row: this.row,
                        column: this.column,
                        value: this.value
                    });
                    if (typeof res === 'string') {
                        cls += res;
                    }
                    else if (typeof res === 'object') {
                        var /** @type {?} */ keys = Object.keys(res);
                        for (var _i = 0, keys_1 = keys; _i < keys_1.length; _i++) {
                            var k = keys_1[_i];
                            if (res[k] === true)
                                cls += " " + k;
                        }
                    }
                }
            }
            if (!this.sortDir)
                cls += ' sort-active';
            if (this.isFocused)
                cls += ' active';
            if (this.sortDir === SortDirection.asc)
                cls += ' sort-asc';
            if (this.sortDir === SortDirection.desc)
                cls += ' sort-desc';
            return cls;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DataTableBodyCellComponent.prototype, "width", {
        /**
         * @return {?}
         */
        get: function () {
            return this.column.width;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DataTableBodyCellComponent.prototype, "height", {
        /**
         * @return {?}
         */
        get: function () {
            var /** @type {?} */ height = this.rowHeight;
            if (isNaN(height))
                return height;
            return height + 'px';
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DataTableBodyCellComponent.prototype, "value", {
        /**
         * @return {?}
         */
        get: function () {
            if (!this.row || !this.column)
                return '';
            var /** @type {?} */ val = this.column.$$valueGetter(this.row, this.column.prop);
            var /** @type {?} */ userPipe = this.column.pipe;
            if (userPipe)
                return userPipe.transform(val);
            if (val !== undefined)
                return val;
            return '';
        },
        enumerable: true,
        configurable: true
    });
    /**
     * @return {?}
     */
    DataTableBodyCellComponent.prototype.ngOnDestroy = function () {
        if (this.cellTemplate) {
            this.cellTemplate.clear();
        }
    };
    /**
     * @return {?}
     */
    DataTableBodyCellComponent.prototype.onFocus = function () {
        this.isFocused = true;
    };
    /**
     * @return {?}
     */
    DataTableBodyCellComponent.prototype.onBlur = function () {
        this.isFocused = false;
    };
    /**
     * @param {?} event
     * @return {?}
     */
    DataTableBodyCellComponent.prototype.onClick = function (event) {
        this.activate.emit({
            type: 'click',
            event: event,
            row: this.row,
            column: this.column,
            value: this.value,
            cellElement: this.element
        });
    };
    /**
     * @param {?} event
     * @return {?}
     */
    DataTableBodyCellComponent.prototype.onDblClick = function (event) {
        this.activate.emit({
            type: 'dblclick',
            event: event,
            row: this.row,
            column: this.column,
            value: this.value,
            cellElement: this.element
        });
    };
    /**
     * @param {?} event
     * @return {?}
     */
    DataTableBodyCellComponent.prototype.onKeyDown = function (event) {
        var /** @type {?} */ keyCode = event.keyCode;
        var /** @type {?} */ isTargetCell = event.target === this.element;
        var /** @type {?} */ isAction = keyCode === Keys.return ||
            keyCode === Keys.down ||
            keyCode === Keys.up ||
            keyCode === Keys.left ||
            keyCode === Keys.right;
        if (isAction && isTargetCell) {
            event.preventDefault();
            event.stopPropagation();
            this.activate.emit({
                type: 'keydown',
                event: event,
                row: this.row,
                column: this.column,
                value: this.value,
                cellElement: this.element
            });
        }
    };
    /**
     * @param {?} event
     * @return {?}
     */
    DataTableBodyCellComponent.prototype.onCheckboxChange = function (event) {
        this.activate.emit({
            type: 'checkbox',
            event: event,
            row: this.row,
            column: this.column,
            value: this.value,
            cellElement: this.element
        });
    };
    /**
     * @param {?} sorts
     * @return {?}
     */
    DataTableBodyCellComponent.prototype.calcSortDir = function (sorts) {
        var _this = this;
        if (!sorts)
            return;
        var /** @type {?} */ sort = sorts.find(function (s) {
            return s.prop === _this.column.prop;
        });
        if (sort)
            return sort.dir;
    };
    return DataTableBodyCellComponent;
}());
DataTableBodyCellComponent.decorators = [
    { type: Component, args: [{
                selector: 'datatable-body-cell',
                template: "\n    <div class=\"datatable-body-cell-label\">\n      <label\n        *ngIf=\"column.checkboxable\" \n        class=\"datatable-checkbox\">\n        <input \n          type=\"checkbox\"\n          [checked]=\"isSelected\"\n          (click)=\"onCheckboxChange($event)\" \n        />\n      </label>\n      <span\n        *ngIf=\"!column.cellTemplate\"\n        [title]=\"value\"\n        [innerHTML]=\"value\">\n      </span>\n      <ng-template #cellTemplate\n        *ngIf=\"column.cellTemplate\"\n        [ngTemplateOutlet]=\"column.cellTemplate\"\n        [ngOutletContext]=\"{ value: value, row: row, column: column }\">\n      </ng-template>\n    </div>\n  ",
                host: {
                    class: 'datatable-body-cell'
                }
            },] },
];
/**
 * @nocollapse
 */
DataTableBodyCellComponent.ctorParameters = function () { return [
    { type: ElementRef, },
]; };
DataTableBodyCellComponent.propDecorators = {
    'row': [{ type: Input },],
    'column': [{ type: Input },],
    'rowHeight': [{ type: Input },],
    'isSelected': [{ type: Input },],
    'sorts': [{ type: Input },],
    'activate': [{ type: Output },],
    'cellTemplate': [{ type: ViewChild, args: ['cellTemplate', { read: ViewContainerRef },] },],
    'columnCssClasses': [{ type: HostBinding, args: ['class',] },],
    'width': [{ type: HostBinding, args: ['style.width.px',] },],
    'height': [{ type: HostBinding, args: ['style.height',] },],
    'onFocus': [{ type: HostListener, args: ['focus',] },],
    'onBlur': [{ type: HostListener, args: ['blur',] },],
    'onClick': [{ type: HostListener, args: ['click', ['$event'],] },],
    'onDblClick': [{ type: HostListener, args: ['dblclick', ['$event'],] },],
    'onKeyDown': [{ type: HostListener, args: ['keydown', ['$event'],] },],
};
var DataTableBodyRowComponent = /** @class */ (function () {
    /**
     * @param {?} scrollbarHelper
     * @param {?} element
     */
    function DataTableBodyRowComponent(scrollbarHelper, element) {
        this.scrollbarHelper = scrollbarHelper;
        this.activate = new EventEmitter();
        this.element = element.nativeElement;
    }
    Object.defineProperty(DataTableBodyRowComponent.prototype, "columns", {
        /**
         * @return {?}
         */
        get: function () {
            return this._columns;
        },
        /**
         * @param {?} val
         * @return {?}
         */
        set: function (val) {
            this._columns = val;
            this.recalculateColumns(val);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DataTableBodyRowComponent.prototype, "innerWidth", {
        /**
         * @return {?}
         */
        get: function () {
            return this._innerWidth;
        },
        /**
         * @param {?} val
         * @return {?}
         */
        set: function (val) {
            this._innerWidth = val;
            this.recalculateColumns();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DataTableBodyRowComponent.prototype, "cssClass", {
        /**
         * @return {?}
         */
        get: function () {
            var /** @type {?} */ cls = 'datatable-body-row';
            if (this.isSelected)
                cls += ' active';
            if (this.row.$$index % 2 !== 0)
                cls += ' datatable-row-odd';
            if (this.row.$$index % 2 === 0)
                cls += ' datatable-row-even';
            if (this.rowClass) {
                var /** @type {?} */ res = this.rowClass(this.row);
                if (typeof res === 'string') {
                    cls += res;
                }
                else if (typeof res === 'object') {
                    var /** @type {?} */ keys = Object.keys(res);
                    for (var _i = 0, keys_2 = keys; _i < keys_2.length; _i++) {
                        var k = keys_2[_i];
                        if (res[k] === true)
                            cls += " " + k;
                    }
                }
            }
            return cls;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DataTableBodyRowComponent.prototype, "columnsTotalWidths", {
        /**
         * @return {?}
         */
        get: function () {
            return this.columnGroupWidths.total;
        },
        enumerable: true,
        configurable: true
    });
    /**
     * @param {?} index
     * @param {?} colGroup
     * @return {?}
     */
    DataTableBodyRowComponent.prototype.trackByGroups = function (index, colGroup) {
        return colGroup.type;
    };
    /**
     * @param {?} index
     * @param {?} column
     * @return {?}
     */
    DataTableBodyRowComponent.prototype.columnTrackingFn = function (index, column) {
        return column.$$id;
    };
    /**
     * @param {?} group
     * @return {?}
     */
    DataTableBodyRowComponent.prototype.stylesByGroup = function (group) {
        var /** @type {?} */ widths = this.columnGroupWidths;
        var /** @type {?} */ offsetX = this.offsetX;
        var /** @type {?} */ styles = {
            width: widths[group] + "px"
        };
        if (group === 'left') {
            translateXY(styles, offsetX, 0);
        }
        else if (group === 'right') {
            var /** @type {?} */ bodyWidth = parseInt(this.innerWidth + '', 0);
            var /** @type {?} */ totalDiff = widths.total - bodyWidth;
            var /** @type {?} */ offsetDiff = totalDiff - offsetX;
            var /** @type {?} */ offset = (offsetDiff + this.scrollbarHelper.width) * -1;
            translateXY(styles, offset, 0);
        }
        return styles;
    };
    /**
     * @param {?} event
     * @param {?} index
     * @return {?}
     */
    DataTableBodyRowComponent.prototype.onActivate = function (event, index) {
        event.cellIndex = index;
        event.rowElement = this.element;
        this.activate.emit(event);
    };
    /**
     * @param {?} event
     * @return {?}
     */
    DataTableBodyRowComponent.prototype.onKeyDown = function (event) {
        var /** @type {?} */ keyCode = event.keyCode;
        var /** @type {?} */ isTargetRow = event.target === this.element;
        var /** @type {?} */ isAction = keyCode === Keys.return ||
            keyCode === Keys.down ||
            keyCode === Keys.up ||
            keyCode === Keys.left ||
            keyCode === Keys.right;
        if (isAction && isTargetRow) {
            event.preventDefault();
            event.stopPropagation();
            this.activate.emit({
                type: 'keydown',
                event: event,
                row: this.row,
                rowElement: this.element
            });
        }
    };
    /**
     * @param {?=} val
     * @return {?}
     */
    DataTableBodyRowComponent.prototype.recalculateColumns = function (val) {
        if (val === void 0) { val = this.columns; }
        var /** @type {?} */ colsByPin = columnsByPin(val);
        this.columnsByPin = columnsByPinArr(val);
        this.columnGroupWidths = columnGroupWidths(colsByPin, val);
    };
    return DataTableBodyRowComponent;
}());
DataTableBodyRowComponent.decorators = [
    { type: Component, args: [{
                selector: 'datatable-body-row',
                template: "\n    <div\n      *ngFor=\"let colGroup of columnsByPin; let i = index; trackBy: trackByGroups\"\n      class=\"datatable-row-{{colGroup.type}} datatable-row-group\"\n      [ngStyle]=\"stylesByGroup(colGroup.type)\">\n      <datatable-body-cell\n        *ngFor=\"let column of colGroup.columns; let ii = index; trackBy: columnTrackingFn\"\n        tabindex=\"-1\"\n        [row]=\"row\"\n        [isSelected]=\"isSelected\"\n        [column]=\"column\"\n        [rowHeight]=\"rowHeight\"\n        (activate)=\"onActivate($event, ii)\">\n      </datatable-body-cell>\n    </div>\n  "
            },] },
];
/**
 * @nocollapse
 */
DataTableBodyRowComponent.ctorParameters = function () { return [
    { type: ScrollbarHelper, },
    { type: ElementRef, },
]; };
DataTableBodyRowComponent.propDecorators = {
    'columns': [{ type: Input },],
    'innerWidth': [{ type: Input },],
    'rowClass': [{ type: Input },],
    'row': [{ type: Input },],
    'offsetX': [{ type: Input },],
    'isSelected': [{ type: Input },],
    'cssClass': [{ type: HostBinding, args: ['class',] },],
    'rowHeight': [{ type: HostBinding, args: ['style.height.px',] }, { type: Input },],
    'columnsTotalWidths': [{ type: HostBinding, args: ['style.width.px',] },],
    'activate': [{ type: Output },],
    'onKeyDown': [{ type: HostListener, args: ['keydown', ['$event'],] },],
};
var ProgressBarComponent = /** @class */ (function () {
    function ProgressBarComponent() {
    }
    return ProgressBarComponent;
}());
ProgressBarComponent.decorators = [
    { type: Component, args: [{
                selector: 'datatable-progress',
                template: "\n    <div class=\"progress-linear\" role=\"progressbar\">\n      <div class=\"container\">\n        <div class=\"bar\"></div>\n      </div>\n    </div>\n  ",
                changeDetection: ChangeDetectionStrategy.OnPush
            },] },
];
/**
 * @nocollapse
 */
ProgressBarComponent.ctorParameters = function () { return []; };
var DataTableRowWrapperComponent = /** @class */ (function () {
    function DataTableRowWrapperComponent() {
        this.expanded = false;
        this.rowContextmenu = new EventEmitter(false);
    }
    /**
     * @param {?} $event
     * @return {?}
     */
    DataTableRowWrapperComponent.prototype.onContextmenu = function ($event) {
        this.rowContextmenu.emit({ event: $event, row: this.row });
    };
    return DataTableRowWrapperComponent;
}());
DataTableRowWrapperComponent.decorators = [
    { type: Component, args: [{
                selector: 'datatable-row-wrapper',
                template: "\n    <ng-content></ng-content>\n    <div \n      *ngIf=\"expanded\"\n      [style.height.px]=\"detailRowHeight\" \n      class=\"datatable-row-detail\">\n      <ng-template\n        *ngIf=\"rowDetail && rowDetail.template\"\n        [ngTemplateOutlet]=\"rowDetail.template\"\n        [ngOutletContext]=\"{ row: row }\">\n      </ng-template>\n    </div>\n  ",
                host: {
                    class: 'datatable-row-wrapper'
                }
            },] },
];
/**
 * @nocollapse
 */
DataTableRowWrapperComponent.ctorParameters = function () { return []; };
DataTableRowWrapperComponent.propDecorators = {
    'rowDetail': [{ type: Input },],
    'detailRowHeight': [{ type: Input },],
    'expanded': [{ type: Input },],
    'row': [{ type: Input },],
    'rowContextmenu': [{ type: Output },],
    'onContextmenu': [{ type: HostListener, args: ['contextmenu', ['$event'],] },],
};
var DataTableSelectionComponent = /** @class */ (function () {
    function DataTableSelectionComponent() {
        this.activate = new EventEmitter();
        this.select = new EventEmitter();
    }
    /**
     * @param {?} event
     * @param {?} index
     * @param {?} row
     * @return {?}
     */
    DataTableSelectionComponent.prototype.selectRow = function (event, index, row) {
        if (!this.selectEnabled)
            return;
        var /** @type {?} */ chkbox = this.selectionType === SelectionType.checkbox;
        var /** @type {?} */ multi = this.selectionType === SelectionType.multi;
        var /** @type {?} */ multiClick = this.selectionType === SelectionType.multiClick;
        var /** @type {?} */ selected = [];
        if (multi || chkbox || multiClick) {
            if (event.shiftKey) {
                selected = selectRowsBetween([], this.rows, index, this.prevIndex, this.getRowSelectedIdx.bind(this));
            }
            else if (event.ctrlKey || multiClick || chkbox) {
                selected = selectRows(this.selected.slice(), row, this.getRowSelectedIdx.bind(this));
            }
            else {
                selected = selectRows([], row, this.getRowSelectedIdx.bind(this));
            }
        }
        else {
            selected = selectRows([], row, this.getRowSelectedIdx.bind(this));
        }
        if (typeof this.selectCheck === 'function') {
            selected = selected.filter(this.selectCheck.bind(this));
        }
        this.selected.splice(0, this.selected.length);
        (_a = this.selected).push.apply(_a, selected);
        this.prevIndex = index;
        this.select.emit({
            selected: selected
        });
        var _a;
    };
    /**
     * @param {?} model
     * @param {?} index
     * @return {?}
     */
    DataTableSelectionComponent.prototype.onActivate = function (model, index) {
        var type = model.type, event = model.event, row = model.row;
        var /** @type {?} */ chkbox = this.selectionType === SelectionType.checkbox;
        var /** @type {?} */ select = (!chkbox && (type === 'click' || type === 'dblclick')) ||
            (chkbox && type === 'checkbox');
        if (select) {
            this.selectRow(event, index, row);
        }
        else if (type === 'keydown') {
            if (((event)).keyCode === Keys.return) {
                this.selectRow(event, index, row);
            }
            else {
                this.onKeyboardFocus(model);
            }
        }
        this.activate.emit(model);
    };
    /**
     * @param {?} model
     * @return {?}
     */
    DataTableSelectionComponent.prototype.onKeyboardFocus = function (model) {
        var keyCode = (model.event).keyCode;
        var /** @type {?} */ shouldFocus = keyCode === Keys.up ||
            keyCode === Keys.down ||
            keyCode === Keys.right ||
            keyCode === Keys.left;
        if (shouldFocus) {
            var /** @type {?} */ isCellSelection = this.selectionType === SelectionType.cell;
            if (!model.cellElement || !isCellSelection) {
                this.focusRow(model.rowElement, keyCode);
            }
            else if (isCellSelection) {
                this.focusCell(model.cellElement, model.rowElement, keyCode, model.cellIndex);
            }
        }
    };
    /**
     * @param {?} rowElement
     * @param {?} keyCode
     * @return {?}
     */
    DataTableSelectionComponent.prototype.focusRow = function (rowElement, keyCode) {
        var /** @type {?} */ nextRowElement = this.getPrevNextRow(rowElement, keyCode);
        if (nextRowElement)
            nextRowElement.focus();
    };
    /**
     * @param {?} rowElement
     * @param {?} keyCode
     * @return {?}
     */
    DataTableSelectionComponent.prototype.getPrevNextRow = function (rowElement, keyCode) {
        var /** @type {?} */ parentElement = rowElement.parentElement;
        if (parentElement) {
            var /** @type {?} */ focusElement = void 0;
            if (keyCode === Keys.up) {
                focusElement = parentElement.previousElementSibling;
            }
            else if (keyCode === Keys.down) {
                focusElement = parentElement.nextElementSibling;
            }
            if (focusElement && focusElement.children.length) {
                return focusElement.children[0];
            }
        }
    };
    /**
     * @param {?} cellElement
     * @param {?} rowElement
     * @param {?} keyCode
     * @param {?} cellIndex
     * @return {?}
     */
    DataTableSelectionComponent.prototype.focusCell = function (cellElement, rowElement, keyCode, cellIndex) {
        var /** @type {?} */ nextCellElement;
        if (keyCode === Keys.left) {
            nextCellElement = cellElement.previousElementSibling;
        }
        else if (keyCode === Keys.right) {
            nextCellElement = cellElement.nextElementSibling;
        }
        else if (keyCode === Keys.up || keyCode === Keys.down) {
            var /** @type {?} */ nextRowElement = this.getPrevNextRow(rowElement, keyCode);
            if (nextRowElement) {
                var /** @type {?} */ children = nextRowElement.getElementsByClassName('datatable-body-cell');
                if (children.length)
                    nextCellElement = children[cellIndex];
            }
        }
        if (nextCellElement)
            nextCellElement.focus();
    };
    /**
     * @param {?} row
     * @return {?}
     */
    DataTableSelectionComponent.prototype.getRowSelected = function (row) {
        return this.getRowSelectedIdx(row, this.selected) > -1;
    };
    /**
     * @param {?} row
     * @param {?} selected
     * @return {?}
     */
    DataTableSelectionComponent.prototype.getRowSelectedIdx = function (row, selected) {
        var _this = this;
        if (!selected || !selected.length)
            return -1;
        var /** @type {?} */ rowId = this.rowIdentity(row);
        return selected.findIndex(function (r) {
            var /** @type {?} */ id = _this.rowIdentity(r);
            return id === rowId;
        });
    };
    return DataTableSelectionComponent;
}());
DataTableSelectionComponent.decorators = [
    { type: Component, args: [{
                selector: 'datatable-selection',
                template: "\n    <ng-content></ng-content>\n  "
            },] },
];
/**
 * @nocollapse
 */
DataTableSelectionComponent.ctorParameters = function () { return []; };
DataTableSelectionComponent.propDecorators = {
    'rows': [{ type: Input },],
    'selected': [{ type: Input },],
    'selectEnabled': [{ type: Input },],
    'selectionType': [{ type: Input },],
    'rowIdentity': [{ type: Input },],
    'selectCheck': [{ type: Input },],
    'activate': [{ type: Output },],
    'select': [{ type: Output },],
};
var DataTableColumnHeaderDirective = /** @class */ (function () {
    /**
     * @param {?} template
     */
    function DataTableColumnHeaderDirective(template) {
        this.template = template;
    }
    return DataTableColumnHeaderDirective;
}());
DataTableColumnHeaderDirective.decorators = [
    { type: Directive, args: [{ selector: '[ngx-datatable-header-template]' },] },
];
/**
 * @nocollapse
 */
DataTableColumnHeaderDirective.ctorParameters = function () { return [
    { type: TemplateRef, },
]; };
var DataTableColumnCellDirective = /** @class */ (function () {
    /**
     * @param {?} template
     */
    function DataTableColumnCellDirective(template) {
        this.template = template;
    }
    return DataTableColumnCellDirective;
}());
DataTableColumnCellDirective.decorators = [
    { type: Directive, args: [{ selector: '[ngx-datatable-cell-template]' },] },
];
/**
 * @nocollapse
 */
DataTableColumnCellDirective.ctorParameters = function () { return [
    { type: TemplateRef, },
]; };
var DataTableColumnDirective = /** @class */ (function () {
    function DataTableColumnDirective() {
    }
    return DataTableColumnDirective;
}());
DataTableColumnDirective.decorators = [
    { type: Directive, args: [{ selector: 'ngx-datatable-column' },] },
];
/**
 * @nocollapse
 */
DataTableColumnDirective.ctorParameters = function () { return []; };
DataTableColumnDirective.propDecorators = {
    'name': [{ type: Input },],
    'prop': [{ type: Input },],
    'frozenLeft': [{ type: Input },],
    'frozenRight': [{ type: Input },],
    'flexGrow': [{ type: Input },],
    'resizeable': [{ type: Input },],
    'comparator': [{ type: Input },],
    'pipe': [{ type: Input },],
    'sortable': [{ type: Input },],
    'draggable': [{ type: Input },],
    'canAutoResize': [{ type: Input },],
    'minWidth': [{ type: Input },],
    'width': [{ type: Input },],
    'maxWidth': [{ type: Input },],
    'checkboxable': [{ type: Input },],
    'headerCheckboxable': [{ type: Input },],
    'headerClass': [{ type: Input },],
    'cellClass': [{ type: Input },],
    'cellTemplate': [{ type: Input }, { type: ContentChild, args: [DataTableColumnCellDirective, { read: TemplateRef },] },],
    'headerTemplate': [{ type: Input }, { type: ContentChild, args: [DataTableColumnHeaderDirective, { read: TemplateRef },] },],
};
var DatatableRowDetailTemplateDirective = /** @class */ (function () {
    /**
     * @param {?} template
     */
    function DatatableRowDetailTemplateDirective(template) {
        this.template = template;
    }
    return DatatableRowDetailTemplateDirective;
}());
DatatableRowDetailTemplateDirective.decorators = [
    { type: Directive, args: [{
                selector: '[ngx-datatable-row-detail-template]'
            },] },
];
/**
 * @nocollapse
 */
DatatableRowDetailTemplateDirective.ctorParameters = function () { return [
    { type: TemplateRef, },
]; };
var DatatableRowDetailDirective = /** @class */ (function () {
    function DatatableRowDetailDirective() {
        /**
         * The detail row height is required especially
         * when virtual scroll is enabled.
         *
         * \@memberOf DatatableComponent
         */
        this.rowHeight = 0;
        /**
         * Row detail row visbility was toggled.
         *
         * \@memberOf DatatableComponent
         */
        this.toggle = new EventEmitter();
    }
    /**
     * Toggle the expansion of the row
     *
     * @param {?} row
     * @return {?}
     */
    DatatableRowDetailDirective.prototype.toggleExpandRow = function (row) {
        this.toggle.emit({
            type: 'row',
            value: row
        });
    };
    /**
     * API method to expand all the rows.
     *
     * \@memberOf DatatableComponent
     * @return {?}
     */
    DatatableRowDetailDirective.prototype.expandAllRows = function () {
        this.toggle.emit({
            type: 'all',
            value: true
        });
    };
    /**
     * API method to collapse all the rows.
     *
     * \@memberOf DatatableComponent
     * @return {?}
     */
    DatatableRowDetailDirective.prototype.collapseAllRows = function () {
        this.toggle.emit({
            type: 'all',
            value: false
        });
    };
    return DatatableRowDetailDirective;
}());
DatatableRowDetailDirective.decorators = [
    { type: Directive, args: [{ selector: 'ngx-datatable-row-detail' },] },
];
/**
 * @nocollapse
 */
DatatableRowDetailDirective.ctorParameters = function () { return []; };
DatatableRowDetailDirective.propDecorators = {
    'rowHeight': [{ type: Input },],
    'template': [{ type: Input }, { type: ContentChild, args: [DatatableRowDetailTemplateDirective, { read: TemplateRef },] },],
    'toggle': [{ type: Output },],
};
var DataTableFooterComponent = /** @class */ (function () {
    function DataTableFooterComponent() {
        this.selectedCount = 0;
        this.page = new EventEmitter();
    }
    Object.defineProperty(DataTableFooterComponent.prototype, "isVisible", {
        /**
         * @return {?}
         */
        get: function () {
            return (this.rowCount / this.pageSize) > 1;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DataTableFooterComponent.prototype, "curPage", {
        /**
         * @return {?}
         */
        get: function () {
            return this.offset + 1;
        },
        enumerable: true,
        configurable: true
    });
    return DataTableFooterComponent;
}());
DataTableFooterComponent.decorators = [
    { type: Component, args: [{
                selector: 'datatable-footer',
                template: "\n    <div\n      class=\"datatable-footer-inner\"\n      [ngClass]=\"{'selected-count': selectedMessage}\"\n      [style.height.px]=\"footerHeight\">\n      <ng-template\n        *ngIf=\"footerTemplate\"\n        [ngTemplateOutlet]=\"footerTemplate.template\"\n        [ngOutletContext]=\"{ \n          rowCount: rowCount, \n          pageSize: pageSize, \n          selectedCount: selectedCount,\n          curPage: curPage,\n          offset: offset\n        }\">\n      </ng-template>\n      <div class=\"page-count\" *ngIf=\"!footerTemplate\">\n        <span *ngIf=\"selectedMessage\">\n          {{selectedCount.toLocaleString()}} {{selectedMessage}} / \n        </span>\n        {{rowCount.toLocaleString()}} {{totalMessage}}\n      </div>\n      <datatable-pager *ngIf=\"!footerTemplate\"\n        [pagerLeftArrowIcon]=\"pagerLeftArrowIcon\"\n        [pagerRightArrowIcon]=\"pagerRightArrowIcon\"\n        [pagerPreviousIcon]=\"pagerPreviousIcon\"\n        [pagerNextIcon]=\"pagerNextIcon\"\n        [page]=\"curPage\"\n        [size]=\"pageSize\"\n        [count]=\"rowCount\"\n        [hidden]=\"!isVisible\"\n        (change)=\"page.emit($event)\">\n      </datatable-pager>\n    </div>\n  ",
                host: {
                    class: 'datatable-footer'
                },
                changeDetection: ChangeDetectionStrategy.OnPush
            },] },
];
/**
 * @nocollapse
 */
DataTableFooterComponent.ctorParameters = function () { return []; };
DataTableFooterComponent.propDecorators = {
    'footerHeight': [{ type: Input },],
    'rowCount': [{ type: Input },],
    'pageSize': [{ type: Input },],
    'offset': [{ type: Input },],
    'pagerLeftArrowIcon': [{ type: Input },],
    'pagerRightArrowIcon': [{ type: Input },],
    'pagerPreviousIcon': [{ type: Input },],
    'pagerNextIcon': [{ type: Input },],
    'totalMessage': [{ type: Input },],
    'footerTemplate': [{ type: Input },],
    'selectedCount': [{ type: Input },],
    'selectedMessage': [{ type: Input },],
    'page': [{ type: Output },],
};
var DataTablePagerComponent = /** @class */ (function () {
    function DataTablePagerComponent() {
        this.change = new EventEmitter();
        this._count = 0;
        this._page = 1;
        this._size = 0;
    }
    Object.defineProperty(DataTablePagerComponent.prototype, "size", {
        /**
         * @return {?}
         */
        get: function () {
            return this._size;
        },
        /**
         * @param {?} val
         * @return {?}
         */
        set: function (val) {
            this._size = val;
            this.pages = this.calcPages();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DataTablePagerComponent.prototype, "count", {
        /**
         * @return {?}
         */
        get: function () {
            return this._count;
        },
        /**
         * @param {?} val
         * @return {?}
         */
        set: function (val) {
            this._count = val;
            this.pages = this.calcPages();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DataTablePagerComponent.prototype, "page", {
        /**
         * @return {?}
         */
        get: function () {
            return this._page;
        },
        /**
         * @param {?} val
         * @return {?}
         */
        set: function (val) {
            this._page = val;
            this.pages = this.calcPages();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DataTablePagerComponent.prototype, "totalPages", {
        /**
         * @return {?}
         */
        get: function () {
            var /** @type {?} */ count = this.size < 1 ? 1 : Math.ceil(this.count / this.size);
            return Math.max(count || 0, 1);
        },
        enumerable: true,
        configurable: true
    });
    /**
     * @return {?}
     */
    DataTablePagerComponent.prototype.canPrevious = function () {
        return this.page > 1;
    };
    /**
     * @return {?}
     */
    DataTablePagerComponent.prototype.canNext = function () {
        return this.page < this.totalPages;
    };
    /**
     * @return {?}
     */
    DataTablePagerComponent.prototype.prevPage = function () {
        this.selectPage(this.page - 1);
    };
    /**
     * @return {?}
     */
    DataTablePagerComponent.prototype.nextPage = function () {
        this.selectPage(this.page + 1);
    };
    /**
     * @param {?} page
     * @return {?}
     */
    DataTablePagerComponent.prototype.selectPage = function (page) {
        if (page > 0 && page <= this.totalPages && page !== this.page) {
            this.page = page;
            this.change.emit({
                page: page
            });
        }
    };
    /**
     * @param {?=} page
     * @return {?}
     */
    DataTablePagerComponent.prototype.calcPages = function (page) {
        var /** @type {?} */ pages = [];
        var /** @type {?} */ startPage = 1;
        var /** @type {?} */ endPage = this.totalPages;
        var /** @type {?} */ maxSize = 5;
        var /** @type {?} */ isMaxSized = maxSize < this.totalPages;
        page = page || this.page;
        if (isMaxSized) {
            startPage = ((Math.ceil(page / maxSize) - 1) * maxSize) + 1;
            endPage = Math.min(startPage + maxSize - 1, this.totalPages);
        }
        for (var /** @type {?} */ num = startPage; num <= endPage; num++) {
            pages.push({
                number: num,
                text: /** @type {?} */ ((num))
            });
        }
        return pages;
    };
    return DataTablePagerComponent;
}());
DataTablePagerComponent.decorators = [
    { type: Component, args: [{
                selector: 'datatable-pager',
                template: "\n    <ul class=\"pager\">\n      <li [class.disabled]=\"!canPrevious()\">\n        <a\n          href=\"javascript:void(0)\"\n          (click)=\"selectPage(1)\">\n          <i class=\"{{pagerPreviousIcon}}\"></i>\n        </a>\n      </li>\n      <li [class.disabled]=\"!canPrevious()\">\n        <a\n          href=\"javascript:void(0)\"\n          (click)=\"prevPage()\">\n          <i class=\"{{pagerLeftArrowIcon}}\"></i>\n        </a>\n      </li>\n      <li\n        class=\"pages\"\n        *ngFor=\"let pg of pages\"\n        [class.active]=\"pg.number === page\">\n        <a\n          href=\"javascript:void(0)\"\n          (click)=\"selectPage(pg.number)\">\n          {{pg.text}}\n        </a>\n      </li>\n      <li [class.disabled]=\"!canNext()\">\n        <a\n          href=\"javascript:void(0)\"\n          (click)=\"nextPage()\">\n          <i class=\"{{pagerRightArrowIcon}}\"></i>\n        </a>\n      </li>\n      <li [class.disabled]=\"!canNext()\">\n        <a\n          href=\"javascript:void(0)\"\n          (click)=\"selectPage(totalPages)\">\n          <i class=\"{{pagerNextIcon}}\"></i>\n        </a>\n      </li>\n    </ul>\n  ",
                host: {
                    class: 'datatable-pager'
                },
                changeDetection: ChangeDetectionStrategy.OnPush
            },] },
];
/**
 * @nocollapse
 */
DataTablePagerComponent.ctorParameters = function () { return []; };
DataTablePagerComponent.propDecorators = {
    'pagerLeftArrowIcon': [{ type: Input },],
    'pagerRightArrowIcon': [{ type: Input },],
    'pagerPreviousIcon': [{ type: Input },],
    'pagerNextIcon': [{ type: Input },],
    'size': [{ type: Input },],
    'count': [{ type: Input },],
    'page': [{ type: Input },],
    'change': [{ type: Output },],
};
var DataTableFooterTemplateDirective = /** @class */ (function () {
    /**
     * @param {?} template
     */
    function DataTableFooterTemplateDirective(template) {
        this.template = template;
    }
    return DataTableFooterTemplateDirective;
}());
DataTableFooterTemplateDirective.decorators = [
    { type: Directive, args: [{ selector: '[ngx-datatable-footer-template]' },] },
];
/**
 * @nocollapse
 */
DataTableFooterTemplateDirective.ctorParameters = function () { return [
    { type: TemplateRef, },
]; };
var DatatableFooterDirective = /** @class */ (function () {
    function DatatableFooterDirective() {
    }
    return DatatableFooterDirective;
}());
DatatableFooterDirective.decorators = [
    { type: Directive, args: [{ selector: 'ngx-datatable-footer' },] },
];
/**
 * @nocollapse
 */
DatatableFooterDirective.ctorParameters = function () { return []; };
DatatableFooterDirective.propDecorators = {
    'footerHeight': [{ type: Input },],
    'totalMessage': [{ type: Input },],
    'selectedMessage': [{ type: Input },],
    'pagerLeftArrowIcon': [{ type: Input },],
    'pagerRightArrowIcon': [{ type: Input },],
    'pagerPreviousIcon': [{ type: Input },],
    'pagerNextIcon': [{ type: Input },],
    'template': [{ type: Input }, { type: ContentChild, args: [DataTableFooterTemplateDirective, { read: TemplateRef },] },],
};
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function")
        r = Reflect.decorate(decorators, target, key, desc);
    else
        for (var i = decorators.length - 1; i >= 0; i--)
            if (d = decorators[i])
                r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var __metadata = (this && this.__metadata) || function (k, v) {
    if (typeof Reflect === "object" && typeof Reflect.metadata === "function")
        return Reflect.metadata(k, v);
};
var DatatableComponent = /** @class */ (function () {
    /**
     * @param {?} scrollbarHelper
     * @param {?} element
     * @param {?} differs
     */
    function DatatableComponent(scrollbarHelper, element, differs) {
        this.scrollbarHelper = scrollbarHelper;
        /**
         * List of row objects that should be
         * represented as selected in the grid.
         * Default value: `[]`
         *
         * \@memberOf DatatableComponent
         */
        this.selected = [];
        /**
         * Enable vertical scrollbars
         *
         * \@memberOf DatatableComponent
         */
        this.scrollbarV = false;
        /**
         * Enable horz scrollbars
         *
         * \@memberOf DatatableComponent
         */
        this.scrollbarH = false;
        /**
         * The row height; which is necessary
         * to calculate the height for the lazy rendering.
         *
         * \@memberOf DatatableComponent
         */
        this.rowHeight = 30;
        /**
         * Type of column width distribution formula.
         * Example: flex, force, standard
         *
         * \@memberOf DatatableComponent
         */
        this.columnMode = ColumnMode.standard;
        /**
         * The minimum header height in pixels.
         * Pass a falsey for no header
         *
         * \@memberOf DatatableComponent
         */
        this.headerHeight = 30;
        /**
         * The minimum footer height in pixels.
         * Pass falsey for no footer
         *
         * \@memberOf DatatableComponent
         */
        this.footerHeight = 0;
        /**
         * If the table should use external paging
         * otherwise its assumed that all data is preloaded.
         *
         * \@memberOf DatatableComponent
         */
        this.externalPaging = false;
        /**
         * If the table should use external sorting or
         * the built-in basic sorting.
         *
         * \@memberOf DatatableComponent
         */
        this.externalSorting = false;
        /**
         * The page size to be shown.
         * Default value: `undefined`
         *
         * \@memberOf DatatableComponent
         */
        this.limit = undefined;
        /**
         * The current offset ( page - 1 ) shown.
         * Default value: `0`
         *
         * \@memberOf DatatableComponent
         */
        this.offset = 0;
        /**
         * Show the linear loading bar.
         * Default value: `false`
         *
         * \@memberOf DatatableComponent
         */
        this.loadingIndicator = false;
        /**
         * Enable/Disable ability to re-order columns
         * by dragging them.
         *
         * \@memberOf DatatableComponent
         */
        this.reorderable = true;
        /**
         * The type of sorting
         *
         * \@memberOf DatatableComponent
         */
        this.sortType = SortType.single;
        /**
         * Array of sorted columns by property and type.
         * Default value: `[]`
         *
         * \@memberOf DatatableComponent
         */
        this.sorts = [];
        /**
         * Css class overrides
         *
         * \@memberOf DatatableComponent
         */
        this.cssClasses = {
            sortAscending: 'icon-up',
            sortDescending: 'icon-down',
            pagerLeftArrow: 'icon-left',
            pagerRightArrow: 'icon-right',
            pagerPrevious: 'icon-prev',
            pagerNext: 'icon-skip'
        };
        /**
         * Message overrides for localization
         *
         * emptyMessage     [default] = 'No data to display'
         * totalMessage     [default] = 'total'
         * selectedMessage  [default] = 'selected'
         *
         * \@memberOf DatatableComponent
         */
        this.messages = {
            // Message to show when array is presented
            // but contains no values
            emptyMessage: 'No data to display',
            // Footer total message
            totalMessage: 'total',
            // Footer selected message
            selectedMessage: 'selected'
        };
        /**
         * This will be used when displaying or selecting rows.
         * when tracking/comparing them, we'll use the value of this fn,
         *
         * (`fn(x) === fn(y)` instead of `x === y`)
         *
         * \@memberOf DatatableComponent
         */
        this.rowIdentity = (function (x) { return x; });
        /**
         * Body was scrolled typically in a `scrollbarV:true` scenario.
         *
         * \@memberOf DatatableComponent
         */
        this.scroll = new EventEmitter();
        /**
         * A cell or row was focused via keyboard or mouse click.
         *
         * \@memberOf DatatableComponent
         */
        this.activate = new EventEmitter();
        /**
         * A cell or row was selected.
         *
         * \@memberOf DatatableComponent
         */
        this.select = new EventEmitter();
        /**
         * Column sort was invoked.
         *
         * \@memberOf DatatableComponent
         */
        this.sort = new EventEmitter();
        /**
         * The table was paged either triggered by the pager or the body scroll.
         *
         * \@memberOf DatatableComponent
         */
        this.page = new EventEmitter();
        /**
         * Columns were re-ordered.
         *
         * \@memberOf DatatableComponent
         */
        this.reorder = new EventEmitter();
        /**
         * Column was resized.
         *
         * \@memberOf DatatableComponent
         */
        this.resize = new EventEmitter();
        /**
         * The context menu was invoked on a row.
         *
         * \@memberOf DatatableComponent
         */
        this.rowContextmenu = new EventEmitter(false);
        this.rowCount = 0;
        this.offsetX = 0;
        this._count = 0;
        // get ref to elm for measuring
        this.element = element.nativeElement;
        this.rowDiffer = differs.find({}).create(null);
    }
    Object.defineProperty(DatatableComponent.prototype, "rows", {
        /**
         * Gets the rows.
         *
         * \@readonly
         * \@memberOf DatatableComponent
         * @return {?}
         */
        get: function () {
            return this._rows;
        },
        /**
         * Rows that are displayed in the table.
         *
         * \@memberOf DatatableComponent
         * @param {?} val
         * @return {?}
         */
        set: function (val) {
            // auto sort on new updates
            if (!this.externalSorting) {
                val = sortRows(val, this.columns, this.sorts);
            }
            this._rows = val;
            // recalculate sizes/etc
            this.recalculate();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DatatableComponent.prototype, "columns", {
        /**
         * Get the columns.
         *
         * \@readonly
         * \@memberOf DatatableComponent
         * @return {?}
         */
        get: function () {
            return this._columns;
        },
        /**
         * Columns to be displayed.
         *
         * \@memberOf DatatableComponent
         * @param {?} val
         * @return {?}
         */
        set: function (val) {
            if (val) {
                setColumnDefaults(val);
                this.recalculateColumns(val);
            }
            this._columns = val;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DatatableComponent.prototype, "count", {
        /**
         * Gets the count.
         *
         * \@readonly
         * \@memberOf DatatableComponent
         * @return {?}
         */
        get: function () {
            return this._count;
        },
        /**
         * The total count of all rows.
         * Default value: `0`
         *
         * \@memberOf DatatableComponent
         * @param {?} val
         * @return {?}
         */
        set: function (val) {
            this._count = val;
            // recalculate sizes/etc
            this.recalculate();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DatatableComponent.prototype, "isFixedHeader", {
        /**
         * CSS class applied if the header height if fixed height.
         *
         * \@readonly
         * \@memberOf DatatableComponent
         * @return {?}
         */
        get: function () {
            var /** @type {?} */ headerHeight = this.headerHeight;
            return (typeof headerHeight === 'string') ?
                ((headerHeight)) !== 'auto' : true;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DatatableComponent.prototype, "isFixedRow", {
        /**
         * CSS class applied to the root element if
         * the row heights are fixed heights.
         *
         * \@readonly
         * \@memberOf DatatableComponent
         * @return {?}
         */
        get: function () {
            var /** @type {?} */ rowHeight = this.rowHeight;
            return (typeof rowHeight === 'string') ?
                ((rowHeight)) !== 'auto' : true;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DatatableComponent.prototype, "isVertScroll", {
        /**
         * CSS class applied to root element if
         * vertical scrolling is enabled.
         *
         * \@readonly
         * \@memberOf DatatableComponent
         * @return {?}
         */
        get: function () {
            return this.scrollbarV;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DatatableComponent.prototype, "isHorScroll", {
        /**
         * CSS class applied to the root element
         * if the horziontal scrolling is enabled.
         *
         * \@readonly
         * \@memberOf DatatableComponent
         * @return {?}
         */
        get: function () {
            return this.scrollbarH;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DatatableComponent.prototype, "isSelectable", {
        /**
         * CSS class applied to root element is selectable.
         *
         * \@readonly
         * \@memberOf DatatableComponent
         * @return {?}
         */
        get: function () {
            return this.selectionType !== undefined;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DatatableComponent.prototype, "isCheckboxSelection", {
        /**
         * CSS class applied to root is checkbox selection.
         *
         * \@readonly
         * \@memberOf DatatableComponent
         * @return {?}
         */
        get: function () {
            return this.selectionType === SelectionType.checkbox;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DatatableComponent.prototype, "isCellSelection", {
        /**
         * CSS class applied to root if cell selection.
         *
         * \@readonly
         * \@memberOf DatatableComponent
         * @return {?}
         */
        get: function () {
            return this.selectionType === SelectionType.cell;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DatatableComponent.prototype, "isSingleSelection", {
        /**
         * CSS class applied to root if single select.
         *
         * \@readonly
         * \@memberOf DatatableComponent
         * @return {?}
         */
        get: function () {
            return this.selectionType === SelectionType.single;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DatatableComponent.prototype, "isMultiSelection", {
        /**
         * CSS class added to root element if mulit select
         *
         * \@readonly
         * \@memberOf DatatableComponent
         * @return {?}
         */
        get: function () {
            return this.selectionType === SelectionType.multi;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DatatableComponent.prototype, "isMultiClickSelection", {
        /**
         * CSS class added to root element if mulit click select
         *
         * \@readonly
         * \@memberOf DatatableComponent
         * @return {?}
         */
        get: function () {
            return this.selectionType === SelectionType.multiClick;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DatatableComponent.prototype, "columnTemplates", {
        /**
         * Returns the column templates.
         *
         * \@readonly
         * \@memberOf DatatableComponent
         * @return {?}
         */
        get: function () {
            return this._columnTemplates;
        },
        /**
         * Column templates gathered from `ContentChildren`
         * if described in your markup.
         *
         * \@memberOf DatatableComponent
         * @param {?} val
         * @return {?}
         */
        set: function (val) {
            this._columnTemplates = val;
            if (val) {
                // only set this if results were brought back
                var /** @type {?} */ arr = val.toArray();
                if (arr.length) {
                    // translate them to normal objects
                    this.columns = translateTemplates(arr);
                }
            }
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DatatableComponent.prototype, "allRowsSelected", {
        /**
         * Returns if all rows are selected.
         *
         * \@readonly
         * \@memberOf DatatableComponent
         * @return {?}
         */
        get: function () {
            return this.selected &&
                this.rows &&
                this.selected.length === this.rows.length;
        },
        enumerable: true,
        configurable: true
    });
    /**
     * Lifecycle hook that is called after data-bound
     * properties of a directive are initialized.
     *
     * \@memberOf DatatableComponent
     * @return {?}
     */
    DatatableComponent.prototype.ngOnInit = function () {
        // need to call this immediatly to size
        // if the table is hidden the visibility
        // listener will invoke this itself upon show
        this.recalculate();
    };
    /**
     * Lifecycle hook that is called after a component's
     * view has been fully initialized.
     *
     * \@memberOf DatatableComponent
     * @return {?}
     */
    DatatableComponent.prototype.ngAfterViewInit = function () {
        var _this = this;
        if (!this.externalSorting) {
            var /** @type {?} */ val = sortRows(this._rows, this.columns, this.sorts);
            this._rows = val;
        }
        // this has to be done to prevent the change detection
        // tree from freaking out because we are readjusting
        setTimeout(function () { return _this.recalculate(); });
    };
    /**
     * Lifecycle hook that is called when Angular dirty checks a directive.
     *
     * \@memberOf DatatableComponent
     * @return {?}
     */
    DatatableComponent.prototype.ngDoCheck = function () {
        if (this.rowDiffer.diff(this.rows)) {
            this.recalculatePages();
        }
    };
    /**
     * Recalc's the sizes of the grid.
     *
     * Updated automatically on changes to:
     *
     *  - Columns
     *  - Rows
     *  - Paging related
     *
     * Also can be manually invoked or upon window resize.
     *
     * \@memberOf DatatableComponent
     * @return {?}
     */
    DatatableComponent.prototype.recalculate = function () {
        this.recalculateDims();
        this.recalculateColumns();
    };
    /**
     * Window resize handler to update sizes.
     *
     * \@memberOf DatatableComponent
     * @return {?}
     */
    DatatableComponent.prototype.onWindowResize = function () {
        this.recalculate();
    };
    /**
     * Recalulcates the column widths based on column width
     * distribution mode and scrollbar offsets.
     *
     *
     * \@memberOf DatatableComponent
     * @param {?=} columns
     * @param {?=} forceIdx
     * @param {?=} allowBleed
     * @return {?}
     */
    DatatableComponent.prototype.recalculateColumns = function (columns, forceIdx, allowBleed) {
        if (columns === void 0) { columns = this.columns; }
        if (forceIdx === void 0) { forceIdx = -1; }
        if (allowBleed === void 0) { allowBleed = this.scrollbarH; }
        if (!columns)
            return;
        var /** @type {?} */ width = this.innerWidth;
        if (this.scrollbarV) {
            width = width - this.scrollbarHelper.width;
        }
        if (this.columnMode === ColumnMode.force) {
            forceFillColumnWidths(columns, width, forceIdx, allowBleed);
        }
        else if (this.columnMode === ColumnMode.flex) {
            adjustColumnWidths(columns, width);
        }
        return columns;
    };
    /**
     * Recalculates the dimensions of the table size.
     * Internally calls the page size and row count calcs too.
     *
     * \@memberOf DatatableComponent
     * @return {?}
     */
    DatatableComponent.prototype.recalculateDims = function () {
        var /** @type {?} */ dims = this.element.getBoundingClientRect();
        this.innerWidth = Math.floor(dims.width);
        if (this.scrollbarV) {
            var /** @type {?} */ height = dims.height;
            if (this.headerHeight)
                height = height - this.headerHeight;
            if (this.footerHeight)
                height = height - this.footerHeight;
            this.bodyHeight = height;
        }
        this.recalculatePages();
    };
    /**
     * Recalculates the pages after a update.
     *
     *
     * \@memberOf DatatableComponent
     * @return {?}
     */
    DatatableComponent.prototype.recalculatePages = function () {
        this.pageSize = this.calcPageSize();
        this.rowCount = this.calcRowCount();
    };
    /**
     * Body triggered a page event.
     *
     *
     * \@memberOf DatatableComponent
     * @param {?} __0
     * @return {?}
     */
    DatatableComponent.prototype.onBodyPage = function (_a) {
        var offset = _a.offset;
        this.offset = offset;
        this.page.emit({
            count: this.count,
            pageSize: this.pageSize,
            limit: this.limit,
            offset: this.offset
        });
    };
    /**
     * The body triggered a scroll event.
     *
     *
     * \@memberOf DatatableComponent
     * @param {?} event
     * @return {?}
     */
    DatatableComponent.prototype.onBodyScroll = function (event) {
        this.offsetX = event.offsetX;
        this.scroll.emit(event);
    };
    /**
     * The footer triggered a page event.
     *
     *
     * \@memberOf DatatableComponent
     * @param {?} event
     * @return {?}
     */
    DatatableComponent.prototype.onFooterPage = function (event) {
        this.offset = event.page - 1;
        this.bodyComponent.updateOffsetY(this.offset);
        this.page.emit({
            count: this.count,
            pageSize: this.pageSize,
            limit: this.limit,
            offset: this.offset
        });
    };
    /**
     * Recalculates the sizes of the page
     *
     *
     * \@memberOf DatatableComponent
     * @param {?=} val
     * @return {?}
     */
    DatatableComponent.prototype.calcPageSize = function (val) {
        if (val === void 0) { val = this.rows; }
        // Keep the page size constant even if the row has been expanded.
        // This is because an expanded row is still considered to be a child of
        // the original row.  Hence calculation would use rowHeight only.
        if (this.scrollbarV) {
            var /** @type {?} */ size = Math.ceil(this.bodyHeight / this.rowHeight);
            return Math.max(size, 0);
        }
        // if limit is passed, we are paging
        if (this.limit !== undefined)
            return this.limit;
        // otherwise use row length
        if (val)
            return val.length;
        // other empty :(
        return 0;
    };
    /**
     * Calculates the row count.
     *
     *
     * \@memberOf DatatableComponent
     * @param {?=} val
     * @return {?}
     */
    DatatableComponent.prototype.calcRowCount = function (val) {
        if (val === void 0) { val = this.rows; }
        if (!this.externalPaging) {
            if (!val)
                return 0;
            return val.length;
        }
        return this.count;
    };
    /**
     * The header triggered a column resize event.
     *
     *
     * \@memberOf DatatableComponent
     * @param {?} __0
     * @return {?}
     */
    DatatableComponent.prototype.onColumnResize = function (_a) {
        var column = _a.column, newValue = _a.newValue;
        /* Safari/iOS 10.2 workaround */
        if (column === undefined) {
            return;
        }
        var /** @type {?} */ idx;
        var /** @type {?} */ cols = this.columns.map(function (c, i) {
            c = Object.assign({}, c);
            if (c.$$id === column.$$id) {
                idx = i;
                c.width = newValue;
                // set this so we can force the column
                // width distribution to be to this value
                c.$$oldWidth = newValue;
            }
            return c;
        });
        this.recalculateColumns(cols, idx);
        this._columns = cols;
        this.resize.emit({
            column: column,
            newValue: newValue
        });
    };
    /**
     * The header triggered a column re-order event.
     *
     *
     * \@memberOf DatatableComponent
     * @param {?} __0
     * @return {?}
     */
    DatatableComponent.prototype.onColumnReorder = function (_a) {
        var column = _a.column, newValue = _a.newValue, prevValue = _a.prevValue;
        var /** @type {?} */ cols = this.columns.map(function (c) {
            return Object.assign({}, c);
        });
        var /** @type {?} */ prevCol = cols[newValue];
        cols[newValue] = column;
        cols[prevValue] = prevCol;
        this.columns = cols;
        this.reorder.emit({
            column: column,
            newValue: newValue,
            prevValue: prevValue
        });
    };
    /**
     * The header triggered a column sort event.
     *
     *
     * \@memberOf DatatableComponent
     * @param {?} event
     * @return {?}
     */
    DatatableComponent.prototype.onColumnSort = function (event) {
        var sorts = event.sorts;
        // this could be optimized better since it will resort
        // the rows again on the 'push' detection...
        if (this.externalSorting === false) {
            // don't use normal setter so we don't resort
            this._rows = sortRows(this.rows, this.columns, sorts);
        }
        this.sorts = sorts;
        // Always go to first page when sorting to see the newly sorted data
        this.offset = 0;
        this.bodyComponent.updateOffsetY(this.offset);
        this.sort.emit(event);
    };
    /**
     * Toggle all row selection
     *
     *
     * \@memberOf DatatableComponent
     * @param {?} event
     * @return {?}
     */
    DatatableComponent.prototype.onHeaderSelect = function (event) {
        // before we splice, chk if we currently have all selected
        var /** @type {?} */ allSelected = this.selected.length === this.rows.length;
        // remove all existing either way
        this.selected = [];
        // do the opposite here
        if (!allSelected) {
            (_a = this.selected).push.apply(_a, this.rows);
        }
        this.select.emit({
            selected: this.selected
        });
        var _a;
    };
    /**
     * A row was selected from body
     *
     *
     * \@memberOf DatatableComponent
     * @param {?} event
     * @return {?}
     */
    DatatableComponent.prototype.onBodySelect = function (event) {
        this.select.emit(event);
    };
    return DatatableComponent;
}());
DatatableComponent.decorators = [
    { type: Component, args: [{
                selector: 'ngx-datatable',
                template: "\n    <div\n      visibility-observer\n      (visible)=\"recalculate()\">\n      <datatable-header\n        *ngIf=\"headerHeight\"\n        [sorts]=\"sorts\"\n        [sortType]=\"sortType\"\n        [scrollbarH]=\"scrollbarH\"\n        [innerWidth]=\"innerWidth\"\n        [offsetX]=\"offsetX\"\n        [columns]=\"columns\"\n        [headerHeight]=\"headerHeight\"\n        [reorderable]=\"reorderable\"\n        [sortAscendingIcon]=\"cssClasses.sortAscending\"\n        [sortDescendingIcon]=\"cssClasses.sortDescending\"\n        [allRowsSelected]=\"allRowsSelected\"\n        [selectionType]=\"selectionType\"\n        (sort)=\"onColumnSort($event)\"\n        (resize)=\"onColumnResize($event)\"\n        (reorder)=\"onColumnReorder($event)\"\n        (select)=\"onHeaderSelect($event)\">\n      </datatable-header>\n      <datatable-body\n        [rows]=\"rows\"\n        [scrollbarV]=\"scrollbarV\"\n        [scrollbarH]=\"scrollbarH\"\n        [loadingIndicator]=\"loadingIndicator\"\n        [rowHeight]=\"rowHeight\"\n        [rowCount]=\"rowCount\"\n        [offset]=\"offset\"\n        [trackByProp]=\"trackByProp\"\n        [columns]=\"columns\"\n        [pageSize]=\"pageSize\"\n        [offsetX]=\"offsetX\"\n        [rowDetail]=\"rowDetail\"\n        [selected]=\"selected\"\n        [innerWidth]=\"innerWidth\"\n        [bodyHeight]=\"bodyHeight\"\n        [selectionType]=\"selectionType\"\n        [emptyMessage]=\"messages.emptyMessage\"\n        [rowIdentity]=\"rowIdentity\"\n        [rowClass]=\"rowClass\"\n        [selectCheck]=\"selectCheck\"\n        (page)=\"onBodyPage($event)\"\n        (activate)=\"activate.emit($event)\"\n        (rowContextmenu)=\"rowContextmenu.emit($event)\"\n        (select)=\"onBodySelect($event)\"\n        (scroll)=\"onBodyScroll($event)\">\n      </datatable-body>\n      <datatable-footer\n        *ngIf=\"footerHeight\"\n        [rowCount]=\"rowCount\"\n        [pageSize]=\"pageSize\"\n        [offset]=\"offset\"\n        [footerHeight]=\"footerHeight\"\n        [footerTemplate]=\"footer\"\n        [totalMessage]=\"messages.totalMessage\"\n        [pagerLeftArrowIcon]=\"cssClasses.pagerLeftArrow\"\n        [pagerRightArrowIcon]=\"cssClasses.pagerRightArrow\"\n        [pagerPreviousIcon]=\"cssClasses.pagerPrevious\"\n        [selectedCount]=\"selected.length\"\n        [selectedMessage]=\"!!selectionType && messages.selectedMessage\"\n        [pagerNextIcon]=\"cssClasses.pagerNext\"\n        (page)=\"onFooterPage($event)\">\n      </datatable-footer>\n    </div>\n  ",
                encapsulation: ViewEncapsulation.None,
                styles: ["\n    .ngx-datatable {\n      display: block;\n      overflow: hidden;\n      -webkit-box-pack: center;\n          -ms-flex-pack: center;\n              justify-content: center;\n      position: relative;\n      -webkit-transform: translate3d(0, 0, 0);\n      /**\n       * Vertical Scrolling Adjustments\n       */\n      /**\n       * Horizontal Scrolling Adjustments\n       */\n      /**\n       * Fixed Header Height Adjustments\n       */\n      /**\n       * Fixed row height adjustments\n       */\n      /**\n       * Shared Styles\n       */\n      /**\n       * Header Styles\n       */\n      /**\n       * Body Styles\n       */\n      /**\n       * Footer Styles\n       */ }\n      .ngx-datatable [hidden] {\n        display: none !important; }\n      .ngx-datatable *, .ngx-datatable *:before, .ngx-datatable *:after {\n        -webkit-box-sizing: border-box;\n        box-sizing: border-box; }\n      .ngx-datatable.scroll-vertical .datatable-body {\n        overflow-y: auto; }\n        .ngx-datatable.scroll-vertical .datatable-body .datatable-row-wrapper {\n          position: absolute; }\n      .ngx-datatable.scroll-horz .datatable-body {\n        overflow-x: auto;\n        -webkit-overflow-scrolling: touch; }\n      .ngx-datatable.fixed-header .datatable-header .datatable-header-inner {\n        white-space: nowrap; }\n        .ngx-datatable.fixed-header .datatable-header .datatable-header-inner .datatable-header-cell {\n          white-space: nowrap;\n          overflow: hidden;\n          text-overflow: ellipsis; }\n      .ngx-datatable.fixed-row .datatable-scroll {\n        white-space: nowrap; }\n        .ngx-datatable.fixed-row .datatable-scroll .datatable-body-row {\n          white-space: nowrap; }\n          .ngx-datatable.fixed-row .datatable-scroll .datatable-body-row .datatable-body-cell {\n            overflow: hidden;\n            white-space: nowrap;\n            text-overflow: ellipsis; }\n      .ngx-datatable .datatable-body-row,\n      .ngx-datatable .datatable-header-inner {\n        display: -webkit-box;\n        display: -ms-flexbox;\n        display: flex;\n        -webkit-box-orient: horizontal;\n        -webkit-box-direction: normal;\n            -ms-flex-direction: row;\n                flex-direction: row;\n        -ms-flex-flow: row;\n        -o-flex-flow: row;\n        flex-flow: row; }\n      .ngx-datatable .datatable-body-cell,\n      .ngx-datatable .datatable-header-cell {\n        vertical-align: top;\n        display: inline-block;\n        line-height: 1.625;\n        -webkit-user-select: none;\n        -moz-user-select: none;\n        -ms-user-select: none;\n        -o-user-select: none;\n        user-select: none; }\n        .ngx-datatable .datatable-body-cell:focus,\n        .ngx-datatable .datatable-header-cell:focus {\n          outline: none; }\n      .ngx-datatable .datatable-row-left,\n      .ngx-datatable .datatable-row-right {\n        z-index: 9; }\n      .ngx-datatable .datatable-row-left,\n      .ngx-datatable .datatable-row-center,\n      .ngx-datatable .datatable-row-right {\n        position: relative; }\n      .ngx-datatable .datatable-header {\n        display: block;\n        overflow: hidden; }\n        .ngx-datatable .datatable-header .datatable-header-inner {\n          -webkit-box-align: stretch;\n              -ms-flex-align: stretch;\n                  align-items: stretch;\n          -webkit-align-items: stretch; }\n        .ngx-datatable .datatable-header .datatable-header-cell {\n          position: relative;\n          display: inline-block; }\n          .ngx-datatable .datatable-header .datatable-header-cell .datatable-header-cell-wrapper {\n            cursor: pointer; }\n          .ngx-datatable .datatable-header .datatable-header-cell.longpress .datatable-header-cell-wrapper {\n            cursor: move; }\n          .ngx-datatable .datatable-header .datatable-header-cell .sort-btn {\n            line-height: 100%;\n            vertical-align: middle;\n            display: inline-block;\n            cursor: pointer; }\n          .ngx-datatable .datatable-header .datatable-header-cell .resize-handle {\n            display: inline-block;\n            position: absolute;\n            right: 0;\n            top: 0;\n            bottom: 0;\n            width: 5px;\n            padding: 0 4px;\n            visibility: hidden;\n            cursor: ew-resize; }\n          .ngx-datatable .datatable-header .datatable-header-cell.resizeable:hover .resize-handle {\n            visibility: visible; }\n      .ngx-datatable .datatable-body {\n        position: relative;\n        z-index: 10;\n        display: block; }\n        .ngx-datatable .datatable-body .datatable-scroll {\n          display: inline-block; }\n        .ngx-datatable .datatable-body .datatable-row-detail {\n          overflow-y: hidden; }\n        .ngx-datatable .datatable-body .datatable-row-wrapper {\n          display: -webkit-box;\n          display: -ms-flexbox;\n          display: flex;\n          -webkit-box-orient: vertical;\n          -webkit-box-direction: normal;\n          -ms-flex-direction: column;\n          flex-direction: column; }\n        .ngx-datatable .datatable-body .datatable-body-row {\n          outline: none; }\n          .ngx-datatable .datatable-body .datatable-body-row > div {\n            display: -webkit-box;\n            display: -ms-flexbox;\n            display: flex; }\n      .ngx-datatable .datatable-footer {\n        display: block;\n        width: 100%; }\n        .ngx-datatable .datatable-footer .datatable-footer-inner {\n          display: -webkit-box;\n          display: -ms-flexbox;\n          display: flex;\n          -webkit-box-align: center;\n              -ms-flex-align: center;\n                  align-items: center;\n          width: 100%; }\n        .ngx-datatable .datatable-footer .selected-count .page-count {\n          -webkit-box-flex: 1;\n              -ms-flex: 1 1 40%;\n                  flex: 1 1 40%; }\n        .ngx-datatable .datatable-footer .selected-count .datatable-pager {\n          -webkit-box-flex: 1;\n              -ms-flex: 1 1 60%;\n                  flex: 1 1 60%; }\n        .ngx-datatable .datatable-footer .page-count {\n          -webkit-box-flex: 1;\n              -ms-flex: 1 1 20%;\n                  flex: 1 1 20%; }\n        .ngx-datatable .datatable-footer .datatable-pager {\n          -webkit-box-flex: 1;\n              -ms-flex: 1 1 80%;\n                  flex: 1 1 80%;\n          text-align: right; }\n          .ngx-datatable .datatable-footer .datatable-pager .pager,\n          .ngx-datatable .datatable-footer .datatable-pager .pager li {\n            padding: 0;\n            margin: 0;\n            display: inline-block;\n            list-style: none; }\n          .ngx-datatable .datatable-footer .datatable-pager .pager li, .ngx-datatable .datatable-footer .datatable-pager .pager li a {\n            outline: none; }\n          .ngx-datatable .datatable-footer .datatable-pager .pager li a {\n            cursor: pointer;\n            display: inline-block; }\n          .ngx-datatable .datatable-footer .datatable-pager .pager li.disabled a {\n            cursor: not-allowed; }\n  "],
                host: {
                    class: 'ngx-datatable'
                }
            },] },
];
/**
 * @nocollapse
 */
DatatableComponent.ctorParameters = function () { return [
    { type: ScrollbarHelper, },
    { type: ElementRef, },
    { type: KeyValueDiffers, },
]; };
DatatableComponent.propDecorators = {
    'rows': [{ type: Input },],
    'columns': [{ type: Input },],
    'selected': [{ type: Input },],
    'scrollbarV': [{ type: Input },],
    'scrollbarH': [{ type: Input },],
    'rowHeight': [{ type: Input },],
    'columnMode': [{ type: Input },],
    'headerHeight': [{ type: Input },],
    'footerHeight': [{ type: Input },],
    'externalPaging': [{ type: Input },],
    'externalSorting': [{ type: Input },],
    'limit': [{ type: Input },],
    'count': [{ type: Input },],
    'offset': [{ type: Input },],
    'loadingIndicator': [{ type: Input },],
    'selectionType': [{ type: Input },],
    'reorderable': [{ type: Input },],
    'sortType': [{ type: Input },],
    'sorts': [{ type: Input },],
    'cssClasses': [{ type: Input },],
    'messages': [{ type: Input },],
    'rowIdentity': [{ type: Input },],
    'rowClass': [{ type: Input },],
    'selectCheck': [{ type: Input },],
    'trackByProp': [{ type: Input },],
    'scroll': [{ type: Output },],
    'activate': [{ type: Output },],
    'select': [{ type: Output },],
    'sort': [{ type: Output },],
    'page': [{ type: Output },],
    'reorder': [{ type: Output },],
    'resize': [{ type: Output },],
    'rowContextmenu': [{ type: Output },],
    'isFixedHeader': [{ type: HostBinding, args: ['class.fixed-header',] },],
    'isFixedRow': [{ type: HostBinding, args: ['class.fixed-row',] },],
    'isVertScroll': [{ type: HostBinding, args: ['class.scroll-vertical',] },],
    'isHorScroll': [{ type: HostBinding, args: ['class.scroll-horz',] },],
    'isSelectable': [{ type: HostBinding, args: ['class.selectable',] },],
    'isCheckboxSelection': [{ type: HostBinding, args: ['class.checkbox-selection',] },],
    'isCellSelection': [{ type: HostBinding, args: ['class.cell-selection',] },],
    'isSingleSelection': [{ type: HostBinding, args: ['class.single-selection',] },],
    'isMultiSelection': [{ type: HostBinding, args: ['class.multi-selection',] },],
    'isMultiClickSelection': [{ type: HostBinding, args: ['class.multi-click-selection',] },],
    'columnTemplates': [{ type: ContentChildren, args: [DataTableColumnDirective,] },],
    'rowDetail': [{ type: ContentChild, args: [DatatableRowDetailDirective,] },],
    'footer': [{ type: ContentChild, args: [DatatableFooterDirective,] },],
    'bodyComponent': [{ type: ViewChild, args: [DataTableBodyComponent,] },],
    'onWindowResize': [{ type: HostListener, args: ['window:resize',] },],
};
__decorate([
    throttleable(5),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", []),
    __metadata("design:returntype", void 0)
], DatatableComponent.prototype, "onWindowResize", null);
var DataTableHeaderComponent = /** @class */ (function () {
    function DataTableHeaderComponent() {
        this.sort = new EventEmitter();
        this.reorder = new EventEmitter();
        this.resize = new EventEmitter();
        this.select = new EventEmitter();
    }
    Object.defineProperty(DataTableHeaderComponent.prototype, "headerHeight", {
        /**
         * @return {?}
         */
        get: function () {
            return this._headerHeight;
        },
        /**
         * @param {?} val
         * @return {?}
         */
        set: function (val) {
            if (val !== 'auto') {
                this._headerHeight = val + "px";
            }
            else {
                this._headerHeight = val;
            }
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DataTableHeaderComponent.prototype, "columns", {
        /**
         * @return {?}
         */
        get: function () {
            return this._columns;
        },
        /**
         * @param {?} val
         * @return {?}
         */
        set: function (val) {
            this._columns = val;
            var /** @type {?} */ colsByPin = columnsByPin(val);
            this.columnsByPin = columnsByPinArr(val);
            this.columnGroupWidths = columnGroupWidths(colsByPin, val);
        },
        enumerable: true,
        configurable: true
    });
    /**
     * @param {?} __0
     * @return {?}
     */
    DataTableHeaderComponent.prototype.onLongPressStart = function (_a) {
        var event = _a.event, model = _a.model;
        model.dragging = true;
        this.dragEventTarget = event;
    };
    /**
     * @param {?} __0
     * @return {?}
     */
    DataTableHeaderComponent.prototype.onLongPressEnd = function (_a) {
        var event = _a.event, model = _a.model;
        this.dragEventTarget = event;
        // delay resetting so sort can be 
        // prevented if we were dragging
        setTimeout(function () {
            model.dragging = false;
        }, 5);
    };
    Object.defineProperty(DataTableHeaderComponent.prototype, "headerWidth", {
        /**
         * @return {?}
         */
        get: function () {
            if (this.scrollbarH) {
                return this.innerWidth + 'px';
            }
            return '100%';
        },
        enumerable: true,
        configurable: true
    });
    /**
     * @param {?} index
     * @param {?} colGroup
     * @return {?}
     */
    DataTableHeaderComponent.prototype.trackByGroups = function (index, colGroup) {
        return colGroup.type;
    };
    /**
     * @param {?} index
     * @param {?} column
     * @return {?}
     */
    DataTableHeaderComponent.prototype.columnTrackingFn = function (index, column) {
        return column.$$id;
    };
    /**
     * @param {?} width
     * @param {?} column
     * @return {?}
     */
    DataTableHeaderComponent.prototype.onColumnResized = function (width, column) {
        if (width <= column.minWidth) {
            width = column.minWidth;
        }
        else if (width >= column.maxWidth) {
            width = column.maxWidth;
        }
        this.resize.emit({
            column: column,
            prevValue: column.width,
            newValue: width
        });
    };
    /**
     * @param {?} __0
     * @return {?}
     */
    DataTableHeaderComponent.prototype.onColumnReordered = function (_a) {
        var prevIndex = _a.prevIndex, newIndex = _a.newIndex, model = _a.model;
        this.reorder.emit({
            column: model,
            prevValue: prevIndex,
            newValue: newIndex
        });
    };
    /**
     * @param {?} __0
     * @return {?}
     */
    DataTableHeaderComponent.prototype.onSort = function (_a) {
        var column = _a.column, prevValue = _a.prevValue, newValue = _a.newValue;
        // if we are dragging don't sort!
        if (column.dragging)
            return;
        var /** @type {?} */ sorts = this.calcNewSorts(column, prevValue, newValue);
        this.sort.emit({
            sorts: sorts,
            column: column,
            prevValue: prevValue,
            newValue: newValue
        });
    };
    /**
     * @param {?} column
     * @param {?} prevValue
     * @param {?} newValue
     * @return {?}
     */
    DataTableHeaderComponent.prototype.calcNewSorts = function (column, prevValue, newValue) {
        var /** @type {?} */ idx = 0;
        var /** @type {?} */ sorts = this.sorts.map(function (s, i) {
            s = Object.assign({}, s);
            if (s.prop === column.prop)
                idx = i;
            return s;
        });
        if (newValue === undefined) {
            sorts.splice(idx, 1);
        }
        else if (prevValue) {
            sorts[idx].dir = newValue;
        }
        else {
            if (this.sortType === SortType.single) {
                sorts.splice(0, this.sorts.length);
            }
            sorts.push({ dir: newValue, prop: column.prop });
        }
        return sorts;
    };
    /**
     * @param {?} group
     * @return {?}
     */
    DataTableHeaderComponent.prototype.stylesByGroup = function (group) {
        var /** @type {?} */ widths = this.columnGroupWidths;
        var /** @type {?} */ offsetX = this.offsetX;
        var /** @type {?} */ styles = {
            width: widths[group] + "px"
        };
        if (group === 'center') {
            translateXY(styles, offsetX * -1, 0);
        }
        else if (group === 'right') {
            var /** @type {?} */ totalDiff = widths.total - this.innerWidth;
            var /** @type {?} */ offset = totalDiff * -1;
            translateXY(styles, offset, 0);
        }
        return styles;
    };
    return DataTableHeaderComponent;
}());
DataTableHeaderComponent.decorators = [
    { type: Component, args: [{
                selector: 'datatable-header',
                template: "\n    <div\n      orderable\n      (reorder)=\"onColumnReordered($event)\"\n      [style.width.px]=\"columnGroupWidths.total\"\n      class=\"datatable-header-inner\">\n      <div\n        *ngFor=\"let colGroup of columnsByPin; trackBy: trackByGroups\"\n        [class]=\"'datatable-row-' + colGroup.type\"\n        [ngStyle]=\"stylesByGroup(colGroup.type)\">\n        <datatable-header-cell\n          *ngFor=\"let column of colGroup.columns; trackBy: columnTrackingFn\"\n          resizeable\n          [resizeEnabled]=\"column.resizeable\"\n          (resize)=\"onColumnResized($event, column)\"\n          long-press\n          [pressModel]=\"column\"\n          [pressEnabled]=\"reorderable && column.draggable\"\n          (longPressStart)=\"onLongPressStart($event)\"\n          (longPressEnd)=\"onLongPressEnd($event)\"\n          draggable\n          [dragX]=\"reorderable && column.draggable && column.dragging\"\n          [dragY]=\"false\"\n          [dragModel]=\"column\"\n          [dragEventTarget]=\"dragEventTarget\"\n          [headerHeight]=\"headerHeight\"\n          [column]=\"column\"\n          [sortType]=\"sortType\"\n          [sorts]=\"sorts\"\n          [selectionType]=\"selectionType\"\n          [sortAscendingIcon]=\"sortAscendingIcon\"\n          [sortDescendingIcon]=\"sortDescendingIcon\"\n          (sort)=\"onSort($event)\"\n          (select)=\"select.emit($event)\">\n        </datatable-header-cell>\n      </div>\n    </div>\n  ",
                host: {
                    class: 'datatable-header'
                }
            },] },
];
/**
 * @nocollapse
 */
DataTableHeaderComponent.ctorParameters = function () { return []; };
DataTableHeaderComponent.propDecorators = {
    'sortAscendingIcon': [{ type: Input },],
    'sortDescendingIcon': [{ type: Input },],
    'scrollbarH': [{ type: Input },],
    'innerWidth': [{ type: Input },],
    'offsetX': [{ type: Input },],
    'sorts': [{ type: Input },],
    'sortType': [{ type: Input },],
    'allRowsSelected': [{ type: Input },],
    'selectionType': [{ type: Input },],
    'reorderable': [{ type: Input },],
    'headerHeight': [{ type: HostBinding, args: ['style.height',] }, { type: Input },],
    'columns': [{ type: Input },],
    'sort': [{ type: Output },],
    'reorder': [{ type: Output },],
    'resize': [{ type: Output },],
    'select': [{ type: Output },],
    'headerWidth': [{ type: HostBinding, args: ['style.width',] },],
};
var DataTableHeaderCellComponent = /** @class */ (function () {
    function DataTableHeaderCellComponent() {
        this.sort = new EventEmitter();
        this.select = new EventEmitter();
        this.sortFn = this.onSort.bind(this);
    }
    Object.defineProperty(DataTableHeaderCellComponent.prototype, "sorts", {
        /**
         * @return {?}
         */
        get: function () {
            return this._sorts;
        },
        /**
         * @param {?} val
         * @return {?}
         */
        set: function (val) {
            this._sorts = val;
            this.sortDir = this.calcSortDir(val);
            this.sortClass = this.calcSortClass(this.sortDir);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DataTableHeaderCellComponent.prototype, "columnCssClasses", {
        /**
         * @return {?}
         */
        get: function () {
            var /** @type {?} */ cls = 'datatable-header-cell';
            if (this.column.sortable)
                cls += ' sortable';
            if (this.column.resizeable)
                cls += ' resizeable';
            if (this.column.headerClass) {
                if (typeof this.column.headerClass === 'string') {
                    cls += ' ' + this.column.headerClass;
                }
                else if (typeof this.column.headerClass === 'function') {
                    var /** @type {?} */ res = this.column.headerClass({
                        column: this.column
                    });
                    if (typeof res === 'string') {
                        cls += res;
                    }
                    else if (typeof res === 'object') {
                        var /** @type {?} */ keys = Object.keys(res);
                        for (var _i = 0, keys_3 = keys; _i < keys_3.length; _i++) {
                            var k = keys_3[_i];
                            if (res[k] === true)
                                cls += " " + k;
                        }
                    }
                }
            }
            var /** @type {?} */ sortDir = this.sortDir;
            if (sortDir) {
                cls += " sort-active sort-" + sortDir;
            }
            return cls;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DataTableHeaderCellComponent.prototype, "name", {
        /**
         * @return {?}
         */
        get: function () {
            // guaranteed to have a value by setColumnDefaults() in column-helper.ts
            return this.column.headerTemplate === undefined ? this.column.name : undefined;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DataTableHeaderCellComponent.prototype, "minWidth", {
        /**
         * @return {?}
         */
        get: function () {
            return this.column.minWidth;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DataTableHeaderCellComponent.prototype, "maxWidth", {
        /**
         * @return {?}
         */
        get: function () {
            return this.column.maxWidth;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DataTableHeaderCellComponent.prototype, "width", {
        /**
         * @return {?}
         */
        get: function () {
            return this.column.width;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DataTableHeaderCellComponent.prototype, "isCheckboxable", {
        /**
         * @return {?}
         */
        get: function () {
            return this.column.checkboxable &&
                this.column.headerCheckboxable &&
                this.selectionType === SelectionType.checkbox;
        },
        enumerable: true,
        configurable: true
    });
    /**
     * @param {?} sorts
     * @return {?}
     */
    DataTableHeaderCellComponent.prototype.calcSortDir = function (sorts) {
        var _this = this;
        if (sorts && this.column) {
            var /** @type {?} */ sort = sorts.find(function (s) {
                return s.prop === _this.column.prop;
            });
            if (sort)
                return sort.dir;
        }
    };
    /**
     * @return {?}
     */
    DataTableHeaderCellComponent.prototype.onSort = function () {
        if (!this.column.sortable)
            return;
        var /** @type {?} */ newValue = nextSortDir(this.sortType, this.sortDir);
        this.sort.emit({
            column: this.column,
            prevValue: this.sortDir,
            newValue: newValue
        });
    };
    /**
     * @param {?} sortDir
     * @return {?}
     */
    DataTableHeaderCellComponent.prototype.calcSortClass = function (sortDir) {
        if (sortDir === SortDirection.asc) {
            return "sort-btn sort-asc " + this.sortAscendingIcon;
        }
        else if (sortDir === SortDirection.desc) {
            return "sort-btn sort-desc " + this.sortDescendingIcon;
        }
        else {
            return "sort-btn";
        }
    };
    return DataTableHeaderCellComponent;
}());
DataTableHeaderCellComponent.decorators = [
    { type: Component, args: [{
                selector: 'datatable-header-cell',
                template: "\n    <div>\n      <label\n        *ngIf=\"isCheckboxable\" \n        class=\"datatable-checkbox\">\n        <input \n          type=\"checkbox\"\n          [attr.checked]=\"allRowsSelected\"\n          (change)=\"select.emit(!allRowsSelected)\" \n        />\n      </label>\n      <span \n        *ngIf=\"!column.headerTemplate\"\n        class=\"datatable-header-cell-wrapper\">\n        <span\n          class=\"datatable-header-cell-label draggable\"\n          (click)=\"onSort()\"\n          [innerHTML]=\"name\">\n        </span>\n      </span>\n      <ng-template\n        *ngIf=\"column.headerTemplate\"\n        [ngTemplateOutlet]=\"column.headerTemplate\"\n        [ngOutletContext]=\"{ \n          column: column, \n          sortDir: sortDir,\n          sortFn: sortFn\n        }\">\n      </ng-template>\n      <span\n        (click)=\"onSort()\"\n        [class]=\"sortClass\">\n      </span>\n    </div>\n  "
            },] },
];
/**
 * @nocollapse
 */
DataTableHeaderCellComponent.ctorParameters = function () { return []; };
DataTableHeaderCellComponent.propDecorators = {
    'sortType': [{ type: Input },],
    'column': [{ type: Input },],
    'sortAscendingIcon': [{ type: Input },],
    'sortDescendingIcon': [{ type: Input },],
    'allRowsSelected': [{ type: Input },],
    'selectionType': [{ type: Input },],
    'headerHeight': [{ type: HostBinding, args: ['style.height.px',] }, { type: Input },],
    'sorts': [{ type: Input },],
    'sort': [{ type: Output },],
    'select': [{ type: Output },],
    'columnCssClasses': [{ type: HostBinding, args: ['class',] },],
    'name': [{ type: HostBinding, args: ['attr.title',] },],
    'minWidth': [{ type: HostBinding, args: ['style.minWidth.px',] },],
    'maxWidth': [{ type: HostBinding, args: ['style.maxWidth.px',] },],
    'width': [{ type: HostBinding, args: ['style.width.px',] },],
};
/**
 * Draggable Directive for Angular2
 *
 * Inspiration:
 *   https://github.com/AngularClass/angular2-examples/blob/master/rx-draggable/directives/draggable.ts
 *   http://stackoverflow.com/questions/35662530/how-to-implement-drag-and-drop-in-angular2
 *
 */
var DraggableDirective = /** @class */ (function () {
    /**
     * @param {?} element
     */
    function DraggableDirective(element) {
        this.dragX = true;
        this.dragY = true;
        this.dragStart = new EventEmitter();
        this.dragging = new EventEmitter();
        this.dragEnd = new EventEmitter();
        this.isDragging = false;
        this.element = element.nativeElement;
    }
    /**
     * @param {?} changes
     * @return {?}
     */
    DraggableDirective.prototype.ngOnChanges = function (changes) {
        if (changes['dragEventTarget'] && changes['dragEventTarget'].currentValue && this.dragModel.dragging) {
            this.onMousedown(changes['dragEventTarget'].currentValue);
        }
    };
    /**
     * @return {?}
     */
    DraggableDirective.prototype.ngOnDestroy = function () {
        this._destroySubscription();
    };
    /**
     * @param {?} event
     * @return {?}
     */
    DraggableDirective.prototype.onMouseup = function (event) {
        if (!this.isDragging)
            return;
        this.isDragging = false;
        this.element.classList.remove('dragging');
        if (this.subscription) {
            this._destroySubscription();
            this.dragEnd.emit({
                event: event,
                element: this.element,
                model: this.dragModel
            });
        }
    };
    /**
     * @param {?} event
     * @return {?}
     */
    DraggableDirective.prototype.onMousedown = function (event) {
        var _this = this;
        // we only want to drag the inner header text
        var /** @type {?} */ isDragElm = ((event.target)).classList.contains('draggable');
        if (isDragElm && (this.dragX || this.dragY)) {
            event.preventDefault();
            this.isDragging = true;
            var /** @type {?} */ mouseDownPos_1 = { x: event.clientX, y: event.clientY };
            var /** @type {?} */ mouseup = Observable$1.fromEvent(document, 'mouseup');
            this.subscription = mouseup
                .subscribe(function (ev) { return _this.onMouseup(ev); });
            var /** @type {?} */ mouseMoveSub = Observable$1.fromEvent(document, 'mousemove')
                .takeUntil(mouseup)
                .subscribe(function (ev) { return _this.move(ev, mouseDownPos_1); });
            this.subscription.add(mouseMoveSub);
            this.dragStart.emit({
                event: event,
                element: this.element,
                model: this.dragModel
            });
        }
    };
    /**
     * @param {?} event
     * @param {?} mouseDownPos
     * @return {?}
     */
    DraggableDirective.prototype.move = function (event, mouseDownPos) {
        if (!this.isDragging)
            return;
        var /** @type {?} */ x = event.clientX - mouseDownPos.x;
        var /** @type {?} */ y = event.clientY - mouseDownPos.y;
        if (this.dragX)
            this.element.style.left = x + "px";
        if (this.dragY)
            this.element.style.top = y + "px";
        this.element.classList.add('dragging');
        this.dragging.emit({
            event: event,
            element: this.element,
            model: this.dragModel
        });
    };
    /**
     * @return {?}
     */
    DraggableDirective.prototype._destroySubscription = function () {
        if (this.subscription) {
            this.subscription.unsubscribe();
            this.subscription = undefined;
        }
    };
    return DraggableDirective;
}());
DraggableDirective.decorators = [
    { type: Directive, args: [{ selector: '[draggable]' },] },
];
/**
 * @nocollapse
 */
DraggableDirective.ctorParameters = function () { return [
    { type: ElementRef, },
]; };
DraggableDirective.propDecorators = {
    'dragEventTarget': [{ type: Input },],
    'dragModel': [{ type: Input },],
    'dragX': [{ type: Input },],
    'dragY': [{ type: Input },],
    'dragStart': [{ type: Output },],
    'dragging': [{ type: Output },],
    'dragEnd': [{ type: Output },],
};
var LongPressDirective = /** @class */ (function () {
    function LongPressDirective() {
        this.pressEnabled = true;
        this.duration = 500;
        this.longPressStart = new EventEmitter();
        this.longPressing = new EventEmitter();
        this.longPressEnd = new EventEmitter();
        this.mouseX = 0;
        this.mouseY = 0;
    }
    Object.defineProperty(LongPressDirective.prototype, "press", {
        /**
         * @return {?}
         */
        get: function () { return this.pressing; },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(LongPressDirective.prototype, "isLongPress", {
        /**
         * @return {?}
         */
        get: function () {
            return this.isLongPressing;
        },
        enumerable: true,
        configurable: true
    });
    /**
     * @param {?} event
     * @return {?}
     */
    LongPressDirective.prototype.onMouseDown = function (event) {
        var _this = this;
        // don't do right/middle clicks
        if (event.which !== 1 || !this.pressEnabled)
            return;
        // don't start drag if its on resize handle
        var /** @type {?} */ target = ((event.target));
        if (target.classList.contains('resize-handle'))
            return;
        this.mouseX = event.clientX;
        this.mouseY = event.clientY;
        this.pressing = true;
        this.isLongPressing = false;
        var /** @type {?} */ mouseup = Observable$1.fromEvent(document, 'mouseup');
        this.subscription = mouseup.subscribe(function (ev) { return _this.onMouseup(); });
        this.timeout = setTimeout(function () {
            _this.isLongPressing = true;
            _this.longPressStart.emit({
                event: event,
                model: _this.pressModel
            });
            _this.subscription.add(Observable$1.fromEvent(document, 'mousemove')
                .takeUntil(mouseup)
                .subscribe(function (mouseEvent) { return _this.onMouseMove(mouseEvent); }));
            _this.loop(event);
        }, this.duration);
        this.loop(event);
    };
    /**
     * @param {?} event
     * @return {?}
     */
    LongPressDirective.prototype.onMouseMove = function (event) {
        if (this.pressing && !this.isLongPressing) {
            var /** @type {?} */ xThres = Math.abs(event.clientX - this.mouseX) > 10;
            var /** @type {?} */ yThres = Math.abs(event.clientY - this.mouseY) > 10;
            if (xThres || yThres) {
                this.endPress();
            }
        }
    };
    /**
     * @param {?} event
     * @return {?}
     */
    LongPressDirective.prototype.loop = function (event) {
        var _this = this;
        if (this.isLongPressing) {
            this.timeout = setTimeout(function () {
                _this.longPressing.emit({
                    event: event,
                    model: _this.pressModel
                });
                _this.loop(event);
            }, 50);
        }
    };
    /**
     * @return {?}
     */
    LongPressDirective.prototype.endPress = function () {
        clearTimeout(this.timeout);
        this.isLongPressing = false;
        this.pressing = false;
        this._destroySubscription();
        this.longPressEnd.emit({
            model: this.pressModel
        });
    };
    /**
     * @return {?}
     */
    LongPressDirective.prototype.onMouseup = function () {
        this.endPress();
    };
    /**
     * @return {?}
     */
    LongPressDirective.prototype.ngOnDestroy = function () {
        this._destroySubscription();
    };
    /**
     * @return {?}
     */
    LongPressDirective.prototype._destroySubscription = function () {
        if (this.subscription) {
            this.subscription.unsubscribe();
            this.subscription = undefined;
        }
    };
    return LongPressDirective;
}());
LongPressDirective.decorators = [
    { type: Directive, args: [{ selector: '[long-press]' },] },
];
/**
 * @nocollapse
 */
LongPressDirective.ctorParameters = function () { return []; };
LongPressDirective.propDecorators = {
    'pressEnabled': [{ type: Input },],
    'pressModel': [{ type: Input },],
    'duration': [{ type: Input },],
    'longPressStart': [{ type: Output },],
    'longPressing': [{ type: Output },],
    'longPressEnd': [{ type: Output },],
    'press': [{ type: HostBinding, args: ['class.press',] },],
    'isLongPress': [{ type: HostBinding, args: ['class.longpress',] },],
    'onMouseDown': [{ type: HostListener, args: ['mousedown', ['$event'],] },],
};
var OrderableDirective = /** @class */ (function () {
    /**
     * @param {?} differs
     * @param {?} document
     */
    function OrderableDirective(differs, document) {
        this.document = document;
        this.reorder = new EventEmitter();
        this.differ = differs.find({}).create(null);
    }
    /**
     * @return {?}
     */
    OrderableDirective.prototype.ngAfterContentInit = function () {
        // HACK: Investigate Better Way
        this.updateSubscriptions();
        this.draggables.changes.subscribe(this.updateSubscriptions.bind(this));
    };
    /**
     * @return {?}
     */
    OrderableDirective.prototype.ngOnDestroy = function () {
        this.draggables.forEach(function (d) {
            d.dragStart.unsubscribe();
            d.dragEnd.unsubscribe();
        });
    };
    /**
     * @return {?}
     */
    OrderableDirective.prototype.updateSubscriptions = function () {
        var _this = this;
        var /** @type {?} */ diffs = this.differ.diff(this.createMapDiffs());
        if (diffs) {
            var /** @type {?} */ subscribe = function (_a) {
                var currentValue = _a.currentValue, previousValue = _a.previousValue;
                unsubscribe_1({ previousValue: previousValue });
                if (currentValue) {
                    currentValue.dragStart.subscribe(_this.onDragStart.bind(_this));
                    currentValue.dragEnd.subscribe(_this.onDragEnd.bind(_this));
                }
            };
            var /** @type {?} */ unsubscribe_1 = function (_a) {
                var previousValue = _a.previousValue;
                if (previousValue) {
                    previousValue.dragStart.unsubscribe();
                    previousValue.dragEnd.unsubscribe();
                }
            };
            diffs.forEachAddedItem(subscribe.bind(this));
            // diffs.forEachChangedItem(subscribe.bind(this));
            diffs.forEachRemovedItem(unsubscribe_1.bind(this));
        }
    };
    /**
     * @return {?}
     */
    OrderableDirective.prototype.onDragStart = function () {
        this.positions = {};
        var /** @type {?} */ i = 0;
        for (var _i = 0, _a = this.draggables.toArray(); _i < _a.length; _i++) {
            var dragger = _a[_i];
            var /** @type {?} */ elm = dragger.element;
            var /** @type {?} */ left = parseInt(elm.offsetLeft.toString(), 0);
            this.positions[dragger.dragModel.prop] = {
                left: left,
                right: left + parseInt(elm.offsetWidth.toString(), 0),
                index: i++,
                element: elm
            };
        }
    };
    /**
     * @param {?} __0
     * @return {?}
     */
    OrderableDirective.prototype.onDragEnd = function (_a) {
        var element = _a.element, model = _a.model, event = _a.event;
        var /** @type {?} */ prevPos = this.positions[model.prop];
        var /** @type {?} */ target = this.isTarget(model, event);
        if (target) {
            this.reorder.emit({
                prevIndex: prevPos.index,
                newIndex: target.i,
                model: model
            });
        }
        element.style.left = 'auto';
    };
    /**
     * @param {?} model
     * @param {?} event
     * @return {?}
     */
    OrderableDirective.prototype.isTarget = function (model, event) {
        var /** @type {?} */ i = 0;
        var /** @type {?} */ x = event.x || event.clientX;
        var /** @type {?} */ y = event.y || event.clientY;
        var /** @type {?} */ targets = this.document.elementsFromPoint(x, y);
        var _loop_1 = function (prop) {
            // current column position which throws event.
            var /** @type {?} */ pos = this_1.positions[prop];
            // since we drag the inner span, we need to find it in the elements at the cursor
            if (model.prop !== prop && targets.find(function (el) { return el === pos.element; })) {
                return { value: {
                        pos: pos,
                        i: i
                    } };
            }
            i++;
        };
        var this_1 = this;
        for (var /** @type {?} */ prop in this.positions) {
            var state_1 = _loop_1(/** @type {?} */ prop);
            if (typeof state_1 === "object")
                return state_1.value;
        }
    };
    /**
     * @return {?}
     */
    OrderableDirective.prototype.createMapDiffs = function () {
        return this.draggables.toArray()
            .reduce(function (acc, curr) {
            acc[curr.dragModel.$$id] = curr;
            return acc;
        }, {});
    };
    return OrderableDirective;
}());
OrderableDirective.decorators = [
    { type: Directive, args: [{ selector: '[orderable]' },] },
];
/**
 * @nocollapse
 */
OrderableDirective.ctorParameters = function () { return [
    { type: KeyValueDiffers, },
    { type: undefined, decorators: [{ type: Inject, args: [DOCUMENT,] },] },
]; };
OrderableDirective.propDecorators = {
    'reorder': [{ type: Output },],
    'draggables': [{ type: ContentChildren, args: [DraggableDirective, { descendants: true },] },],
};
var ResizeableDirective = /** @class */ (function () {
    /**
     * @param {?} element
     */
    function ResizeableDirective(element) {
        this.resizeEnabled = true;
        this.resize = new EventEmitter();
        this.resizing = false;
        this.element = element.nativeElement;
    }
    /**
     * @return {?}
     */
    ResizeableDirective.prototype.ngAfterViewInit = function () {
        if (this.resizeEnabled) {
            var /** @type {?} */ node = document.createElement('span');
            node.classList.add('resize-handle');
            this.element.appendChild(node);
        }
    };
    /**
     * @return {?}
     */
    ResizeableDirective.prototype.ngOnDestroy = function () {
        this._destroySubscription();
    };
    /**
     * @return {?}
     */
    ResizeableDirective.prototype.onMouseup = function () {
        this.resizing = false;
        if (this.subscription && !this.subscription.closed) {
            this._destroySubscription();
            this.resize.emit(this.element.clientWidth);
        }
    };
    /**
     * @param {?} event
     * @return {?}
     */
    ResizeableDirective.prototype.onMousedown = function (event) {
        var _this = this;
        var /** @type {?} */ isHandle = (((event.target))).classList.contains('resize-handle');
        var /** @type {?} */ initialWidth = this.element.clientWidth;
        var /** @type {?} */ mouseDownScreenX = event.screenX;
        if (isHandle) {
            event.stopPropagation();
            this.resizing = true;
            var /** @type {?} */ mouseup = Observable$1.fromEvent(document, 'mouseup');
            this.subscription = mouseup
                .subscribe(function (ev) { return _this.onMouseup(); });
            var /** @type {?} */ mouseMoveSub = Observable$1.fromEvent(document, 'mousemove')
                .takeUntil(mouseup)
                .subscribe(function (e) { return _this.move(e, initialWidth, mouseDownScreenX); });
            this.subscription.add(mouseMoveSub);
        }
    };
    /**
     * @param {?} event
     * @param {?} initialWidth
     * @param {?} mouseDownScreenX
     * @return {?}
     */
    ResizeableDirective.prototype.move = function (event, initialWidth, mouseDownScreenX) {
        var /** @type {?} */ movementX = event.screenX - mouseDownScreenX;
        var /** @type {?} */ newWidth = initialWidth + movementX;
        var /** @type {?} */ overMinWidth = !this.minWidth || newWidth >= this.minWidth;
        var /** @type {?} */ underMaxWidth = !this.maxWidth || newWidth <= this.maxWidth;
        if (overMinWidth && underMaxWidth) {
            this.element.style.width = newWidth + "px";
        }
    };
    /**
     * @return {?}
     */
    ResizeableDirective.prototype._destroySubscription = function () {
        if (this.subscription) {
            this.subscription.unsubscribe();
            this.subscription = undefined;
        }
    };
    return ResizeableDirective;
}());
ResizeableDirective.decorators = [
    { type: Directive, args: [{
                selector: '[resizeable]',
                host: {
                    '[class.resizeable]': 'resizeEnabled'
                }
            },] },
];
/**
 * @nocollapse
 */
ResizeableDirective.ctorParameters = function () { return [
    { type: ElementRef, },
]; };
ResizeableDirective.propDecorators = {
    'resizeEnabled': [{ type: Input },],
    'minWidth': [{ type: Input },],
    'maxWidth': [{ type: Input },],
    'resize': [{ type: Output },],
    'onMousedown': [{ type: HostListener, args: ['mousedown', ['$event'],] },],
};
/**
 * Visibility Observer Directive
 *
 * Usage:
 *
 * 		<div
 * 			visibility-observer
 * 			(visible)="onVisible($event)">
 * 		</div>
 *
 */
var VisibilityDirective = /** @class */ (function () {
    /**
     * @param {?} element
     * @param {?} zone
     */
    function VisibilityDirective(element, zone) {
        this.element = element;
        this.zone = zone;
        this.isVisible = false;
        this.visible = new EventEmitter();
    }
    /**
     * @return {?}
     */
    VisibilityDirective.prototype.ngOnInit = function () {
        this.runCheck();
    };
    /**
     * @return {?}
     */
    VisibilityDirective.prototype.ngOnDestroy = function () {
        clearTimeout(this.timeout);
    };
    /**
     * @return {?}
     */
    VisibilityDirective.prototype.onVisibilityChange = function () {
        var _this = this;
        // trigger zone recalc for columns
        this.zone.run(function () {
            _this.isVisible = true;
            _this.visible.emit(true);
        });
    };
    /**
     * @return {?}
     */
    VisibilityDirective.prototype.runCheck = function () {
        var _this = this;
        var /** @type {?} */ check = function () {
            // https://davidwalsh.name/offsetheight-visibility
            var _a = _this.element.nativeElement, offsetHeight = _a.offsetHeight, offsetWidth = _a.offsetWidth;
            if (offsetHeight && offsetWidth) {
                clearTimeout(_this.timeout);
                _this.onVisibilityChange();
            }
            else {
                clearTimeout(_this.timeout);
                _this.zone.runOutsideAngular(function () {
                    _this.timeout = setTimeout(function () { return check(); }, 50);
                });
            }
        };
        setTimeout(function () { return check(); });
    };
    return VisibilityDirective;
}());
VisibilityDirective.decorators = [
    { type: Directive, args: [{ selector: '[visibility-observer]' },] },
];
/**
 * @nocollapse
 */
VisibilityDirective.ctorParameters = function () { return [
    { type: ElementRef, },
    { type: NgZone, },
]; };
VisibilityDirective.propDecorators = {
    'isVisible': [{ type: HostBinding, args: ['class.visible',] },],
    'visible': [{ type: Output },],
};
var NgxDatatableModule = /** @class */ (function () {
    function NgxDatatableModule() {
    }
    return NgxDatatableModule;
}());
NgxDatatableModule.decorators = [
    { type: NgModule, args: [{
                imports: [
                    CommonModule
                ],
                providers: [
                    ScrollbarHelper
                ],
                declarations: [
                    DataTableFooterTemplateDirective,
                    VisibilityDirective,
                    DraggableDirective,
                    ResizeableDirective,
                    OrderableDirective,
                    LongPressDirective,
                    ScrollerComponent,
                    DatatableComponent,
                    DataTableColumnDirective,
                    DataTableHeaderComponent,
                    DataTableHeaderCellComponent,
                    DataTableBodyComponent,
                    DataTableFooterComponent,
                    DataTablePagerComponent,
                    ProgressBarComponent,
                    DataTableBodyRowComponent,
                    DataTableRowWrapperComponent,
                    DatatableRowDetailDirective,
                    DatatableRowDetailTemplateDirective,
                    DataTableBodyCellComponent,
                    DataTableSelectionComponent,
                    DataTableColumnHeaderDirective,
                    DataTableColumnCellDirective,
                    DatatableFooterDirective
                ],
                exports: [
                    DatatableComponent,
                    DatatableRowDetailDirective,
                    DatatableRowDetailTemplateDirective,
                    DataTableColumnDirective,
                    DataTableColumnHeaderDirective,
                    DataTableColumnCellDirective,
                    DataTableFooterTemplateDirective,
                    DatatableFooterDirective
                ]
            },] },
];
/**
 * @nocollapse
 */
NgxDatatableModule.ctorParameters = function () { return []; };
/**
 * Generated bundle index. Do not edit.
 */
export { NgxDatatableModule, ColumnMode, SortType, SortDirection, SelectionType, ClickType, DatatableComponent, DataTableHeaderComponent, DataTableHeaderCellComponent, DataTableBodyComponent, DataTableBodyCellComponent, DataTableBodyRowComponent, ProgressBarComponent, ScrollerComponent, DataTableRowWrapperComponent, DataTableSelectionComponent, DataTableFooterComponent, DataTablePagerComponent, DatatableFooterDirective, DataTableFooterTemplateDirective, DataTableColumnDirective, DataTableColumnHeaderDirective, DataTableColumnCellDirective, DatatableRowDetailDirective, DatatableRowDetailTemplateDirective, DataTableBodyCellComponent as ɵu, DataTableBodyComponent as ɵm, DataTableBodyRowComponent as ɵq, DataTableColumnCellDirective as ɵx, DataTableColumnDirective as ɵj, DataTableColumnHeaderDirective as ɵw, DataTableFooterComponent as ɵn, DataTableFooterTemplateDirective as ɵb, DataTableHeaderCellComponent as ɵl, DataTableHeaderComponent as ɵk, DataTablePagerComponent as ɵo, DataTableRowWrapperComponent as ɵr, DataTableSelectionComponent as ɵv, DatatableComponent as ɵi, DatatableFooterDirective as ɵy, DatatableRowDetailDirective as ɵs, DatatableRowDetailTemplateDirective as ɵt, ProgressBarComponent as ɵp, ScrollerComponent as ɵh, DraggableDirective as ɵd, LongPressDirective as ɵg, OrderableDirective as ɵf, ResizeableDirective as ɵe, VisibilityDirective as ɵc, ScrollbarHelper as ɵa };
//# sourceMappingURL=ngx-datatable-ng-packagr.es5.js.map
