import { ChangeDetectionStrategy, Component, ContentChild, ContentChildren, Directive, ElementRef, EventEmitter, HostBinding, HostListener, Inject, Injectable, Input, KeyValueDiffers, NgModule, NgZone, Output, Renderer, TemplateRef, ViewChild, ViewContainerRef, ViewEncapsulation } from '@angular/core';
import { CommonModule } from '@angular/common';
import 'rxjs/add/observable/fromEvent';
import { DOCUMENT } from '@angular/platform-browser';
import { Observable as Observable$1 } from 'rxjs/Observable';
import 'rxjs/add/operator/takeUntil';

/**
 * Creates a unique object id.
 * http://stackoverflow.com/questions/6248666/how-to-generate-short-uid-like-ax4j9z-in-js
 * @return {?}
 */
function id() {
    return ('0000' + (Math.random() * Math.pow(36, 4) << 0).toString(36)).slice(-4);
}

/**
 * Returns the columns by pin.
 * @param {?} cols
 * @return {?}
 */
function columnsByPin(cols) {
    const /** @type {?} */ ret = {
        left: [],
        center: [],
        right: []
    };
    if (cols) {
        for (const /** @type {?} */ col of cols) {
            if (col.frozenLeft) {
                ret.left.push(col);
            }
            else if (col.frozenRight) {
                ret.right.push(col);
            }
            else {
                ret.center.push(col);
            }
        }
    }
    return ret;
}
/**
 * Returns the widths of all group sets of a column
 * @param {?} groups
 * @param {?} all
 * @return {?}
 */
function columnGroupWidths(groups, all) {
    return {
        left: columnTotalWidth(groups.left),
        center: columnTotalWidth(groups.center),
        right: columnTotalWidth(groups.right),
        total: columnTotalWidth(all)
    };
}
/**
 * Calculates the total width of all columns and their groups
 * @param {?} columns
 * @param {?=} prop
 * @return {?}
 */
function columnTotalWidth(columns, prop) {
    let /** @type {?} */ totalWidth = 0;
    if (columns) {
        for (const /** @type {?} */ c of columns) {
            const /** @type {?} */ has = prop && c[prop];
            const /** @type {?} */ width = has ? c[prop] : c.width;
            totalWidth = totalWidth + parseFloat(width);
        }
    }
    return totalWidth;
}
/**
 * Calculates the total width of all columns and their groups
 * @param {?} columns
 * @param {?=} prop
 * @return {?}
 */
function columnsTotalWidth(columns, prop) {
    let /** @type {?} */ totalWidth = 0;
    for (const /** @type {?} */ column of columns) {
        const /** @type {?} */ has = prop && column[prop];
        totalWidth = totalWidth + (has ? column[prop] : column.width);
    }
    return totalWidth;
}
/**
 * @param {?} val
 * @return {?}
 */
function columnsByPinArr(val) {
    const /** @type {?} */ colsByPinArr = [];
    const /** @type {?} */ colsByPin = columnsByPin(val);
    colsByPinArr.push({ type: 'left', columns: colsByPin['left'] });
    colsByPinArr.push({ type: 'center', columns: colsByPin['center'] });
    colsByPinArr.push({ type: 'right', columns: colsByPin['right'] });
    return colsByPinArr;
}

// maybe rename this file to prop-getters.ts
/**
 * Always returns the empty string ''
 * @return {?}
 */
function emptyStringGetter() {
    return '';
}
/**
 * Returns the appropriate getter function for this kind of prop.
 * If prop == null, returns the emptyStringGetter.
 * @param {?} prop
 * @return {?}
 */
function getterForProp(prop) {
    if (prop == null)
        return emptyStringGetter;
    if (typeof prop === 'number') {
        return numericIndexGetter;
    }
    else {
        // deep or simple
        if (prop.indexOf('.') !== -1) {
            return deepValueGetter;
        }
        else {
            return shallowValueGetter;
        }
    }
}
/**
 * Returns the value at this numeric index.
 * @param {?} row array of values
 * @param {?} index numeric index
 * @return {?}
 */
function numericIndexGetter(row, index) {
    // mimic behavior of deepValueGetter
    if (!row || index == null)
        return row;
    const /** @type {?} */ value = row[index];
    if (value == null)
        return '';
    return value;
}
/**
 * Returns the value of a field.
 * (more efficient than deepValueGetter)
 * @param {?} obj object containing the field
 * @param {?} fieldName field name string
 * @return {?}
 */
function shallowValueGetter(obj, fieldName) {
    if (!obj || !fieldName)
        return obj;
    const /** @type {?} */ value = obj[fieldName];
    if (value == null)
        return '';
    return value;
}
/**
 * Returns a deep object given a string. zoo['animal.type']
 * @param {?} obj
 * @param {?} path
 * @return {?}
 */
function deepValueGetter(obj, path) {
    if (!obj || !path)
        return obj;
    // check if path matches a root-level field
    // { "a.b.c": 123 }
    let /** @type {?} */ current = obj[path];
    if (current !== undefined)
        return current;
    current = obj;
    const /** @type {?} */ split = path.split('.');
    if (split.length) {
        for (let /** @type {?} */ i = 0; i < split.length; i++) {
            current = current[split[i]];
            // if found undefined, return empty string
            if (current === undefined || current === null)
                return '';
        }
    }
    return current;
}

/**
 * Converts strings from something to camel case
 * http://stackoverflow.com/questions/10425287/convert-dash-separated-string-to-camelcase
 * @param {?} str
 * @return {?}
 */
function camelCase(str) {
    // Replace special characters with a space
    str = str.replace(/[^a-zA-Z0-9 ]/g, ' ');
    // put a space before an uppercase letter
    str = str.replace(/([a-z](?=[A-Z]))/g, '$1 ');
    // Lower case first character and some other stuff
    str = str.replace(/([^a-zA-Z0-9 ])|^[0-9]+/g, '').trim().toLowerCase();
    // uppercase characters preceded by a space or number
    str = str.replace(/([ 0-9]+)([a-zA-Z])/g, function (a, b, c) {
        return b.trim() + c.toUpperCase();
    });
    return str;
}
/**
 * Converts strings from camel case to words
 * http://stackoverflow.com/questions/7225407/convert-camelcasetext-to-camel-case-text
 *
 * @export
 * @param {?} str
 * @return {?} string
 */
function deCamelCase(str) {
    return str
        .replace(/([A-Z])/g, (match) => ` ${match}`)
        .replace(/^./, (match) => match.toUpperCase());
}

let Keys = {};
Keys.up = 38;
Keys.down = 40;
Keys.return = 13;
Keys.escape = 27;
Keys.left = 37;
Keys.right = 39;
Keys[Keys.up] = "up";
Keys[Keys.down] = "down";
Keys[Keys.return] = "return";
Keys[Keys.escape] = "escape";
Keys[Keys.left] = "left";
Keys[Keys.right] = "right";

/**
 * Calculates the Total Flex Grow
 * @param {?} columns
 * @return {?}
 */
function getTotalFlexGrow(columns) {
    let /** @type {?} */ totalFlexGrow = 0;
    for (const /** @type {?} */ c of columns) {
        totalFlexGrow += c.flexGrow || 0;
    }
    return totalFlexGrow;
}
/**
 * Adjusts the column widths.
 * Inspired by: https://github.com/facebook/fixed-data-table/blob/master/src/FixedDataTableWidthHelper.js
 * @param {?} allColumns
 * @param {?} expectedWidth
 * @return {?}
 */
function adjustColumnWidths(allColumns, expectedWidth) {
    const /** @type {?} */ columnsWidth = columnsTotalWidth(allColumns);
    const /** @type {?} */ totalFlexGrow = getTotalFlexGrow(allColumns);
    const /** @type {?} */ colsByGroup = columnsByPin(allColumns);
    if (columnsWidth !== expectedWidth) {
        scaleColumns(colsByGroup, expectedWidth, totalFlexGrow);
    }
}
/**
 * Resizes columns based on the flexGrow property, while respecting manually set widths
 * @param {?} colsByGroup
 * @param {?} maxWidth
 * @param {?} totalFlexGrow
 * @return {?}
 */
function scaleColumns(colsByGroup, maxWidth, totalFlexGrow) {
    // calculate total width and flexgrow points for coulumns that can be resized
    for (const /** @type {?} */ attr in colsByGroup) {
        for (const /** @type {?} */ column of colsByGroup[attr]) {
            if (!column.canAutoResize) {
                maxWidth -= column.width;
                totalFlexGrow -= column.flexGrow;
            }
            else {
                column.width = 0;
            }
        }
    }
    const /** @type {?} */ hasMinWidth = {};
    let /** @type {?} */ remainingWidth = maxWidth;
    // resize columns until no width is left to be distributed
    do {
        const /** @type {?} */ widthPerFlexPoint = remainingWidth / totalFlexGrow;
        remainingWidth = 0;
        for (const /** @type {?} */ attr in colsByGroup) {
            for (const /** @type {?} */ column of colsByGroup[attr]) {
                // if the column can be resize and it hasn't reached its minimum width yet
                if (column.canAutoResize && !hasMinWidth[column.prop]) {
                    const /** @type {?} */ newWidth = column.width + column.flexGrow * widthPerFlexPoint;
                    if (column.minWidth !== undefined && newWidth < column.minWidth) {
                        remainingWidth += newWidth - column.minWidth;
                        column.width = column.minWidth;
                        hasMinWidth[column.prop] = true;
                    }
                    else {
                        column.width = newWidth;
                    }
                }
            }
        }
    } while (remainingWidth !== 0);
}
/**
 * Forces the width of the columns to
 * distribute equally but overflowing when necessary
 *
 * Rules:
 *
 *  - If combined withs are less than the total width of the grid,
 *    proportion the widths given the min / max / normal widths to fill the width.
 *
 *  - If the combined widths, exceed the total width of the grid,
 *    use the standard widths.
 *
 *  - If a column is resized, it should always use that width
 *
 *  - The proportional widths should never fall below min size if specified.
 *
 *  - If the grid starts off small but then becomes greater than the size ( + / - )
 *    the width should use the original width; not the newly proportioned widths.
 *
 * @param {?} allColumns
 * @param {?} expectedWidth
 * @param {?} startIdx
 * @param {?} allowBleed
 * @param {?=} defaultColWidth
 * @return {?}
 */
function forceFillColumnWidths(allColumns, expectedWidth, startIdx, allowBleed, defaultColWidth = 300) {
    const /** @type {?} */ columnsToResize = allColumns
        .slice(startIdx + 1, allColumns.length)
        .filter((c) => {
        return c.canAutoResize !== false;
    });
    for (const /** @type {?} */ column of columnsToResize) {
        if (!column.$$oldWidth) {
            column.$$oldWidth = column.width;
        }
    }
    let /** @type {?} */ additionWidthPerColumn = 0;
    let /** @type {?} */ exceedsWindow = false;
    let /** @type {?} */ contentWidth = getContentWidth(allColumns, defaultColWidth);
    let /** @type {?} */ remainingWidth = expectedWidth - contentWidth;
    const /** @type {?} */ columnsProcessed = [];
    // This loop takes care of the
    do {
        additionWidthPerColumn = remainingWidth / columnsToResize.length;
        exceedsWindow = contentWidth >= expectedWidth;
        for (const /** @type {?} */ column of columnsToResize) {
            if (exceedsWindow && allowBleed) {
                column.width = column.$$oldWidth || column.width || defaultColWidth;
            }
            else {
                const /** @type {?} */ newSize = (column.width || defaultColWidth) + additionWidthPerColumn;
                if (column.minWidth && newSize < column.minWidth) {
                    column.width = column.minWidth;
                    columnsProcessed.push(column);
                }
                else if (column.maxWidth && newSize > column.maxWidth) {
                    column.width = column.maxWidth;
                    columnsProcessed.push(column);
                }
                else {
                    column.width = newSize;
                }
            }
            column.width = Math.max(0, column.width);
        }
        contentWidth = getContentWidth(allColumns);
        remainingWidth = expectedWidth - contentWidth;
        removeProcessedColumns(columnsToResize, columnsProcessed);
    } while (remainingWidth > 0 && columnsToResize.length !== 0);
}
/**
 * Remove the processed columns from the current active columns.
 *
 * @param {?} columnsToResize  Array containing the columns that need to be resized.
 * @param {?} columnsProcessed Array containing the columns that have already been processed.
 * @return {?}
 */
function removeProcessedColumns(columnsToResize, columnsProcessed) {
    for (const /** @type {?} */ column of columnsProcessed) {
        const /** @type {?} */ index = columnsToResize.indexOf(column);
        columnsToResize.splice(index, 1);
    }
}
/**
 * Gets the width of the columns
 *
 * @param {?} allColumns
 * @param {?=} defaultColWidth
 * @return {?}
 */
function getContentWidth(allColumns, defaultColWidth = 300) {
    let /** @type {?} */ contentWidth = 0;
    for (const /** @type {?} */ column of allColumns) {
        contentWidth += (column.width || defaultColWidth);
    }
    return contentWidth;
}

const cache = {};
const testStyle = document.createElement('div').style;
// Get Prefix
// http://davidwalsh.name/vendor-prefix
const prefix = (function () {
    const /** @type {?} */ styles = window.getComputedStyle(document.documentElement, '');
    const /** @type {?} */ pre = (Array.prototype.slice.call(styles).join('').match(/-(moz|webkit|ms)-/))[1];
    const /** @type {?} */ dom = ('WebKit|Moz|MS|O').match(new RegExp('(' + pre + ')', 'i'))[1];
    return {
        dom,
        lowercase: pre,
        css: `-${pre}-`,
        js: pre[0].toUpperCase() + pre.substr(1)
    };
})();
/**
 * @param {?} property
 * @return {?}
 */
function getVendorPrefixedName(property) {
    const /** @type {?} */ name = camelCase(property);
    if (!cache[name]) {
        if (testStyle[prefix.css + property] !== undefined) {
            cache[name] = prefix.css + property;
        }
        else if (testStyle[property] !== undefined) {
            cache[name] = property;
        }
    }
    return cache[name];
}

/**
 * @param {?} selected
 * @param {?} row
 * @param {?} comparefn
 * @return {?}
 */
function selectRows(selected, row, comparefn) {
    const /** @type {?} */ selectedIndex = comparefn(row, selected);
    if (selectedIndex > -1) {
        selected.splice(selectedIndex, 1);
    }
    else {
        selected.push(row);
    }
    return selected;
}
/**
 * @param {?} selected
 * @param {?} rows
 * @param {?} index
 * @param {?} prevIndex
 * @param {?} comparefn
 * @return {?}
 */
function selectRowsBetween(selected, rows, index, prevIndex, comparefn) {
    const /** @type {?} */ reverse = index < prevIndex;
    for (let /** @type {?} */ i = 0; i < rows.length; i++) {
        const /** @type {?} */ row = rows[i];
        const /** @type {?} */ greater = i >= prevIndex && i <= index;
        const /** @type {?} */ lesser = i <= prevIndex && i >= index;
        let /** @type {?} */ range = { start: 0, end: 0 };
        if (reverse) {
            range = {
                start: index,
                end: (prevIndex - index)
            };
        }
        else {
            range = {
                start: prevIndex,
                end: index + 1
            };
        }
        if ((reverse && lesser) || (!reverse && greater)) {
            // if in the positive range to be added to `selected`, and
            // not already in the selected array, add it
            if (i >= range.start && i <= range.end) {
                selected.push(row);
            }
        }
    }
    return selected;
}

// browser detection and prefixing tools
const transform = getVendorPrefixedName('transform');
const backfaceVisibility = getVendorPrefixedName('backfaceVisibility');
const hasCSSTransforms = !!getVendorPrefixedName('transform');
const hasCSS3DTransforms = !!getVendorPrefixedName('perspective');
const ua = window.navigator.userAgent;
const isSafari = (/Safari\//).test(ua) && !(/Chrome\//).test(ua);
/**
 * @param {?} styles
 * @param {?} x
 * @param {?} y
 * @return {?}
 */
function translateXY(styles, x, y) {
    if (hasCSSTransforms) {
        if (!isSafari && hasCSS3DTransforms) {
            styles[transform] = `translate3d(${x}px, ${y}px, 0)`;
            styles[backfaceVisibility] = 'hidden';
        }
        else {
            styles[camelCase(transform)] = `translate(${x}px, ${y}px)`;
        }
    }
    else {
        styles.top = `${y}px`;
        styles.left = `${x}px`;
    }
}

/**
 * Throttle a function
 *
 * @export
 * @param {?} func
 * @param {?} wait
 * @param {?=} options
 * @return {?}
 */
function throttle(func, wait, options) {
    options = options || {};
    let /** @type {?} */ context;
    let /** @type {?} */ args;
    let /** @type {?} */ result;
    let /** @type {?} */ timeout = null;
    let /** @type {?} */ previous = 0;
    /**
     * @return {?}
     */
    function later() {
        previous = options.leading === false ? 0 : +new Date();
        timeout = null;
        result = func.apply(context, args);
    }
    return function () {
        const /** @type {?} */ now = +new Date();
        if (!previous && options.leading === false) {
            previous = now;
        }
        const /** @type {?} */ remaining = wait - (now - previous);
        context = this;
        args = arguments;
        if (remaining <= 0) {
            clearTimeout(timeout);
            timeout = null;
            previous = now;
            result = func.apply(context, args);
        }
        else if (!timeout && options.trailing !== false) {
            timeout = setTimeout(later, remaining);
        }
        return result;
    };
}
/**
 * Throttle decorator
 *
 *  class MyClass {
 *    throttleable(10)
 *    myFn() { ... }
 *  }
 *
 * @export
 * @param {?} duration
 * @param {?=} options
 * @return {?}
 */
function throttleable(duration, options) {
    return function innerDecorator(target, key, descriptor) {
        return {
            configurable: true,
            enumerable: descriptor.enumerable,
            get: function getter() {
                Object.defineProperty(this, key, {
                    configurable: true,
                    enumerable: descriptor.enumerable,
                    value: throttle(descriptor.value, duration, options)
                });
                return this[key];
            }
        };
    };
}

let ColumnMode = {};
ColumnMode.standard = ('standard');
ColumnMode.flex = ('flex');
ColumnMode.force = ('force');
ColumnMode[ColumnMode.standard] = "standard";
ColumnMode[ColumnMode.flex] = "flex";
ColumnMode[ColumnMode.force] = "force";

let SortType = {};
SortType.single = ('single');
SortType.multi = ('multi');
SortType[SortType.single] = "single";
SortType[SortType.multi] = "multi";

let SortDirection = {};
SortDirection.asc = ('asc');
SortDirection.desc = ('desc');
SortDirection[SortDirection.asc] = "asc";
SortDirection[SortDirection.desc] = "desc";

let SelectionType = {};
SelectionType.single = ('single');
SelectionType.multi = ('multi');
SelectionType.multiClick = ('multiClick');
SelectionType.cell = ('cell');
SelectionType.checkbox = ('checkbox');
SelectionType[SelectionType.single] = "single";
SelectionType[SelectionType.multi] = "multi";
SelectionType[SelectionType.multiClick] = "multiClick";
SelectionType[SelectionType.cell] = "cell";
SelectionType[SelectionType.checkbox] = "checkbox";

let ClickType = {};
ClickType.single = ('single');
ClickType.double = ('double');
ClickType[ClickType.single] = "single";
ClickType[ClickType.double] = "double";

/**
 * Gets the next sort direction
 * @param {?} sortType
 * @param {?} current
 * @return {?}
 */
function nextSortDir(sortType, current) {
    if (sortType === SortType.single) {
        if (current === SortDirection.asc) {
            return SortDirection.desc;
        }
        else {
            return SortDirection.asc;
        }
    }
    else {
        if (!current) {
            return SortDirection.asc;
        }
        else if (current === SortDirection.asc) {
            return SortDirection.desc;
        }
        else if (current === SortDirection.desc) {
            return undefined;
        }
    }
}
/**
 * Adapted from fueld-ui on 6/216
 * https://github.com/FuelInteractive/fuel-ui/tree/master/src/pipes/OrderBy
 * @param {?} a
 * @param {?} b
 * @return {?}
 */
function orderByComparator(a, b) {
    if (a === null || typeof a === 'undefined')
        a = 0;
    if (b === null || typeof b === 'undefined')
        b = 0;
    if ((isNaN(parseFloat(a)) || !isFinite(a)) || (isNaN(parseFloat(b)) || !isFinite(b))) {
        // Convert to string in case of a=0 or b=0
        a = String(a);
        b = String(b);
        // Isn't a number so lowercase the string to properly compare
        if (a.toLowerCase() < b.toLowerCase())
            return -1;
        if (a.toLowerCase() > b.toLowerCase())
            return 1;
    }
    else {
        // Parse strings as numbers to compare properly
        if (parseFloat(a) < parseFloat(b))
            return -1;
        if (parseFloat(a) > parseFloat(b))
            return 1;
    }
    // equal each other
    return 0;
}
/**
 * Sorts the rows
 *
 * @export
 * @param {?} rows
 * @param {?} columns
 * @param {?} dirs
 * @return {?}
 */
function sortRows(rows, columns, dirs) {
    if (!rows || !dirs || !dirs.length || !columns)
        return rows;
    const /** @type {?} */ temp = [...rows];
    const /** @type {?} */ cols = columns.reduce((obj, col) => {
        if (col.comparator && typeof col.comparator === 'function') {
            obj[col.prop] = col.comparator;
        }
        return obj;
    }, {});
    // cache valueGetter and compareFn so that they
    // do not need to be looked-up in the sort function body
    const /** @type {?} */ cachedDirs = dirs.map(dir => {
        const /** @type {?} */ prop = dir.prop;
        return {
            prop,
            dir: dir.dir,
            valueGetter: getterForProp(prop),
            compareFn: cols[prop] || orderByComparator
        };
    });
    return temp.sort(function (a, b) {
        for (const /** @type {?} */ cachedDir of cachedDirs) {
            const { prop, valueGetter } = cachedDir;
            const /** @type {?} */ propA = valueGetter(a, prop);
            const /** @type {?} */ propB = valueGetter(b, prop);
            const /** @type {?} */ comparison = cachedDir.dir !== SortDirection.desc ?
                cachedDir.compareFn(propA, propB) :
                -cachedDir.compareFn(propA, propB);
            // Don't return 0 yet in case of needing to sort by next property
            if (comparison !== 0)
                return comparison;
        }
        // equal each other
        return 0;
    });
}

/**
 * This object contains the cache of the various row heights that are present inside
 * the data table.   Its based on Fenwick tree data structure that helps with
 * querying sums that have time complexity of log n.
 *
 * Fenwick Tree Credits: http://petr-mitrichev.blogspot.com/2013/05/fenwick-tree-range-updates.html
 * https://github.com/mikolalysenko/fenwick-tree
 *
 */
class RowHeightCache {
    constructor() {
        /**
         * Tree Array stores the cumulative information of the row heights to perform efficient
         * range queries and updates.  Currently the tree is initialized to the base row
         * height instead of the detail row height.
         */
        this.treeArray = [];
    }
    /**
     * Clear the Tree array.
     * @return {?}
     */
    clearCache() {
        this.treeArray = [];
    }
    /**
     * Initialize the Fenwick tree with row Heights.
     *
     * @param {?} rows The array of rows which contain the expanded status.
     * @param {?} rowHeight The row height.
     * @param {?} detailRowHeight The detail row height.
     * @return {?}
     */
    initCache(rows, rowHeight, detailRowHeight) {
        const /** @type {?} */ isFn = typeof rowHeight === 'function';
        const /** @type {?} */ isDetailFn = typeof detailRowHeight === 'function';
        if (!isFn && isNaN(rowHeight)) {
            throw new Error(`Row Height cache initialization failed. Please ensure that 'rowHeight' is a
        valid number or function value: (${rowHeight}) when 'scrollbarV' is enabled.`);
        }
        // Add this additional guard in case detailRowHeight is set to 'auto' as it wont work.
        if (!isDetailFn && isNaN(detailRowHeight)) {
            throw new Error(`Row Height cache initialization failed. Please ensure that 'detailRowHeight' is a
        valid number or function value: (${detailRowHeight}) when 'scrollbarV' is enabled.`);
        }
        const /** @type {?} */ n = rows.length;
        this.treeArray = new Array(n);
        for (let /** @type {?} */ i = 0; i < n; ++i) {
            this.treeArray[i] = 0;
        }
        for (let /** @type {?} */ i = 0; i < n; ++i) {
            const /** @type {?} */ row = rows[i];
            let /** @type {?} */ currentRowHeight = rowHeight;
            if (isFn) {
                currentRowHeight = rowHeight(row);
            }
            // Add the detail row height to the already expanded rows.
            // This is useful for the table that goes through a filter or sort.
            if (row && row.$$expanded === 1) {
                if (isDetailFn) {
                    currentRowHeight += detailRowHeight(row, row.$$index);
                }
                else {
                    currentRowHeight += detailRowHeight;
                }
            }
            this.update(i, currentRowHeight);
        }
    }
    /**
     * Given the ScrollY position i.e. sum, provide the rowIndex
     * that is present in the current view port.  Below handles edge cases.
     *
     * @param {?} scrollY - The scrollY position.
     * @return {?}
     */
    getRowIndex(scrollY) {
        if (scrollY === 0)
            return 0;
        return this.calcRowIndex(scrollY);
    }
    /**
     * When a row is expanded or rowHeight is changed, update the height.  This can
     * be utilized in future when Angular Data table supports dynamic row heights.
     *
     *
     * @param {?} atRowIndex Update the data at this index row in the grid.
     * @param {?} byRowHeight Update by the rowHeight provided.
     * @return {?}
     */
    update(atRowIndex, byRowHeight) {
        if (!this.treeArray.length) {
            throw new Error(`Update at index ${atRowIndex} with value ${byRowHeight} failed:
        Row Height cache not initialized.`);
        }
        const /** @type {?} */ n = this.treeArray.length;
        atRowIndex |= 0;
        while (atRowIndex < n) {
            this.treeArray[atRowIndex] += byRowHeight;
            atRowIndex |= (atRowIndex + 1);
        }
    }
    /**
     * Range Sum query from 1 to the rowIndex
     *
     * @param {?} atIndex The row index until which the total height needs to be obtained.
     * @return {?}
     */
    query(atIndex) {
        if (!this.treeArray.length) {
            throw new Error(`query at index ${atIndex} failed: Fenwick tree array not initialized.`);
        }
        let /** @type {?} */ sum = 0;
        atIndex |= 0;
        while (atIndex >= 0) {
            sum += this.treeArray[atIndex];
            atIndex = (atIndex & (atIndex + 1)) - 1;
        }
        return sum;
    }
    /**
     * Find the total height between 2 row indexes
     * @param {?} atIndexA The row index from
     * @param {?} atIndexB The row index to
     * @return {?}
     */
    queryBetween(atIndexA, atIndexB) {
        return this.query(atIndexB) - this.query(atIndexA - 1);
    }
    /**
     * Given the ScrollY position i.e. sum, provide the rowIndex
     * that is present in the current view port.
     *
     * @param {?} sum - The scrollY position.
     * @return {?}
     */
    calcRowIndex(sum) {
        if (!this.treeArray.length)
            return 0;
        let /** @type {?} */ pos = -1;
        const /** @type {?} */ dataLength = this.treeArray.length;
        // Get the highest bit for the block size.
        const /** @type {?} */ highestBit = Math.pow(2, dataLength.toString(2).length - 1);
        for (let /** @type {?} */ blockSize = highestBit; blockSize !== 0; blockSize >>= 1) {
            const /** @type {?} */ nextPos = pos + blockSize;
            if (nextPos < dataLength && sum >= this.treeArray[nextPos]) {
                sum -= this.treeArray[nextPos];
                pos = nextPos;
            }
        }
        return pos + 1;
    }
}

/**
 * Sets the column defaults
 *
 * @export
 * @param {?} columns
 * @return {?}
 */
function setColumnDefaults(columns) {
    if (!columns)
        return;
    for (const /** @type {?} */ column of columns) {
        if (!column.$$id) {
            column.$$id = id();
        }
        // prop can be numeric; zero is valid not a missing prop
        // translate name => prop
        if (column.prop == null && column.name) {
            column.prop = camelCase(column.name);
        }
        if (!column.$$valueGetter) {
            column.$$valueGetter = getterForProp(column.prop);
        }
        // format props if no name passed
        if (column.prop != null && !column.name) {
            column.name = deCamelCase(String(column.prop));
        }
        if (!column.hasOwnProperty('resizeable')) {
            column.resizeable = true;
        }
        if (!column.hasOwnProperty('sortable')) {
            column.sortable = true;
        }
        if (!column.hasOwnProperty('draggable')) {
            column.draggable = true;
        }
        if (!column.hasOwnProperty('canAutoResize')) {
            column.canAutoResize = true;
        }
        if (!column.hasOwnProperty('width')) {
            column.width = 150;
        }
    }
}
/**
 * Translates templates definitions to objects
 *
 * @export
 * @param {?} templates
 * @return {?}
 */
function translateTemplates(templates) {
    const /** @type {?} */ result = [];
    for (const /** @type {?} */ temp of templates) {
        const /** @type {?} */ col = {};
        const /** @type {?} */ props = Object.getOwnPropertyNames(temp);
        for (const /** @type {?} */ prop of props) {
            col[prop] = temp[prop];
        }
        if (temp.headerTemplate) {
            col.headerTemplate = temp.headerTemplate;
        }
        if (temp.cellTemplate) {
            col.cellTemplate = temp.cellTemplate;
        }
        result.push(col);
    }
    return result;
}

if (!document.elementsFromPoint) {
    document.elementsFromPoint = elementsFromPoint;
}
/**
 * Polyfill for `elementsFromPoint`
 *
 * https://developer.mozilla.org/en-US/docs/Web/API/Document/elementsFromPoint
 * https://gist.github.com/iddan/54d5d9e58311b0495a91bf06de661380
 * https://gist.github.com/oslego/7265412
 *
 * @export
 * @param {?} x
 * @param {?} y
 * @return {?}
 */
function elementsFromPoint(x, y) {
    const /** @type {?} */ elements = [];
    const /** @type {?} */ previousPointerEvents = [];
    let /** @type {?} */ current; // TODO: window.getComputedStyle should be used with inferred type (Element)
    let /** @type {?} */ i;
    let /** @type {?} */ d;
    // get all elements via elementFromPoint, and remove them from hit-testing in order
    while ((current = document.elementFromPoint(x, y)) && elements.indexOf(current) === -1 && current != null) {
        // push the element and its current style
        elements.push(current);
        previousPointerEvents.push({
            value: current.style.getPropertyValue('pointer-events'),
            priority: current.style.getPropertyPriority('pointer-events')
        });
        // add "pointer-events: none", to get to the underlying element
        current.style.setProperty('pointer-events', 'none', 'important');
    }
    // restore the previous pointer-events values
    for (i = previousPointerEvents.length; d = previousPointerEvents[--i];) {
        elements[i].style.setProperty('pointer-events', d.value ? d.value : '', d.priority);
    }
    // return our results
    return elements;
}
/*tslint:enable*/

/**
 * Gets the width of the scrollbar.  Nesc for windows
 * http://stackoverflow.com/a/13382873/888165
 *
 * @export
 * \@class ScrollbarHelper
 */
class ScrollbarHelper {
    /**
     * @param {?} document
     */
    constructor(document) {
        this.document = document;
        this.width = this.getWidth();
    }
    /**
     * @return {?}
     */
    getWidth() {
        const /** @type {?} */ outer = this.document.createElement('div');
        outer.style.visibility = 'hidden';
        outer.style.width = '100px';
        outer.style.msOverflowStyle = 'scrollbar';
        this.document.body.appendChild(outer);
        const /** @type {?} */ widthNoScroll = outer.offsetWidth;
        outer.style.overflow = 'scroll';
        const /** @type {?} */ inner = this.document.createElement('div');
        inner.style.width = '100%';
        outer.appendChild(inner);
        const /** @type {?} */ widthWithScroll = inner.offsetWidth;
        outer.parentNode.removeChild(outer);
        return widthNoScroll - widthWithScroll;
    }
}
ScrollbarHelper.decorators = [
    { type: Injectable },
];
/**
 * @nocollapse
 */
ScrollbarHelper.ctorParameters = () => [
    { type: undefined, decorators: [{ type: Inject, args: [DOCUMENT,] },] },
];

class ScrollerComponent {
    /**
     * @param {?} element
     * @param {?} renderer
     */
    constructor(element, renderer) {
        this.renderer = renderer;
        this.scrollbarV = false;
        this.scrollbarH = false;
        this.scroll = new EventEmitter();
        this.scrollYPos = 0;
        this.scrollXPos = 0;
        this.prevScrollYPos = 0;
        this.prevScrollXPos = 0;
        this.element = element.nativeElement;
    }
    /**
     * @return {?}
     */
    ngOnInit() {
        // manual bind so we don't always listen
        if (this.scrollbarV || this.scrollbarH) {
            this.parentElement = this.element.parentElement.parentElement;
            this.onScrollListener = this.renderer.listen(this.parentElement, 'scroll', this.onScrolled.bind(this));
        }
    }
    /**
     * @return {?}
     */
    ngOnDestroy() {
        if (this.scrollbarV || this.scrollbarH) {
            this.onScrollListener();
        }
    }
    /**
     * @param {?} offsetY
     * @return {?}
     */
    setOffset(offsetY) {
        if (this.parentElement) {
            this.parentElement.scrollTop = offsetY;
        }
    }
    /**
     * @param {?} event
     * @return {?}
     */
    onScrolled(event) {
        const /** @type {?} */ dom = (event.currentTarget);
        this.scrollYPos = dom.scrollTop;
        this.scrollXPos = dom.scrollLeft;
        requestAnimationFrame(this.updateOffset.bind(this));
    }
    /**
     * @return {?}
     */
    updateOffset() {
        let /** @type {?} */ direction;
        if (this.scrollYPos < this.prevScrollYPos) {
            direction = 'down';
        }
        else if (this.scrollYPos > this.prevScrollYPos) {
            direction = 'up';
        }
        this.scroll.emit({
            direction,
            scrollYPos: this.scrollYPos,
            scrollXPos: this.scrollXPos
        });
        this.prevScrollYPos = this.scrollYPos;
        this.prevScrollXPos = this.scrollXPos;
    }
}
ScrollerComponent.decorators = [
    { type: Component, args: [{
                selector: 'datatable-scroller',
                template: `
    <ng-content></ng-content>
  `,
                host: {
                    class: 'datatable-scroll'
                }
            },] },
];
/**
 * @nocollapse
 */
ScrollerComponent.ctorParameters = () => [
    { type: ElementRef, },
    { type: Renderer, },
];
ScrollerComponent.propDecorators = {
    'scrollbarV': [{ type: Input },],
    'scrollbarH': [{ type: Input },],
    'scrollHeight': [{ type: HostBinding, args: ['style.height.px',] }, { type: Input },],
    'scrollWidth': [{ type: HostBinding, args: ['style.width.px',] }, { type: Input },],
    'scroll': [{ type: Output },],
};

class DataTableBodyComponent {
    /**
     * Creates an instance of DataTableBodyComponent.
     *
     * \@memberOf DataTableBodyComponent
     */
    constructor() {
        this.selected = [];
        this.scroll = new EventEmitter();
        this.page = new EventEmitter();
        this.activate = new EventEmitter();
        this.select = new EventEmitter();
        this.detailToggle = new EventEmitter();
        this.rowContextmenu = new EventEmitter(false);
        this.rowHeightsCache = new RowHeightCache();
        this.temp = [];
        this.offsetY = 0;
        this.indexes = {};
        /**
         * Get the height of the detail row.
         *
         *
         * \@memberOf DataTableBodyComponent
         */
        this.getDetailRowHeight = (row, index) => {
            if (!this.rowDetail)
                return 0;
            const rowHeight = this.rowDetail.rowHeight;
            return typeof rowHeight === 'function' ? rowHeight(row, index) : rowHeight;
        };
        // declare fn here so we can get access to the `this` property
        this.rowTrackingFn = function (index, row) {
            if (this.trackByProp) {
                return `${row.$$index}-${this.trackByProp}`;
            }
            else {
                return row.$$index;
            }
        }.bind(this);
    }
    /**
     * @param {?} val
     * @return {?}
     */
    set pageSize(val) {
        this._pageSize = val;
        this.recalcLayout();
    }
    /**
     * @return {?}
     */
    get pageSize() {
        return this._pageSize;
    }
    /**
     * @param {?} val
     * @return {?}
     */
    set rows(val) {
        this._rows = val;
        this.recalcLayout();
    }
    /**
     * @return {?}
     */
    get rows() {
        return this._rows;
    }
    /**
     * @param {?} val
     * @return {?}
     */
    set columns(val) {
        this._columns = val;
        const /** @type {?} */ colsByPin = columnsByPin(val);
        this.columnGroupWidths = columnGroupWidths(colsByPin, val);
    }
    /**
     * @return {?}
     */
    get columns() {
        return this._columns;
    }
    /**
     * @param {?} val
     * @return {?}
     */
    set offset(val) {
        this._offset = val;
        this.recalcLayout();
    }
    /**
     * @return {?}
     */
    get offset() {
        return this._offset;
    }
    /**
     * @param {?} val
     * @return {?}
     */
    set rowCount(val) {
        this._rowCount = val;
        this.recalcLayout();
    }
    /**
     * @return {?}
     */
    get rowCount() {
        return this._rowCount;
    }
    /**
     * @return {?}
     */
    get bodyWidth() {
        if (this.scrollbarH) {
            return this.innerWidth + 'px';
        }
        else {
            return '100%';
        }
    }
    /**
     * @param {?} val
     * @return {?}
     */
    set bodyHeight(val) {
        if (this.scrollbarV) {
            this._bodyHeight = val + 'px';
        }
        else {
            this._bodyHeight = 'auto';
        }
        this.recalcLayout();
    }
    /**
     * @return {?}
     */
    get bodyHeight() {
        return this._bodyHeight;
    }
    /**
     * Returns if selection is enabled.
     *
     * \@readonly
     * \@memberOf DataTableBodyComponent
     * @return {?}
     */
    get selectEnabled() {
        return !!this.selectionType;
    }
    /**
     * Property that would calculate the height of scroll bar
     * based on the row heights cache for virtual scroll. Other scenarios
     * calculate scroll height automatically (as height will be undefined).
     *
     * \@readonly
     * \@memberOf DataTableBodyComponent
     * @return {?}
     */
    get scrollHeight() {
        if (this.scrollbarV) {
            return this.rowHeightsCache.query(this.rowCount - 1);
        }
    }
    /**
     * Called after the constructor, initializing input properties
     *
     * \@memberOf DataTableBodyComponent
     * @return {?}
     */
    ngOnInit() {
        if (this.rowDetail) {
            this.listener = this.rowDetail.toggle
                .subscribe(({ type, value }) => {
                if (type === 'row')
                    this.toggleRowExpansion(value);
                if (type === 'all')
                    this.toggleAllRows(value);
            });
        }
    }
    /**
     * Called once, before the instance is destroyed.
     *
     * \@memberOf DataTableBodyComponent
     * @return {?}
     */
    ngOnDestroy() {
        if (this.rowDetail)
            this.listener.unsubscribe();
    }
    /**
     * Updates the Y offset given a new offset.
     *
     *
     * \@memberOf DataTableBodyComponent
     * @param {?=} offset
     * @return {?}
     */
    updateOffsetY(offset) {
        // scroller is missing on empty table
        if (!this.scroller)
            return;
        if (this.scrollbarV && offset) {
            // First get the row Index that we need to move to.
            const /** @type {?} */ rowIndex = this.pageSize * offset;
            offset = this.rowHeightsCache.query(rowIndex - 1);
        }
        this.scroller.setOffset(offset || 0);
    }
    /**
     * Body was scrolled, this is mainly useful for
     * when a user is server-side pagination via virtual scroll.
     *
     *
     * \@memberOf DataTableBodyComponent
     * @param {?} event
     * @return {?}
     */
    onBodyScroll(event) {
        const /** @type {?} */ scrollYPos = event.scrollYPos;
        const /** @type {?} */ scrollXPos = event.scrollXPos;
        // if scroll change, trigger update
        // this is mainly used for header cell positions
        if (this.offsetY !== scrollYPos || this.offsetX !== scrollXPos) {
            this.scroll.emit({
                offsetY: scrollYPos,
                offsetX: scrollXPos
            });
        }
        this.offsetY = scrollYPos;
        this.offsetX = scrollXPos;
        this.updateIndexes();
        this.updatePage(event.direction);
        this.updateRows();
    }
    /**
     * Updates the page given a direction.
     *
     *
     * \@memberOf DataTableBodyComponent
     * @param {?} direction
     * @return {?}
     */
    updatePage(direction) {
        let /** @type {?} */ offset = this.indexes.first / this.pageSize;
        if (direction === 'up') {
            offset = Math.ceil(offset);
        }
        else if (direction === 'down') {
            offset = Math.ceil(offset);
        }
        if (direction !== undefined && !isNaN(offset)) {
            this.page.emit({ offset });
        }
    }
    /**
     * Updates the rows in the view port
     *
     * \@memberOf DataTableBodyComponent
     * @return {?}
     */
    updateRows() {
        const { first, last } = this.indexes;
        let /** @type {?} */ rowIndex = first;
        let /** @type {?} */ idx = 0;
        const /** @type {?} */ temp = [];
        while (rowIndex < last && rowIndex < this.rowCount) {
            const /** @type {?} */ row = this.rows[rowIndex];
            if (row) {
                row.$$index = rowIndex;
                temp[idx] = row;
            }
            idx++;
            rowIndex++;
        }
        this.temp = temp;
    }
    /**
     * Get the row height
     *
     *
     * \@memberOf DataTableBodyComponent
     * @param {?} row
     * @return {?}
     */
    getRowHeight(row) {
        let /** @type {?} */ rowHeight = this.rowHeight;
        // if its a function return it
        if (typeof this.rowHeight === 'function') {
            rowHeight = this.rowHeight(row);
        }
        return rowHeight;
    }
    /**
     * Calculate row height based on the expanded state of the row.
     *
     *
     * \@memberOf DataTableBodyComponent
     * @param {?} row
     * @return {?}
     */
    getRowAndDetailHeight(row) {
        let /** @type {?} */ rowHeight = this.getRowHeight(row);
        // Adding detail row height if its expanded.
        if (row.$$expanded === 1) {
            rowHeight += this.getDetailRowHeight(row);
        }
        return rowHeight;
    }
    /**
     * Calculates the styles for the row so that the rows can be moved in 2D space
     * during virtual scroll inside the DOM.   In the below case the Y position is
     * manipulated.   As an example, if the height of row 0 is 30 px and row 1 is
     * 100 px then following styles are generated:
     *
     * transform: translate3d(0px, 0px, 0px);    ->  row0
     * transform: translate3d(0px, 30px, 0px);   ->  row1
     * transform: translate3d(0px, 130px, 0px);  ->  row2
     *
     * Row heights have to be calculated based on the row heights cache as we wont
     * be able to determine which row is of what height before hand.  In the above
     * case the positionY of the translate3d for row2 would be the sum of all the
     * heights of the rows before it (i.e. row0 and row1).
     *
     *
     * \@memberOf DataTableBodyComponent
     * @param {?} row
     * @return {?}
     */
    getRowsStyles(row) {
        const /** @type {?} */ rowHeight = this.getRowAndDetailHeight(row);
        const /** @type {?} */ styles = {
            height: rowHeight + 'px'
        };
        if (this.scrollbarV) {
            const /** @type {?} */ idx = row ? row.$$index : 0;
            // const pos = idx * rowHeight;
            // The position of this row would be the sum of all row heights
            // until the previous row position.
            const /** @type {?} */ pos = this.rowHeightsCache.query(idx - 1);
            translateXY(styles, 0, pos);
        }
        return styles;
    }
    /**
     * Hides the loading indicator
     *
     *
     * \@memberOf DataTableBodyComponent
     * @return {?}
     */
    hideIndicator() {
        setTimeout(() => this.loadingIndicator = false, 500);
    }
    /**
     * Updates the index of the rows in the viewport
     *
     * \@memberOf DataTableBodyComponent
     * @return {?}
     */
    updateIndexes() {
        let /** @type {?} */ first = 0;
        let /** @type {?} */ last = 0;
        if (this.scrollbarV) {
            // Calculation of the first and last indexes will be based on where the
            // scrollY position would be at.  The last index would be the one
            // that shows up inside the view port the last.
            const /** @type {?} */ height = parseInt(this.bodyHeight, 0);
            first = this.rowHeightsCache.getRowIndex(this.offsetY);
            last = this.rowHeightsCache.getRowIndex(height + this.offsetY) + 1;
        }
        else {
            first = Math.max(this.offset * this.pageSize, 0);
            last = Math.min((first + this.pageSize), this.rowCount);
        }
        this.indexes = { first, last };
    }
    /**
     * Refreshes the full Row Height cache.  Should be used
     * when the entire row array state has changed.
     *
     *
     * \@memberOf DataTableBodyComponent
     * @return {?}
     */
    refreshRowHeightCache() {
        if (!this.scrollbarV)
            return;
        // clear the previous row height cache if already present.
        // this is useful during sorts, filters where the state of the
        // rows array is changed.
        this.rowHeightsCache.clearCache();
        // Initialize the tree only if there are rows inside the tree.
        if (this.rows && this.rows.length) {
            this.rowHeightsCache.initCache(this.rows, this.rowHeight, this.getDetailRowHeight);
        }
    }
    /**
     * Gets the index for the view port
     *
     *
     * \@memberOf DataTableBodyComponent
     * @return {?}
     */
    getAdjustedViewPortIndex() {
        // Capture the row index of the first row that is visible on the viewport.
        // If the scroll bar is just below the row which is highlighted then make that as the
        // first index.
        const /** @type {?} */ viewPortFirstRowIndex = this.indexes.first;
        if (this.scrollbarV) {
            const /** @type {?} */ offsetScroll = this.rowHeightsCache.query(viewPortFirstRowIndex - 1);
            return offsetScroll <= this.offsetY ? viewPortFirstRowIndex - 1 : viewPortFirstRowIndex;
        }
        return viewPortFirstRowIndex;
    }
    /**
     * Toggle the Expansion of the row i.e. if the row is expanded then it will
     * collapse and vice versa.   Note that the expanded status is stored as
     * a part of the row object itself as we have to preserve the expanded row
     * status in case of sorting and filtering of the row set.
     *
     *
     * \@memberOf DataTableBodyComponent
     * @param {?} row
     * @return {?}
     */
    toggleRowExpansion(row) {
        // Capture the row index of the first row that is visible on the viewport.
        const /** @type {?} */ viewPortFirstRowIndex = this.getAdjustedViewPortIndex();
        // If the detailRowHeight is auto --> only in case of non-virtualized scroll
        if (this.scrollbarV) {
            const /** @type {?} */ detailRowHeight = this.getDetailRowHeight(row) * (row.$$expanded ? -1 : 1);
            this.rowHeightsCache.update(row.$$index, detailRowHeight);
        }
        // Update the toggled row and update the heights in the cache.
        row.$$expanded ^= 1;
        this.detailToggle.emit({
            rows: [row],
            currentIndex: viewPortFirstRowIndex
        });
    }
    /**
     * Expand/Collapse all the rows no matter what their state is.
     *
     *
     * \@memberOf DataTableBodyComponent
     * @param {?} expanded
     * @return {?}
     */
    toggleAllRows(expanded) {
        const /** @type {?} */ rowExpanded = expanded ? 1 : 0;
        // Capture the row index of the first row that is visible on the viewport.
        const /** @type {?} */ viewPortFirstRowIndex = this.getAdjustedViewPortIndex();
        for (const /** @type {?} */ row of this.rows) {
            row.$$expanded = rowExpanded;
        }
        if (this.scrollbarV) {
            // Refresh the full row heights cache since every row was affected.
            this.recalcLayout();
        }
        // Emit all rows that have been expanded.
        this.detailToggle.emit({
            rows: this.rows,
            currentIndex: viewPortFirstRowIndex
        });
    }
    /**
     * Recalculates the table
     *
     * \@memberOf DataTableBodyComponent
     * @return {?}
     */
    recalcLayout() {
        this.refreshRowHeightCache();
        this.updateIndexes();
        this.updateRows();
    }
}
DataTableBodyComponent.decorators = [
    { type: Component, args: [{
                selector: 'datatable-body',
                template: `
    <datatable-selection
      #selector
      [selected]="selected"
      [rows]="temp"
      [selectCheck]="selectCheck"
      [selectEnabled]="selectEnabled"
      [selectionType]="selectionType"
      [rowIdentity]="rowIdentity"
      (select)="select.emit($event)"
      (activate)="activate.emit($event)">
      <datatable-progress
        *ngIf="loadingIndicator">
      </datatable-progress>
      <datatable-scroller
        *ngIf="rows?.length"
        [scrollbarV]="scrollbarV"
        [scrollbarH]="scrollbarH"
        [scrollHeight]="scrollHeight"
        [scrollWidth]="columnGroupWidths.total"
        (scroll)="onBodyScroll($event)">
        <datatable-row-wrapper
          *ngFor="let row of temp; let i = index; trackBy: rowTrackingFn;"
          [ngStyle]="getRowsStyles(row)"
          [rowDetail]="rowDetail"
          [detailRowHeight]="getDetailRowHeight(row,i)"
          [row]="row"
          [expanded]="row.$$expanded === 1"
          (rowContextmenu)="rowContextmenu.emit($event)">
          <datatable-body-row
            tabindex="-1"
            [isSelected]="selector.getRowSelected(row)"
            [innerWidth]="innerWidth"
            [offsetX]="offsetX"
            [columns]="columns"
            [rowHeight]="getRowHeight(row)"
            [row]="row"
            [rowClass]="rowClass"
            (activate)="selector.onActivate($event, i)">
          </datatable-body-row>
        </datatable-row-wrapper>
      </datatable-scroller>
      <div
        class="empty-row"
        *ngIf="!rows?.length"
        [innerHTML]="emptyMessage">
      </div>
    </datatable-selection>
  `,
                host: {
                    class: 'datatable-body'
                }
            },] },
];
/**
 * @nocollapse
 */
DataTableBodyComponent.ctorParameters = () => [];
DataTableBodyComponent.propDecorators = {
    'scrollbarV': [{ type: Input },],
    'scrollbarH': [{ type: Input },],
    'loadingIndicator': [{ type: Input },],
    'rowHeight': [{ type: Input },],
    'offsetX': [{ type: Input },],
    'emptyMessage': [{ type: Input },],
    'selectionType': [{ type: Input },],
    'selected': [{ type: Input },],
    'rowIdentity': [{ type: Input },],
    'rowDetail': [{ type: Input },],
    'selectCheck': [{ type: Input },],
    'trackByProp': [{ type: Input },],
    'rowClass': [{ type: Input },],
    'pageSize': [{ type: Input },],
    'rows': [{ type: Input },],
    'columns': [{ type: Input },],
    'offset': [{ type: Input },],
    'rowCount': [{ type: Input },],
    'innerWidth': [{ type: Input },],
    'bodyWidth': [{ type: HostBinding, args: ['style.width',] },],
    'bodyHeight': [{ type: Input }, { type: HostBinding, args: ['style.height',] },],
    'scroll': [{ type: Output },],
    'page': [{ type: Output },],
    'activate': [{ type: Output },],
    'select': [{ type: Output },],
    'detailToggle': [{ type: Output },],
    'rowContextmenu': [{ type: Output },],
    'scroller': [{ type: ViewChild, args: [ScrollerComponent,] },],
};

class DataTableBodyCellComponent {
    /**
     * @param {?} element
     */
    constructor(element) {
        this.activate = new EventEmitter();
        this.isFocused = false;
        this.element = element.nativeElement;
    }
    /**
     * @param {?} val
     * @return {?}
     */
    set sorts(val) {
        this._sorts = val;
        this.calcSortDir = this.calcSortDir(val);
    }
    /**
     * @return {?}
     */
    get sorts() {
        return this._sorts;
    }
    /**
     * @return {?}
     */
    get columnCssClasses() {
        let /** @type {?} */ cls = 'datatable-body-cell';
        if (this.column.cellClass) {
            if (typeof this.column.cellClass === 'string') {
                cls += ' ' + this.column.cellClass;
            }
            else if (typeof this.column.cellClass === 'function') {
                const /** @type {?} */ res = this.column.cellClass({
                    row: this.row,
                    column: this.column,
                    value: this.value
                });
                if (typeof res === 'string') {
                    cls += res;
                }
                else if (typeof res === 'object') {
                    const /** @type {?} */ keys = Object.keys(res);
                    for (const /** @type {?} */ k of keys) {
                        if (res[k] === true)
                            cls += ` ${k}`;
                    }
                }
            }
        }
        if (!this.sortDir)
            cls += ' sort-active';
        if (this.isFocused)
            cls += ' active';
        if (this.sortDir === SortDirection.asc)
            cls += ' sort-asc';
        if (this.sortDir === SortDirection.desc)
            cls += ' sort-desc';
        return cls;
    }
    /**
     * @return {?}
     */
    get width() {
        return this.column.width;
    }
    /**
     * @return {?}
     */
    get height() {
        const /** @type {?} */ height = this.rowHeight;
        if (isNaN(height))
            return height;
        return height + 'px';
    }
    /**
     * @return {?}
     */
    get value() {
        if (!this.row || !this.column)
            return '';
        const /** @type {?} */ val = this.column.$$valueGetter(this.row, this.column.prop);
        const /** @type {?} */ userPipe = this.column.pipe;
        if (userPipe)
            return userPipe.transform(val);
        if (val !== undefined)
            return val;
        return '';
    }
    /**
     * @return {?}
     */
    ngOnDestroy() {
        if (this.cellTemplate) {
            this.cellTemplate.clear();
        }
    }
    /**
     * @return {?}
     */
    onFocus() {
        this.isFocused = true;
    }
    /**
     * @return {?}
     */
    onBlur() {
        this.isFocused = false;
    }
    /**
     * @param {?} event
     * @return {?}
     */
    onClick(event) {
        this.activate.emit({
            type: 'click',
            event,
            row: this.row,
            column: this.column,
            value: this.value,
            cellElement: this.element
        });
    }
    /**
     * @param {?} event
     * @return {?}
     */
    onDblClick(event) {
        this.activate.emit({
            type: 'dblclick',
            event,
            row: this.row,
            column: this.column,
            value: this.value,
            cellElement: this.element
        });
    }
    /**
     * @param {?} event
     * @return {?}
     */
    onKeyDown(event) {
        const /** @type {?} */ keyCode = event.keyCode;
        const /** @type {?} */ isTargetCell = event.target === this.element;
        const /** @type {?} */ isAction = keyCode === Keys.return ||
            keyCode === Keys.down ||
            keyCode === Keys.up ||
            keyCode === Keys.left ||
            keyCode === Keys.right;
        if (isAction && isTargetCell) {
            event.preventDefault();
            event.stopPropagation();
            this.activate.emit({
                type: 'keydown',
                event,
                row: this.row,
                column: this.column,
                value: this.value,
                cellElement: this.element
            });
        }
    }
    /**
     * @param {?} event
     * @return {?}
     */
    onCheckboxChange(event) {
        this.activate.emit({
            type: 'checkbox',
            event,
            row: this.row,
            column: this.column,
            value: this.value,
            cellElement: this.element
        });
    }
    /**
     * @param {?} sorts
     * @return {?}
     */
    calcSortDir(sorts) {
        if (!sorts)
            return;
        const /** @type {?} */ sort = sorts.find((s) => {
            return s.prop === this.column.prop;
        });
        if (sort)
            return sort.dir;
    }
}
DataTableBodyCellComponent.decorators = [
    { type: Component, args: [{
                selector: 'datatable-body-cell',
                template: `
    <div class="datatable-body-cell-label">
      <label
        *ngIf="column.checkboxable" 
        class="datatable-checkbox">
        <input 
          type="checkbox"
          [checked]="isSelected"
          (click)="onCheckboxChange($event)" 
        />
      </label>
      <span
        *ngIf="!column.cellTemplate"
        [title]="value"
        [innerHTML]="value">
      </span>
      <ng-template #cellTemplate
        *ngIf="column.cellTemplate"
        [ngTemplateOutlet]="column.cellTemplate"
        [ngOutletContext]="{ value: value, row: row, column: column }">
      </ng-template>
    </div>
  `,
                host: {
                    class: 'datatable-body-cell'
                }
            },] },
];
/**
 * @nocollapse
 */
DataTableBodyCellComponent.ctorParameters = () => [
    { type: ElementRef, },
];
DataTableBodyCellComponent.propDecorators = {
    'row': [{ type: Input },],
    'column': [{ type: Input },],
    'rowHeight': [{ type: Input },],
    'isSelected': [{ type: Input },],
    'sorts': [{ type: Input },],
    'activate': [{ type: Output },],
    'cellTemplate': [{ type: ViewChild, args: ['cellTemplate', { read: ViewContainerRef },] },],
    'columnCssClasses': [{ type: HostBinding, args: ['class',] },],
    'width': [{ type: HostBinding, args: ['style.width.px',] },],
    'height': [{ type: HostBinding, args: ['style.height',] },],
    'onFocus': [{ type: HostListener, args: ['focus',] },],
    'onBlur': [{ type: HostListener, args: ['blur',] },],
    'onClick': [{ type: HostListener, args: ['click', ['$event'],] },],
    'onDblClick': [{ type: HostListener, args: ['dblclick', ['$event'],] },],
    'onKeyDown': [{ type: HostListener, args: ['keydown', ['$event'],] },],
};

class DataTableBodyRowComponent {
    /**
     * @param {?} scrollbarHelper
     * @param {?} element
     */
    constructor(scrollbarHelper, element) {
        this.scrollbarHelper = scrollbarHelper;
        this.activate = new EventEmitter();
        this.element = element.nativeElement;
    }
    /**
     * @param {?} val
     * @return {?}
     */
    set columns(val) {
        this._columns = val;
        this.recalculateColumns(val);
    }
    /**
     * @return {?}
     */
    get columns() {
        return this._columns;
    }
    /**
     * @param {?} val
     * @return {?}
     */
    set innerWidth(val) {
        this._innerWidth = val;
        this.recalculateColumns();
    }
    /**
     * @return {?}
     */
    get innerWidth() {
        return this._innerWidth;
    }
    /**
     * @return {?}
     */
    get cssClass() {
        let /** @type {?} */ cls = 'datatable-body-row';
        if (this.isSelected)
            cls += ' active';
        if (this.row.$$index % 2 !== 0)
            cls += ' datatable-row-odd';
        if (this.row.$$index % 2 === 0)
            cls += ' datatable-row-even';
        if (this.rowClass) {
            const /** @type {?} */ res = this.rowClass(this.row);
            if (typeof res === 'string') {
                cls += res;
            }
            else if (typeof res === 'object') {
                const /** @type {?} */ keys = Object.keys(res);
                for (const /** @type {?} */ k of keys) {
                    if (res[k] === true)
                        cls += ` ${k}`;
                }
            }
        }
        return cls;
    }
    /**
     * @return {?}
     */
    get columnsTotalWidths() {
        return this.columnGroupWidths.total;
    }
    /**
     * @param {?} index
     * @param {?} colGroup
     * @return {?}
     */
    trackByGroups(index, colGroup) {
        return colGroup.type;
    }
    /**
     * @param {?} index
     * @param {?} column
     * @return {?}
     */
    columnTrackingFn(index, column) {
        return column.$$id;
    }
    /**
     * @param {?} group
     * @return {?}
     */
    stylesByGroup(group) {
        const /** @type {?} */ widths = this.columnGroupWidths;
        const /** @type {?} */ offsetX = this.offsetX;
        const /** @type {?} */ styles = {
            width: `${widths[group]}px`
        };
        if (group === 'left') {
            translateXY(styles, offsetX, 0);
        }
        else if (group === 'right') {
            const /** @type {?} */ bodyWidth = parseInt(this.innerWidth + '', 0);
            const /** @type {?} */ totalDiff = widths.total - bodyWidth;
            const /** @type {?} */ offsetDiff = totalDiff - offsetX;
            const /** @type {?} */ offset = (offsetDiff + this.scrollbarHelper.width) * -1;
            translateXY(styles, offset, 0);
        }
        return styles;
    }
    /**
     * @param {?} event
     * @param {?} index
     * @return {?}
     */
    onActivate(event, index) {
        event.cellIndex = index;
        event.rowElement = this.element;
        this.activate.emit(event);
    }
    /**
     * @param {?} event
     * @return {?}
     */
    onKeyDown(event) {
        const /** @type {?} */ keyCode = event.keyCode;
        const /** @type {?} */ isTargetRow = event.target === this.element;
        const /** @type {?} */ isAction = keyCode === Keys.return ||
            keyCode === Keys.down ||
            keyCode === Keys.up ||
            keyCode === Keys.left ||
            keyCode === Keys.right;
        if (isAction && isTargetRow) {
            event.preventDefault();
            event.stopPropagation();
            this.activate.emit({
                type: 'keydown',
                event,
                row: this.row,
                rowElement: this.element
            });
        }
    }
    /**
     * @param {?=} val
     * @return {?}
     */
    recalculateColumns(val = this.columns) {
        const /** @type {?} */ colsByPin = columnsByPin(val);
        this.columnsByPin = columnsByPinArr(val);
        this.columnGroupWidths = columnGroupWidths(colsByPin, val);
    }
}
DataTableBodyRowComponent.decorators = [
    { type: Component, args: [{
                selector: 'datatable-body-row',
                template: `
    <div
      *ngFor="let colGroup of columnsByPin; let i = index; trackBy: trackByGroups"
      class="datatable-row-{{colGroup.type}} datatable-row-group"
      [ngStyle]="stylesByGroup(colGroup.type)">
      <datatable-body-cell
        *ngFor="let column of colGroup.columns; let ii = index; trackBy: columnTrackingFn"
        tabindex="-1"
        [row]="row"
        [isSelected]="isSelected"
        [column]="column"
        [rowHeight]="rowHeight"
        (activate)="onActivate($event, ii)">
      </datatable-body-cell>
    </div>
  `
            },] },
];
/**
 * @nocollapse
 */
DataTableBodyRowComponent.ctorParameters = () => [
    { type: ScrollbarHelper, },
    { type: ElementRef, },
];
DataTableBodyRowComponent.propDecorators = {
    'columns': [{ type: Input },],
    'innerWidth': [{ type: Input },],
    'rowClass': [{ type: Input },],
    'row': [{ type: Input },],
    'offsetX': [{ type: Input },],
    'isSelected': [{ type: Input },],
    'cssClass': [{ type: HostBinding, args: ['class',] },],
    'rowHeight': [{ type: HostBinding, args: ['style.height.px',] }, { type: Input },],
    'columnsTotalWidths': [{ type: HostBinding, args: ['style.width.px',] },],
    'activate': [{ type: Output },],
    'onKeyDown': [{ type: HostListener, args: ['keydown', ['$event'],] },],
};

class ProgressBarComponent {
}
ProgressBarComponent.decorators = [
    { type: Component, args: [{
                selector: 'datatable-progress',
                template: `
    <div class="progress-linear" role="progressbar">
      <div class="container">
        <div class="bar"></div>
      </div>
    </div>
  `,
                changeDetection: ChangeDetectionStrategy.OnPush
            },] },
];
/**
 * @nocollapse
 */
ProgressBarComponent.ctorParameters = () => [];

class DataTableRowWrapperComponent {
    constructor() {
        this.expanded = false;
        this.rowContextmenu = new EventEmitter(false);
    }
    /**
     * @param {?} $event
     * @return {?}
     */
    onContextmenu($event) {
        this.rowContextmenu.emit({ event: $event, row: this.row });
    }
}
DataTableRowWrapperComponent.decorators = [
    { type: Component, args: [{
                selector: 'datatable-row-wrapper',
                template: `
    <ng-content></ng-content>
    <div 
      *ngIf="expanded"
      [style.height.px]="detailRowHeight" 
      class="datatable-row-detail">
      <ng-template
        *ngIf="rowDetail && rowDetail.template"
        [ngTemplateOutlet]="rowDetail.template"
        [ngOutletContext]="{ row: row }">
      </ng-template>
    </div>
  `,
                host: {
                    class: 'datatable-row-wrapper'
                }
            },] },
];
/**
 * @nocollapse
 */
DataTableRowWrapperComponent.ctorParameters = () => [];
DataTableRowWrapperComponent.propDecorators = {
    'rowDetail': [{ type: Input },],
    'detailRowHeight': [{ type: Input },],
    'expanded': [{ type: Input },],
    'row': [{ type: Input },],
    'rowContextmenu': [{ type: Output },],
    'onContextmenu': [{ type: HostListener, args: ['contextmenu', ['$event'],] },],
};

class DataTableSelectionComponent {
    constructor() {
        this.activate = new EventEmitter();
        this.select = new EventEmitter();
    }
    /**
     * @param {?} event
     * @param {?} index
     * @param {?} row
     * @return {?}
     */
    selectRow(event, index, row) {
        if (!this.selectEnabled)
            return;
        const /** @type {?} */ chkbox = this.selectionType === SelectionType.checkbox;
        const /** @type {?} */ multi = this.selectionType === SelectionType.multi;
        const /** @type {?} */ multiClick = this.selectionType === SelectionType.multiClick;
        let /** @type {?} */ selected = [];
        if (multi || chkbox || multiClick) {
            if (event.shiftKey) {
                selected = selectRowsBetween([], this.rows, index, this.prevIndex, this.getRowSelectedIdx.bind(this));
            }
            else if (event.ctrlKey || multiClick || chkbox) {
                selected = selectRows([...this.selected], row, this.getRowSelectedIdx.bind(this));
            }
            else {
                selected = selectRows([], row, this.getRowSelectedIdx.bind(this));
            }
        }
        else {
            selected = selectRows([], row, this.getRowSelectedIdx.bind(this));
        }
        if (typeof this.selectCheck === 'function') {
            selected = selected.filter(this.selectCheck.bind(this));
        }
        this.selected.splice(0, this.selected.length);
        this.selected.push(...selected);
        this.prevIndex = index;
        this.select.emit({
            selected
        });
    }
    /**
     * @param {?} model
     * @param {?} index
     * @return {?}
     */
    onActivate(model, index) {
        const { type, event, row } = model;
        const /** @type {?} */ chkbox = this.selectionType === SelectionType.checkbox;
        const /** @type {?} */ select = (!chkbox && (type === 'click' || type === 'dblclick')) ||
            (chkbox && type === 'checkbox');
        if (select) {
            this.selectRow(event, index, row);
        }
        else if (type === 'keydown') {
            if (((event)).keyCode === Keys.return) {
                this.selectRow(event, index, row);
            }
            else {
                this.onKeyboardFocus(model);
            }
        }
        this.activate.emit(model);
    }
    /**
     * @param {?} model
     * @return {?}
     */
    onKeyboardFocus(model) {
        const { keyCode } = (model.event);
        const /** @type {?} */ shouldFocus = keyCode === Keys.up ||
            keyCode === Keys.down ||
            keyCode === Keys.right ||
            keyCode === Keys.left;
        if (shouldFocus) {
            const /** @type {?} */ isCellSelection = this.selectionType === SelectionType.cell;
            if (!model.cellElement || !isCellSelection) {
                this.focusRow(model.rowElement, keyCode);
            }
            else if (isCellSelection) {
                this.focusCell(model.cellElement, model.rowElement, keyCode, model.cellIndex);
            }
        }
    }
    /**
     * @param {?} rowElement
     * @param {?} keyCode
     * @return {?}
     */
    focusRow(rowElement, keyCode) {
        const /** @type {?} */ nextRowElement = this.getPrevNextRow(rowElement, keyCode);
        if (nextRowElement)
            nextRowElement.focus();
    }
    /**
     * @param {?} rowElement
     * @param {?} keyCode
     * @return {?}
     */
    getPrevNextRow(rowElement, keyCode) {
        const /** @type {?} */ parentElement = rowElement.parentElement;
        if (parentElement) {
            let /** @type {?} */ focusElement;
            if (keyCode === Keys.up) {
                focusElement = parentElement.previousElementSibling;
            }
            else if (keyCode === Keys.down) {
                focusElement = parentElement.nextElementSibling;
            }
            if (focusElement && focusElement.children.length) {
                return focusElement.children[0];
            }
        }
    }
    /**
     * @param {?} cellElement
     * @param {?} rowElement
     * @param {?} keyCode
     * @param {?} cellIndex
     * @return {?}
     */
    focusCell(cellElement, rowElement, keyCode, cellIndex) {
        let /** @type {?} */ nextCellElement;
        if (keyCode === Keys.left) {
            nextCellElement = cellElement.previousElementSibling;
        }
        else if (keyCode === Keys.right) {
            nextCellElement = cellElement.nextElementSibling;
        }
        else if (keyCode === Keys.up || keyCode === Keys.down) {
            const /** @type {?} */ nextRowElement = this.getPrevNextRow(rowElement, keyCode);
            if (nextRowElement) {
                const /** @type {?} */ children = nextRowElement.getElementsByClassName('datatable-body-cell');
                if (children.length)
                    nextCellElement = children[cellIndex];
            }
        }
        if (nextCellElement)
            nextCellElement.focus();
    }
    /**
     * @param {?} row
     * @return {?}
     */
    getRowSelected(row) {
        return this.getRowSelectedIdx(row, this.selected) > -1;
    }
    /**
     * @param {?} row
     * @param {?} selected
     * @return {?}
     */
    getRowSelectedIdx(row, selected) {
        if (!selected || !selected.length)
            return -1;
        const /** @type {?} */ rowId = this.rowIdentity(row);
        return selected.findIndex((r) => {
            const /** @type {?} */ id = this.rowIdentity(r);
            return id === rowId;
        });
    }
}
DataTableSelectionComponent.decorators = [
    { type: Component, args: [{
                selector: 'datatable-selection',
                template: `
    <ng-content></ng-content>
  `
            },] },
];
/**
 * @nocollapse
 */
DataTableSelectionComponent.ctorParameters = () => [];
DataTableSelectionComponent.propDecorators = {
    'rows': [{ type: Input },],
    'selected': [{ type: Input },],
    'selectEnabled': [{ type: Input },],
    'selectionType': [{ type: Input },],
    'rowIdentity': [{ type: Input },],
    'selectCheck': [{ type: Input },],
    'activate': [{ type: Output },],
    'select': [{ type: Output },],
};

class DataTableColumnHeaderDirective {
    /**
     * @param {?} template
     */
    constructor(template) {
        this.template = template;
    }
}
DataTableColumnHeaderDirective.decorators = [
    { type: Directive, args: [{ selector: '[ngx-datatable-header-template]' },] },
];
/**
 * @nocollapse
 */
DataTableColumnHeaderDirective.ctorParameters = () => [
    { type: TemplateRef, },
];

class DataTableColumnCellDirective {
    /**
     * @param {?} template
     */
    constructor(template) {
        this.template = template;
    }
}
DataTableColumnCellDirective.decorators = [
    { type: Directive, args: [{ selector: '[ngx-datatable-cell-template]' },] },
];
/**
 * @nocollapse
 */
DataTableColumnCellDirective.ctorParameters = () => [
    { type: TemplateRef, },
];

class DataTableColumnDirective {
}
DataTableColumnDirective.decorators = [
    { type: Directive, args: [{ selector: 'ngx-datatable-column' },] },
];
/**
 * @nocollapse
 */
DataTableColumnDirective.ctorParameters = () => [];
DataTableColumnDirective.propDecorators = {
    'name': [{ type: Input },],
    'prop': [{ type: Input },],
    'frozenLeft': [{ type: Input },],
    'frozenRight': [{ type: Input },],
    'flexGrow': [{ type: Input },],
    'resizeable': [{ type: Input },],
    'comparator': [{ type: Input },],
    'pipe': [{ type: Input },],
    'sortable': [{ type: Input },],
    'draggable': [{ type: Input },],
    'canAutoResize': [{ type: Input },],
    'minWidth': [{ type: Input },],
    'width': [{ type: Input },],
    'maxWidth': [{ type: Input },],
    'checkboxable': [{ type: Input },],
    'headerCheckboxable': [{ type: Input },],
    'headerClass': [{ type: Input },],
    'cellClass': [{ type: Input },],
    'cellTemplate': [{ type: Input }, { type: ContentChild, args: [DataTableColumnCellDirective, { read: TemplateRef },] },],
    'headerTemplate': [{ type: Input }, { type: ContentChild, args: [DataTableColumnHeaderDirective, { read: TemplateRef },] },],
};

class DatatableRowDetailTemplateDirective {
    /**
     * @param {?} template
     */
    constructor(template) {
        this.template = template;
    }
}
DatatableRowDetailTemplateDirective.decorators = [
    { type: Directive, args: [{
                selector: '[ngx-datatable-row-detail-template]'
            },] },
];
/**
 * @nocollapse
 */
DatatableRowDetailTemplateDirective.ctorParameters = () => [
    { type: TemplateRef, },
];

class DatatableRowDetailDirective {
    constructor() {
        /**
         * The detail row height is required especially
         * when virtual scroll is enabled.
         *
         * \@memberOf DatatableComponent
         */
        this.rowHeight = 0;
        /**
         * Row detail row visbility was toggled.
         *
         * \@memberOf DatatableComponent
         */
        this.toggle = new EventEmitter();
    }
    /**
     * Toggle the expansion of the row
     *
     * @param {?} row
     * @return {?}
     */
    toggleExpandRow(row) {
        this.toggle.emit({
            type: 'row',
            value: row
        });
    }
    /**
     * API method to expand all the rows.
     *
     * \@memberOf DatatableComponent
     * @return {?}
     */
    expandAllRows() {
        this.toggle.emit({
            type: 'all',
            value: true
        });
    }
    /**
     * API method to collapse all the rows.
     *
     * \@memberOf DatatableComponent
     * @return {?}
     */
    collapseAllRows() {
        this.toggle.emit({
            type: 'all',
            value: false
        });
    }
}
DatatableRowDetailDirective.decorators = [
    { type: Directive, args: [{ selector: 'ngx-datatable-row-detail' },] },
];
/**
 * @nocollapse
 */
DatatableRowDetailDirective.ctorParameters = () => [];
DatatableRowDetailDirective.propDecorators = {
    'rowHeight': [{ type: Input },],
    'template': [{ type: Input }, { type: ContentChild, args: [DatatableRowDetailTemplateDirective, { read: TemplateRef },] },],
    'toggle': [{ type: Output },],
};

class DataTableFooterComponent {
    constructor() {
        this.selectedCount = 0;
        this.page = new EventEmitter();
    }
    /**
     * @return {?}
     */
    get isVisible() {
        return (this.rowCount / this.pageSize) > 1;
    }
    /**
     * @return {?}
     */
    get curPage() {
        return this.offset + 1;
    }
}
DataTableFooterComponent.decorators = [
    { type: Component, args: [{
                selector: 'datatable-footer',
                template: `
    <div
      class="datatable-footer-inner"
      [ngClass]="{'selected-count': selectedMessage}"
      [style.height.px]="footerHeight">
      <ng-template
        *ngIf="footerTemplate"
        [ngTemplateOutlet]="footerTemplate.template"
        [ngOutletContext]="{ 
          rowCount: rowCount, 
          pageSize: pageSize, 
          selectedCount: selectedCount,
          curPage: curPage,
          offset: offset
        }">
      </ng-template>
      <div class="page-count" *ngIf="!footerTemplate">
        <span *ngIf="selectedMessage">
          {{selectedCount.toLocaleString()}} {{selectedMessage}} / 
        </span>
        {{rowCount.toLocaleString()}} {{totalMessage}}
      </div>
      <datatable-pager *ngIf="!footerTemplate"
        [pagerLeftArrowIcon]="pagerLeftArrowIcon"
        [pagerRightArrowIcon]="pagerRightArrowIcon"
        [pagerPreviousIcon]="pagerPreviousIcon"
        [pagerNextIcon]="pagerNextIcon"
        [page]="curPage"
        [size]="pageSize"
        [count]="rowCount"
        [hidden]="!isVisible"
        (change)="page.emit($event)">
      </datatable-pager>
    </div>
  `,
                host: {
                    class: 'datatable-footer'
                },
                changeDetection: ChangeDetectionStrategy.OnPush
            },] },
];
/**
 * @nocollapse
 */
DataTableFooterComponent.ctorParameters = () => [];
DataTableFooterComponent.propDecorators = {
    'footerHeight': [{ type: Input },],
    'rowCount': [{ type: Input },],
    'pageSize': [{ type: Input },],
    'offset': [{ type: Input },],
    'pagerLeftArrowIcon': [{ type: Input },],
    'pagerRightArrowIcon': [{ type: Input },],
    'pagerPreviousIcon': [{ type: Input },],
    'pagerNextIcon': [{ type: Input },],
    'totalMessage': [{ type: Input },],
    'footerTemplate': [{ type: Input },],
    'selectedCount': [{ type: Input },],
    'selectedMessage': [{ type: Input },],
    'page': [{ type: Output },],
};

class DataTablePagerComponent {
    constructor() {
        this.change = new EventEmitter();
        this._count = 0;
        this._page = 1;
        this._size = 0;
    }
    /**
     * @param {?} val
     * @return {?}
     */
    set size(val) {
        this._size = val;
        this.pages = this.calcPages();
    }
    /**
     * @return {?}
     */
    get size() {
        return this._size;
    }
    /**
     * @param {?} val
     * @return {?}
     */
    set count(val) {
        this._count = val;
        this.pages = this.calcPages();
    }
    /**
     * @return {?}
     */
    get count() {
        return this._count;
    }
    /**
     * @param {?} val
     * @return {?}
     */
    set page(val) {
        this._page = val;
        this.pages = this.calcPages();
    }
    /**
     * @return {?}
     */
    get page() {
        return this._page;
    }
    /**
     * @return {?}
     */
    get totalPages() {
        const /** @type {?} */ count = this.size < 1 ? 1 : Math.ceil(this.count / this.size);
        return Math.max(count || 0, 1);
    }
    /**
     * @return {?}
     */
    canPrevious() {
        return this.page > 1;
    }
    /**
     * @return {?}
     */
    canNext() {
        return this.page < this.totalPages;
    }
    /**
     * @return {?}
     */
    prevPage() {
        this.selectPage(this.page - 1);
    }
    /**
     * @return {?}
     */
    nextPage() {
        this.selectPage(this.page + 1);
    }
    /**
     * @param {?} page
     * @return {?}
     */
    selectPage(page) {
        if (page > 0 && page <= this.totalPages && page !== this.page) {
            this.page = page;
            this.change.emit({
                page
            });
        }
    }
    /**
     * @param {?=} page
     * @return {?}
     */
    calcPages(page) {
        const /** @type {?} */ pages = [];
        let /** @type {?} */ startPage = 1;
        let /** @type {?} */ endPage = this.totalPages;
        const /** @type {?} */ maxSize = 5;
        const /** @type {?} */ isMaxSized = maxSize < this.totalPages;
        page = page || this.page;
        if (isMaxSized) {
            startPage = ((Math.ceil(page / maxSize) - 1) * maxSize) + 1;
            endPage = Math.min(startPage + maxSize - 1, this.totalPages);
        }
        for (let /** @type {?} */ num = startPage; num <= endPage; num++) {
            pages.push({
                number: num,
                text: /** @type {?} */ ((num))
            });
        }
        return pages;
    }
}
DataTablePagerComponent.decorators = [
    { type: Component, args: [{
                selector: 'datatable-pager',
                template: `
    <ul class="pager">
      <li [class.disabled]="!canPrevious()">
        <a
          href="javascript:void(0)"
          (click)="selectPage(1)">
          <i class="{{pagerPreviousIcon}}"></i>
        </a>
      </li>
      <li [class.disabled]="!canPrevious()">
        <a
          href="javascript:void(0)"
          (click)="prevPage()">
          <i class="{{pagerLeftArrowIcon}}"></i>
        </a>
      </li>
      <li
        class="pages"
        *ngFor="let pg of pages"
        [class.active]="pg.number === page">
        <a
          href="javascript:void(0)"
          (click)="selectPage(pg.number)">
          {{pg.text}}
        </a>
      </li>
      <li [class.disabled]="!canNext()">
        <a
          href="javascript:void(0)"
          (click)="nextPage()">
          <i class="{{pagerRightArrowIcon}}"></i>
        </a>
      </li>
      <li [class.disabled]="!canNext()">
        <a
          href="javascript:void(0)"
          (click)="selectPage(totalPages)">
          <i class="{{pagerNextIcon}}"></i>
        </a>
      </li>
    </ul>
  `,
                host: {
                    class: 'datatable-pager'
                },
                changeDetection: ChangeDetectionStrategy.OnPush
            },] },
];
/**
 * @nocollapse
 */
DataTablePagerComponent.ctorParameters = () => [];
DataTablePagerComponent.propDecorators = {
    'pagerLeftArrowIcon': [{ type: Input },],
    'pagerRightArrowIcon': [{ type: Input },],
    'pagerPreviousIcon': [{ type: Input },],
    'pagerNextIcon': [{ type: Input },],
    'size': [{ type: Input },],
    'count': [{ type: Input },],
    'page': [{ type: Input },],
    'change': [{ type: Output },],
};

class DataTableFooterTemplateDirective {
    /**
     * @param {?} template
     */
    constructor(template) {
        this.template = template;
    }
}
DataTableFooterTemplateDirective.decorators = [
    { type: Directive, args: [{ selector: '[ngx-datatable-footer-template]' },] },
];
/**
 * @nocollapse
 */
DataTableFooterTemplateDirective.ctorParameters = () => [
    { type: TemplateRef, },
];

class DatatableFooterDirective {
}
DatatableFooterDirective.decorators = [
    { type: Directive, args: [{ selector: 'ngx-datatable-footer' },] },
];
/**
 * @nocollapse
 */
DatatableFooterDirective.ctorParameters = () => [];
DatatableFooterDirective.propDecorators = {
    'footerHeight': [{ type: Input },],
    'totalMessage': [{ type: Input },],
    'selectedMessage': [{ type: Input },],
    'pagerLeftArrowIcon': [{ type: Input },],
    'pagerRightArrowIcon': [{ type: Input },],
    'pagerPreviousIcon': [{ type: Input },],
    'pagerNextIcon': [{ type: Input },],
    'template': [{ type: Input }, { type: ContentChild, args: [DataTableFooterTemplateDirective, { read: TemplateRef },] },],
};

var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var __metadata = (this && this.__metadata) || function (k, v) {
    if (typeof Reflect === "object" && typeof Reflect.metadata === "function") return Reflect.metadata(k, v);
};
class DatatableComponent {
    /**
     * @param {?} scrollbarHelper
     * @param {?} element
     * @param {?} differs
     */
    constructor(scrollbarHelper, element, differs) {
        this.scrollbarHelper = scrollbarHelper;
        /**
         * List of row objects that should be
         * represented as selected in the grid.
         * Default value: `[]`
         *
         * \@memberOf DatatableComponent
         */
        this.selected = [];
        /**
         * Enable vertical scrollbars
         *
         * \@memberOf DatatableComponent
         */
        this.scrollbarV = false;
        /**
         * Enable horz scrollbars
         *
         * \@memberOf DatatableComponent
         */
        this.scrollbarH = false;
        /**
         * The row height; which is necessary
         * to calculate the height for the lazy rendering.
         *
         * \@memberOf DatatableComponent
         */
        this.rowHeight = 30;
        /**
         * Type of column width distribution formula.
         * Example: flex, force, standard
         *
         * \@memberOf DatatableComponent
         */
        this.columnMode = ColumnMode.standard;
        /**
         * The minimum header height in pixels.
         * Pass a falsey for no header
         *
         * \@memberOf DatatableComponent
         */
        this.headerHeight = 30;
        /**
         * The minimum footer height in pixels.
         * Pass falsey for no footer
         *
         * \@memberOf DatatableComponent
         */
        this.footerHeight = 0;
        /**
         * If the table should use external paging
         * otherwise its assumed that all data is preloaded.
         *
         * \@memberOf DatatableComponent
         */
        this.externalPaging = false;
        /**
         * If the table should use external sorting or
         * the built-in basic sorting.
         *
         * \@memberOf DatatableComponent
         */
        this.externalSorting = false;
        /**
         * The page size to be shown.
         * Default value: `undefined`
         *
         * \@memberOf DatatableComponent
         */
        this.limit = undefined;
        /**
         * The current offset ( page - 1 ) shown.
         * Default value: `0`
         *
         * \@memberOf DatatableComponent
         */
        this.offset = 0;
        /**
         * Show the linear loading bar.
         * Default value: `false`
         *
         * \@memberOf DatatableComponent
         */
        this.loadingIndicator = false;
        /**
         * Enable/Disable ability to re-order columns
         * by dragging them.
         *
         * \@memberOf DatatableComponent
         */
        this.reorderable = true;
        /**
         * The type of sorting
         *
         * \@memberOf DatatableComponent
         */
        this.sortType = SortType.single;
        /**
         * Array of sorted columns by property and type.
         * Default value: `[]`
         *
         * \@memberOf DatatableComponent
         */
        this.sorts = [];
        /**
         * Css class overrides
         *
         * \@memberOf DatatableComponent
         */
        this.cssClasses = {
            sortAscending: 'icon-up',
            sortDescending: 'icon-down',
            pagerLeftArrow: 'icon-left',
            pagerRightArrow: 'icon-right',
            pagerPrevious: 'icon-prev',
            pagerNext: 'icon-skip'
        };
        /**
         * Message overrides for localization
         *
         * emptyMessage     [default] = 'No data to display'
         * totalMessage     [default] = 'total'
         * selectedMessage  [default] = 'selected'
         *
         * \@memberOf DatatableComponent
         */
        this.messages = {
            // Message to show when array is presented
            // but contains no values
            emptyMessage: 'No data to display',
            // Footer total message
            totalMessage: 'total',
            // Footer selected message
            selectedMessage: 'selected'
        };
        /**
         * This will be used when displaying or selecting rows.
         * when tracking/comparing them, we'll use the value of this fn,
         *
         * (`fn(x) === fn(y)` instead of `x === y`)
         *
         * \@memberOf DatatableComponent
         */
        this.rowIdentity = ((x) => x);
        /**
         * Body was scrolled typically in a `scrollbarV:true` scenario.
         *
         * \@memberOf DatatableComponent
         */
        this.scroll = new EventEmitter();
        /**
         * A cell or row was focused via keyboard or mouse click.
         *
         * \@memberOf DatatableComponent
         */
        this.activate = new EventEmitter();
        /**
         * A cell or row was selected.
         *
         * \@memberOf DatatableComponent
         */
        this.select = new EventEmitter();
        /**
         * Column sort was invoked.
         *
         * \@memberOf DatatableComponent
         */
        this.sort = new EventEmitter();
        /**
         * The table was paged either triggered by the pager or the body scroll.
         *
         * \@memberOf DatatableComponent
         */
        this.page = new EventEmitter();
        /**
         * Columns were re-ordered.
         *
         * \@memberOf DatatableComponent
         */
        this.reorder = new EventEmitter();
        /**
         * Column was resized.
         *
         * \@memberOf DatatableComponent
         */
        this.resize = new EventEmitter();
        /**
         * The context menu was invoked on a row.
         *
         * \@memberOf DatatableComponent
         */
        this.rowContextmenu = new EventEmitter(false);
        this.rowCount = 0;
        this.offsetX = 0;
        this._count = 0;
        // get ref to elm for measuring
        this.element = element.nativeElement;
        this.rowDiffer = differs.find({}).create(null);
    }
    /**
     * Rows that are displayed in the table.
     *
     * \@memberOf DatatableComponent
     * @param {?} val
     * @return {?}
     */
    set rows(val) {
        // auto sort on new updates
        if (!this.externalSorting) {
            val = sortRows(val, this.columns, this.sorts);
        }
        this._rows = val;
        // recalculate sizes/etc
        this.recalculate();
    }
    /**
     * Gets the rows.
     *
     * \@readonly
     * \@memberOf DatatableComponent
     * @return {?}
     */
    get rows() {
        return this._rows;
    }
    /**
     * Columns to be displayed.
     *
     * \@memberOf DatatableComponent
     * @param {?} val
     * @return {?}
     */
    set columns(val) {
        if (val) {
            setColumnDefaults(val);
            this.recalculateColumns(val);
        }
        this._columns = val;
    }
    /**
     * Get the columns.
     *
     * \@readonly
     * \@memberOf DatatableComponent
     * @return {?}
     */
    get columns() {
        return this._columns;
    }
    /**
     * The total count of all rows.
     * Default value: `0`
     *
     * \@memberOf DatatableComponent
     * @param {?} val
     * @return {?}
     */
    set count(val) {
        this._count = val;
        // recalculate sizes/etc
        this.recalculate();
    }
    /**
     * Gets the count.
     *
     * \@readonly
     * \@memberOf DatatableComponent
     * @return {?}
     */
    get count() {
        return this._count;
    }
    /**
     * CSS class applied if the header height if fixed height.
     *
     * \@readonly
     * \@memberOf DatatableComponent
     * @return {?}
     */
    get isFixedHeader() {
        const /** @type {?} */ headerHeight = this.headerHeight;
        return (typeof headerHeight === 'string') ?
            ((headerHeight)) !== 'auto' : true;
    }
    /**
     * CSS class applied to the root element if
     * the row heights are fixed heights.
     *
     * \@readonly
     * \@memberOf DatatableComponent
     * @return {?}
     */
    get isFixedRow() {
        const /** @type {?} */ rowHeight = this.rowHeight;
        return (typeof rowHeight === 'string') ?
            ((rowHeight)) !== 'auto' : true;
    }
    /**
     * CSS class applied to root element if
     * vertical scrolling is enabled.
     *
     * \@readonly
     * \@memberOf DatatableComponent
     * @return {?}
     */
    get isVertScroll() {
        return this.scrollbarV;
    }
    /**
     * CSS class applied to the root element
     * if the horziontal scrolling is enabled.
     *
     * \@readonly
     * \@memberOf DatatableComponent
     * @return {?}
     */
    get isHorScroll() {
        return this.scrollbarH;
    }
    /**
     * CSS class applied to root element is selectable.
     *
     * \@readonly
     * \@memberOf DatatableComponent
     * @return {?}
     */
    get isSelectable() {
        return this.selectionType !== undefined;
    }
    /**
     * CSS class applied to root is checkbox selection.
     *
     * \@readonly
     * \@memberOf DatatableComponent
     * @return {?}
     */
    get isCheckboxSelection() {
        return this.selectionType === SelectionType.checkbox;
    }
    /**
     * CSS class applied to root if cell selection.
     *
     * \@readonly
     * \@memberOf DatatableComponent
     * @return {?}
     */
    get isCellSelection() {
        return this.selectionType === SelectionType.cell;
    }
    /**
     * CSS class applied to root if single select.
     *
     * \@readonly
     * \@memberOf DatatableComponent
     * @return {?}
     */
    get isSingleSelection() {
        return this.selectionType === SelectionType.single;
    }
    /**
     * CSS class added to root element if mulit select
     *
     * \@readonly
     * \@memberOf DatatableComponent
     * @return {?}
     */
    get isMultiSelection() {
        return this.selectionType === SelectionType.multi;
    }
    /**
     * CSS class added to root element if mulit click select
     *
     * \@readonly
     * \@memberOf DatatableComponent
     * @return {?}
     */
    get isMultiClickSelection() {
        return this.selectionType === SelectionType.multiClick;
    }
    /**
     * Column templates gathered from `ContentChildren`
     * if described in your markup.
     *
     * \@memberOf DatatableComponent
     * @param {?} val
     * @return {?}
     */
    set columnTemplates(val) {
        this._columnTemplates = val;
        if (val) {
            // only set this if results were brought back
            const /** @type {?} */ arr = val.toArray();
            if (arr.length) {
                // translate them to normal objects
                this.columns = translateTemplates(arr);
            }
        }
    }
    /**
     * Returns the column templates.
     *
     * \@readonly
     * \@memberOf DatatableComponent
     * @return {?}
     */
    get columnTemplates() {
        return this._columnTemplates;
    }
    /**
     * Returns if all rows are selected.
     *
     * \@readonly
     * \@memberOf DatatableComponent
     * @return {?}
     */
    get allRowsSelected() {
        return this.selected &&
            this.rows &&
            this.selected.length === this.rows.length;
    }
    /**
     * Lifecycle hook that is called after data-bound
     * properties of a directive are initialized.
     *
     * \@memberOf DatatableComponent
     * @return {?}
     */
    ngOnInit() {
        // need to call this immediatly to size
        // if the table is hidden the visibility
        // listener will invoke this itself upon show
        this.recalculate();
    }
    /**
     * Lifecycle hook that is called after a component's
     * view has been fully initialized.
     *
     * \@memberOf DatatableComponent
     * @return {?}
     */
    ngAfterViewInit() {
        if (!this.externalSorting) {
            const /** @type {?} */ val = sortRows(this._rows, this.columns, this.sorts);
            this._rows = val;
        }
        // this has to be done to prevent the change detection
        // tree from freaking out because we are readjusting
        setTimeout(() => this.recalculate());
    }
    /**
     * Lifecycle hook that is called when Angular dirty checks a directive.
     *
     * \@memberOf DatatableComponent
     * @return {?}
     */
    ngDoCheck() {
        if (this.rowDiffer.diff(this.rows)) {
            this.recalculatePages();
        }
    }
    /**
     * Recalc's the sizes of the grid.
     *
     * Updated automatically on changes to:
     *
     *  - Columns
     *  - Rows
     *  - Paging related
     *
     * Also can be manually invoked or upon window resize.
     *
     * \@memberOf DatatableComponent
     * @return {?}
     */
    recalculate() {
        this.recalculateDims();
        this.recalculateColumns();
    }
    /**
     * Window resize handler to update sizes.
     *
     * \@memberOf DatatableComponent
     * @return {?}
     */
    onWindowResize() {
        this.recalculate();
    }
    /**
     * Recalulcates the column widths based on column width
     * distribution mode and scrollbar offsets.
     *
     *
     * \@memberOf DatatableComponent
     * @param {?=} columns
     * @param {?=} forceIdx
     * @param {?=} allowBleed
     * @return {?}
     */
    recalculateColumns(columns = this.columns, forceIdx = -1, allowBleed = this.scrollbarH) {
        if (!columns)
            return;
        let /** @type {?} */ width = this.innerWidth;
        if (this.scrollbarV) {
            width = width - this.scrollbarHelper.width;
        }
        if (this.columnMode === ColumnMode.force) {
            forceFillColumnWidths(columns, width, forceIdx, allowBleed);
        }
        else if (this.columnMode === ColumnMode.flex) {
            adjustColumnWidths(columns, width);
        }
        return columns;
    }
    /**
     * Recalculates the dimensions of the table size.
     * Internally calls the page size and row count calcs too.
     *
     * \@memberOf DatatableComponent
     * @return {?}
     */
    recalculateDims() {
        const /** @type {?} */ dims = this.element.getBoundingClientRect();
        this.innerWidth = Math.floor(dims.width);
        if (this.scrollbarV) {
            let /** @type {?} */ height = dims.height;
            if (this.headerHeight)
                height = height - this.headerHeight;
            if (this.footerHeight)
                height = height - this.footerHeight;
            this.bodyHeight = height;
        }
        this.recalculatePages();
    }
    /**
     * Recalculates the pages after a update.
     *
     *
     * \@memberOf DatatableComponent
     * @return {?}
     */
    recalculatePages() {
        this.pageSize = this.calcPageSize();
        this.rowCount = this.calcRowCount();
    }
    /**
     * Body triggered a page event.
     *
     *
     * \@memberOf DatatableComponent
     * @param {?} __0
     * @return {?}
     */
    onBodyPage({ offset }) {
        this.offset = offset;
        this.page.emit({
            count: this.count,
            pageSize: this.pageSize,
            limit: this.limit,
            offset: this.offset
        });
    }
    /**
     * The body triggered a scroll event.
     *
     *
     * \@memberOf DatatableComponent
     * @param {?} event
     * @return {?}
     */
    onBodyScroll(event) {
        this.offsetX = event.offsetX;
        this.scroll.emit(event);
    }
    /**
     * The footer triggered a page event.
     *
     *
     * \@memberOf DatatableComponent
     * @param {?} event
     * @return {?}
     */
    onFooterPage(event) {
        this.offset = event.page - 1;
        this.bodyComponent.updateOffsetY(this.offset);
        this.page.emit({
            count: this.count,
            pageSize: this.pageSize,
            limit: this.limit,
            offset: this.offset
        });
    }
    /**
     * Recalculates the sizes of the page
     *
     *
     * \@memberOf DatatableComponent
     * @param {?=} val
     * @return {?}
     */
    calcPageSize(val = this.rows) {
        // Keep the page size constant even if the row has been expanded.
        // This is because an expanded row is still considered to be a child of
        // the original row.  Hence calculation would use rowHeight only.
        if (this.scrollbarV) {
            const /** @type {?} */ size = Math.ceil(this.bodyHeight / this.rowHeight);
            return Math.max(size, 0);
        }
        // if limit is passed, we are paging
        if (this.limit !== undefined)
            return this.limit;
        // otherwise use row length
        if (val)
            return val.length;
        // other empty :(
        return 0;
    }
    /**
     * Calculates the row count.
     *
     *
     * \@memberOf DatatableComponent
     * @param {?=} val
     * @return {?}
     */
    calcRowCount(val = this.rows) {
        if (!this.externalPaging) {
            if (!val)
                return 0;
            return val.length;
        }
        return this.count;
    }
    /**
     * The header triggered a column resize event.
     *
     *
     * \@memberOf DatatableComponent
     * @param {?} __0
     * @return {?}
     */
    onColumnResize({ column, newValue }) {
        /* Safari/iOS 10.2 workaround */
        if (column === undefined) {
            return;
        }
        let /** @type {?} */ idx;
        const /** @type {?} */ cols = this.columns.map((c, i) => {
            c = Object.assign({}, c);
            if (c.$$id === column.$$id) {
                idx = i;
                c.width = newValue;
                // set this so we can force the column
                // width distribution to be to this value
                c.$$oldWidth = newValue;
            }
            return c;
        });
        this.recalculateColumns(cols, idx);
        this._columns = cols;
        this.resize.emit({
            column,
            newValue
        });
    }
    /**
     * The header triggered a column re-order event.
     *
     *
     * \@memberOf DatatableComponent
     * @param {?} __0
     * @return {?}
     */
    onColumnReorder({ column, newValue, prevValue }) {
        const /** @type {?} */ cols = this.columns.map(c => {
            return Object.assign({}, c);
        });
        const /** @type {?} */ prevCol = cols[newValue];
        cols[newValue] = column;
        cols[prevValue] = prevCol;
        this.columns = cols;
        this.reorder.emit({
            column,
            newValue,
            prevValue
        });
    }
    /**
     * The header triggered a column sort event.
     *
     *
     * \@memberOf DatatableComponent
     * @param {?} event
     * @return {?}
     */
    onColumnSort(event) {
        const { sorts } = event;
        // this could be optimized better since it will resort
        // the rows again on the 'push' detection...
        if (this.externalSorting === false) {
            // don't use normal setter so we don't resort
            this._rows = sortRows(this.rows, this.columns, sorts);
        }
        this.sorts = sorts;
        // Always go to first page when sorting to see the newly sorted data
        this.offset = 0;
        this.bodyComponent.updateOffsetY(this.offset);
        this.sort.emit(event);
    }
    /**
     * Toggle all row selection
     *
     *
     * \@memberOf DatatableComponent
     * @param {?} event
     * @return {?}
     */
    onHeaderSelect(event) {
        // before we splice, chk if we currently have all selected
        const /** @type {?} */ allSelected = this.selected.length === this.rows.length;
        // remove all existing either way
        this.selected = [];
        // do the opposite here
        if (!allSelected) {
            this.selected.push(...this.rows);
        }
        this.select.emit({
            selected: this.selected
        });
    }
    /**
     * A row was selected from body
     *
     *
     * \@memberOf DatatableComponent
     * @param {?} event
     * @return {?}
     */
    onBodySelect(event) {
        this.select.emit(event);
    }
}
DatatableComponent.decorators = [
    { type: Component, args: [{
                selector: 'ngx-datatable',
                template: `
    <div
      visibility-observer
      (visible)="recalculate()">
      <datatable-header
        *ngIf="headerHeight"
        [sorts]="sorts"
        [sortType]="sortType"
        [scrollbarH]="scrollbarH"
        [innerWidth]="innerWidth"
        [offsetX]="offsetX"
        [columns]="columns"
        [headerHeight]="headerHeight"
        [reorderable]="reorderable"
        [sortAscendingIcon]="cssClasses.sortAscending"
        [sortDescendingIcon]="cssClasses.sortDescending"
        [allRowsSelected]="allRowsSelected"
        [selectionType]="selectionType"
        (sort)="onColumnSort($event)"
        (resize)="onColumnResize($event)"
        (reorder)="onColumnReorder($event)"
        (select)="onHeaderSelect($event)">
      </datatable-header>
      <datatable-body
        [rows]="rows"
        [scrollbarV]="scrollbarV"
        [scrollbarH]="scrollbarH"
        [loadingIndicator]="loadingIndicator"
        [rowHeight]="rowHeight"
        [rowCount]="rowCount"
        [offset]="offset"
        [trackByProp]="trackByProp"
        [columns]="columns"
        [pageSize]="pageSize"
        [offsetX]="offsetX"
        [rowDetail]="rowDetail"
        [selected]="selected"
        [innerWidth]="innerWidth"
        [bodyHeight]="bodyHeight"
        [selectionType]="selectionType"
        [emptyMessage]="messages.emptyMessage"
        [rowIdentity]="rowIdentity"
        [rowClass]="rowClass"
        [selectCheck]="selectCheck"
        (page)="onBodyPage($event)"
        (activate)="activate.emit($event)"
        (rowContextmenu)="rowContextmenu.emit($event)"
        (select)="onBodySelect($event)"
        (scroll)="onBodyScroll($event)">
      </datatable-body>
      <datatable-footer
        *ngIf="footerHeight"
        [rowCount]="rowCount"
        [pageSize]="pageSize"
        [offset]="offset"
        [footerHeight]="footerHeight"
        [footerTemplate]="footer"
        [totalMessage]="messages.totalMessage"
        [pagerLeftArrowIcon]="cssClasses.pagerLeftArrow"
        [pagerRightArrowIcon]="cssClasses.pagerRightArrow"
        [pagerPreviousIcon]="cssClasses.pagerPrevious"
        [selectedCount]="selected.length"
        [selectedMessage]="!!selectionType && messages.selectedMessage"
        [pagerNextIcon]="cssClasses.pagerNext"
        (page)="onFooterPage($event)">
      </datatable-footer>
    </div>
  `,
                encapsulation: ViewEncapsulation.None,
                styles: [`
    .ngx-datatable {
      display: block;
      overflow: hidden;
      -webkit-box-pack: center;
          -ms-flex-pack: center;
              justify-content: center;
      position: relative;
      -webkit-transform: translate3d(0, 0, 0);
      /**
       * Vertical Scrolling Adjustments
       */
      /**
       * Horizontal Scrolling Adjustments
       */
      /**
       * Fixed Header Height Adjustments
       */
      /**
       * Fixed row height adjustments
       */
      /**
       * Shared Styles
       */
      /**
       * Header Styles
       */
      /**
       * Body Styles
       */
      /**
       * Footer Styles
       */ }
      .ngx-datatable [hidden] {
        display: none !important; }
      .ngx-datatable *, .ngx-datatable *:before, .ngx-datatable *:after {
        -webkit-box-sizing: border-box;
        box-sizing: border-box; }
      .ngx-datatable.scroll-vertical .datatable-body {
        overflow-y: auto; }
        .ngx-datatable.scroll-vertical .datatable-body .datatable-row-wrapper {
          position: absolute; }
      .ngx-datatable.scroll-horz .datatable-body {
        overflow-x: auto;
        -webkit-overflow-scrolling: touch; }
      .ngx-datatable.fixed-header .datatable-header .datatable-header-inner {
        white-space: nowrap; }
        .ngx-datatable.fixed-header .datatable-header .datatable-header-inner .datatable-header-cell {
          white-space: nowrap;
          overflow: hidden;
          text-overflow: ellipsis; }
      .ngx-datatable.fixed-row .datatable-scroll {
        white-space: nowrap; }
        .ngx-datatable.fixed-row .datatable-scroll .datatable-body-row {
          white-space: nowrap; }
          .ngx-datatable.fixed-row .datatable-scroll .datatable-body-row .datatable-body-cell {
            overflow: hidden;
            white-space: nowrap;
            text-overflow: ellipsis; }
      .ngx-datatable .datatable-body-row,
      .ngx-datatable .datatable-header-inner {
        display: -webkit-box;
        display: -ms-flexbox;
        display: flex;
        -webkit-box-orient: horizontal;
        -webkit-box-direction: normal;
            -ms-flex-direction: row;
                flex-direction: row;
        -ms-flex-flow: row;
        -o-flex-flow: row;
        flex-flow: row; }
      .ngx-datatable .datatable-body-cell,
      .ngx-datatable .datatable-header-cell {
        vertical-align: top;
        display: inline-block;
        line-height: 1.625;
        -webkit-user-select: none;
        -moz-user-select: none;
        -ms-user-select: none;
        -o-user-select: none;
        user-select: none; }
        .ngx-datatable .datatable-body-cell:focus,
        .ngx-datatable .datatable-header-cell:focus {
          outline: none; }
      .ngx-datatable .datatable-row-left,
      .ngx-datatable .datatable-row-right {
        z-index: 9; }
      .ngx-datatable .datatable-row-left,
      .ngx-datatable .datatable-row-center,
      .ngx-datatable .datatable-row-right {
        position: relative; }
      .ngx-datatable .datatable-header {
        display: block;
        overflow: hidden; }
        .ngx-datatable .datatable-header .datatable-header-inner {
          -webkit-box-align: stretch;
              -ms-flex-align: stretch;
                  align-items: stretch;
          -webkit-align-items: stretch; }
        .ngx-datatable .datatable-header .datatable-header-cell {
          position: relative;
          display: inline-block; }
          .ngx-datatable .datatable-header .datatable-header-cell .datatable-header-cell-wrapper {
            cursor: pointer; }
          .ngx-datatable .datatable-header .datatable-header-cell.longpress .datatable-header-cell-wrapper {
            cursor: move; }
          .ngx-datatable .datatable-header .datatable-header-cell .sort-btn {
            line-height: 100%;
            vertical-align: middle;
            display: inline-block;
            cursor: pointer; }
          .ngx-datatable .datatable-header .datatable-header-cell .resize-handle {
            display: inline-block;
            position: absolute;
            right: 0;
            top: 0;
            bottom: 0;
            width: 5px;
            padding: 0 4px;
            visibility: hidden;
            cursor: ew-resize; }
          .ngx-datatable .datatable-header .datatable-header-cell.resizeable:hover .resize-handle {
            visibility: visible; }
      .ngx-datatable .datatable-body {
        position: relative;
        z-index: 10;
        display: block; }
        .ngx-datatable .datatable-body .datatable-scroll {
          display: inline-block; }
        .ngx-datatable .datatable-body .datatable-row-detail {
          overflow-y: hidden; }
        .ngx-datatable .datatable-body .datatable-row-wrapper {
          display: -webkit-box;
          display: -ms-flexbox;
          display: flex;
          -webkit-box-orient: vertical;
          -webkit-box-direction: normal;
          -ms-flex-direction: column;
          flex-direction: column; }
        .ngx-datatable .datatable-body .datatable-body-row {
          outline: none; }
          .ngx-datatable .datatable-body .datatable-body-row > div {
            display: -webkit-box;
            display: -ms-flexbox;
            display: flex; }
      .ngx-datatable .datatable-footer {
        display: block;
        width: 100%; }
        .ngx-datatable .datatable-footer .datatable-footer-inner {
          display: -webkit-box;
          display: -ms-flexbox;
          display: flex;
          -webkit-box-align: center;
              -ms-flex-align: center;
                  align-items: center;
          width: 100%; }
        .ngx-datatable .datatable-footer .selected-count .page-count {
          -webkit-box-flex: 1;
              -ms-flex: 1 1 40%;
                  flex: 1 1 40%; }
        .ngx-datatable .datatable-footer .selected-count .datatable-pager {
          -webkit-box-flex: 1;
              -ms-flex: 1 1 60%;
                  flex: 1 1 60%; }
        .ngx-datatable .datatable-footer .page-count {
          -webkit-box-flex: 1;
              -ms-flex: 1 1 20%;
                  flex: 1 1 20%; }
        .ngx-datatable .datatable-footer .datatable-pager {
          -webkit-box-flex: 1;
              -ms-flex: 1 1 80%;
                  flex: 1 1 80%;
          text-align: right; }
          .ngx-datatable .datatable-footer .datatable-pager .pager,
          .ngx-datatable .datatable-footer .datatable-pager .pager li {
            padding: 0;
            margin: 0;
            display: inline-block;
            list-style: none; }
          .ngx-datatable .datatable-footer .datatable-pager .pager li, .ngx-datatable .datatable-footer .datatable-pager .pager li a {
            outline: none; }
          .ngx-datatable .datatable-footer .datatable-pager .pager li a {
            cursor: pointer;
            display: inline-block; }
          .ngx-datatable .datatable-footer .datatable-pager .pager li.disabled a {
            cursor: not-allowed; }
  `],
                host: {
                    class: 'ngx-datatable'
                }
            },] },
];
/**
 * @nocollapse
 */
DatatableComponent.ctorParameters = () => [
    { type: ScrollbarHelper, },
    { type: ElementRef, },
    { type: KeyValueDiffers, },
];
DatatableComponent.propDecorators = {
    'rows': [{ type: Input },],
    'columns': [{ type: Input },],
    'selected': [{ type: Input },],
    'scrollbarV': [{ type: Input },],
    'scrollbarH': [{ type: Input },],
    'rowHeight': [{ type: Input },],
    'columnMode': [{ type: Input },],
    'headerHeight': [{ type: Input },],
    'footerHeight': [{ type: Input },],
    'externalPaging': [{ type: Input },],
    'externalSorting': [{ type: Input },],
    'limit': [{ type: Input },],
    'count': [{ type: Input },],
    'offset': [{ type: Input },],
    'loadingIndicator': [{ type: Input },],
    'selectionType': [{ type: Input },],
    'reorderable': [{ type: Input },],
    'sortType': [{ type: Input },],
    'sorts': [{ type: Input },],
    'cssClasses': [{ type: Input },],
    'messages': [{ type: Input },],
    'rowIdentity': [{ type: Input },],
    'rowClass': [{ type: Input },],
    'selectCheck': [{ type: Input },],
    'trackByProp': [{ type: Input },],
    'scroll': [{ type: Output },],
    'activate': [{ type: Output },],
    'select': [{ type: Output },],
    'sort': [{ type: Output },],
    'page': [{ type: Output },],
    'reorder': [{ type: Output },],
    'resize': [{ type: Output },],
    'rowContextmenu': [{ type: Output },],
    'isFixedHeader': [{ type: HostBinding, args: ['class.fixed-header',] },],
    'isFixedRow': [{ type: HostBinding, args: ['class.fixed-row',] },],
    'isVertScroll': [{ type: HostBinding, args: ['class.scroll-vertical',] },],
    'isHorScroll': [{ type: HostBinding, args: ['class.scroll-horz',] },],
    'isSelectable': [{ type: HostBinding, args: ['class.selectable',] },],
    'isCheckboxSelection': [{ type: HostBinding, args: ['class.checkbox-selection',] },],
    'isCellSelection': [{ type: HostBinding, args: ['class.cell-selection',] },],
    'isSingleSelection': [{ type: HostBinding, args: ['class.single-selection',] },],
    'isMultiSelection': [{ type: HostBinding, args: ['class.multi-selection',] },],
    'isMultiClickSelection': [{ type: HostBinding, args: ['class.multi-click-selection',] },],
    'columnTemplates': [{ type: ContentChildren, args: [DataTableColumnDirective,] },],
    'rowDetail': [{ type: ContentChild, args: [DatatableRowDetailDirective,] },],
    'footer': [{ type: ContentChild, args: [DatatableFooterDirective,] },],
    'bodyComponent': [{ type: ViewChild, args: [DataTableBodyComponent,] },],
    'onWindowResize': [{ type: HostListener, args: ['window:resize',] },],
};
__decorate([
    throttleable(5),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", []),
    __metadata("design:returntype", void 0)
], DatatableComponent.prototype, "onWindowResize", null);

class DataTableHeaderComponent {
    constructor() {
        this.sort = new EventEmitter();
        this.reorder = new EventEmitter();
        this.resize = new EventEmitter();
        this.select = new EventEmitter();
    }
    /**
     * @param {?} val
     * @return {?}
     */
    set headerHeight(val) {
        if (val !== 'auto') {
            this._headerHeight = `${val}px`;
        }
        else {
            this._headerHeight = val;
        }
    }
    /**
     * @return {?}
     */
    get headerHeight() {
        return this._headerHeight;
    }
    /**
     * @param {?} val
     * @return {?}
     */
    set columns(val) {
        this._columns = val;
        const /** @type {?} */ colsByPin = columnsByPin(val);
        this.columnsByPin = columnsByPinArr(val);
        this.columnGroupWidths = columnGroupWidths(colsByPin, val);
    }
    /**
     * @return {?}
     */
    get columns() {
        return this._columns;
    }
    /**
     * @param {?} __0
     * @return {?}
     */
    onLongPressStart({ event, model }) {
        model.dragging = true;
        this.dragEventTarget = event;
    }
    /**
     * @param {?} __0
     * @return {?}
     */
    onLongPressEnd({ event, model }) {
        this.dragEventTarget = event;
        // delay resetting so sort can be 
        // prevented if we were dragging
        setTimeout(() => {
            model.dragging = false;
        }, 5);
    }
    /**
     * @return {?}
     */
    get headerWidth() {
        if (this.scrollbarH) {
            return this.innerWidth + 'px';
        }
        return '100%';
    }
    /**
     * @param {?} index
     * @param {?} colGroup
     * @return {?}
     */
    trackByGroups(index, colGroup) {
        return colGroup.type;
    }
    /**
     * @param {?} index
     * @param {?} column
     * @return {?}
     */
    columnTrackingFn(index, column) {
        return column.$$id;
    }
    /**
     * @param {?} width
     * @param {?} column
     * @return {?}
     */
    onColumnResized(width, column) {
        if (width <= column.minWidth) {
            width = column.minWidth;
        }
        else if (width >= column.maxWidth) {
            width = column.maxWidth;
        }
        this.resize.emit({
            column,
            prevValue: column.width,
            newValue: width
        });
    }
    /**
     * @param {?} __0
     * @return {?}
     */
    onColumnReordered({ prevIndex, newIndex, model }) {
        this.reorder.emit({
            column: model,
            prevValue: prevIndex,
            newValue: newIndex
        });
    }
    /**
     * @param {?} __0
     * @return {?}
     */
    onSort({ column, prevValue, newValue }) {
        // if we are dragging don't sort!
        if (column.dragging)
            return;
        const /** @type {?} */ sorts = this.calcNewSorts(column, prevValue, newValue);
        this.sort.emit({
            sorts,
            column,
            prevValue,
            newValue
        });
    }
    /**
     * @param {?} column
     * @param {?} prevValue
     * @param {?} newValue
     * @return {?}
     */
    calcNewSorts(column, prevValue, newValue) {
        let /** @type {?} */ idx = 0;
        const /** @type {?} */ sorts = this.sorts.map((s, i) => {
            s = Object.assign({}, s);
            if (s.prop === column.prop)
                idx = i;
            return s;
        });
        if (newValue === undefined) {
            sorts.splice(idx, 1);
        }
        else if (prevValue) {
            sorts[idx].dir = newValue;
        }
        else {
            if (this.sortType === SortType.single) {
                sorts.splice(0, this.sorts.length);
            }
            sorts.push({ dir: newValue, prop: column.prop });
        }
        return sorts;
    }
    /**
     * @param {?} group
     * @return {?}
     */
    stylesByGroup(group) {
        const /** @type {?} */ widths = this.columnGroupWidths;
        const /** @type {?} */ offsetX = this.offsetX;
        const /** @type {?} */ styles = {
            width: `${widths[group]}px`
        };
        if (group === 'center') {
            translateXY(styles, offsetX * -1, 0);
        }
        else if (group === 'right') {
            const /** @type {?} */ totalDiff = widths.total - this.innerWidth;
            const /** @type {?} */ offset = totalDiff * -1;
            translateXY(styles, offset, 0);
        }
        return styles;
    }
}
DataTableHeaderComponent.decorators = [
    { type: Component, args: [{
                selector: 'datatable-header',
                template: `
    <div
      orderable
      (reorder)="onColumnReordered($event)"
      [style.width.px]="columnGroupWidths.total"
      class="datatable-header-inner">
      <div
        *ngFor="let colGroup of columnsByPin; trackBy: trackByGroups"
        [class]="'datatable-row-' + colGroup.type"
        [ngStyle]="stylesByGroup(colGroup.type)">
        <datatable-header-cell
          *ngFor="let column of colGroup.columns; trackBy: columnTrackingFn"
          resizeable
          [resizeEnabled]="column.resizeable"
          (resize)="onColumnResized($event, column)"
          long-press
          [pressModel]="column"
          [pressEnabled]="reorderable && column.draggable"
          (longPressStart)="onLongPressStart($event)"
          (longPressEnd)="onLongPressEnd($event)"
          draggable
          [dragX]="reorderable && column.draggable && column.dragging"
          [dragY]="false"
          [dragModel]="column"
          [dragEventTarget]="dragEventTarget"
          [headerHeight]="headerHeight"
          [column]="column"
          [sortType]="sortType"
          [sorts]="sorts"
          [selectionType]="selectionType"
          [sortAscendingIcon]="sortAscendingIcon"
          [sortDescendingIcon]="sortDescendingIcon"
          (sort)="onSort($event)"
          (select)="select.emit($event)">
        </datatable-header-cell>
      </div>
    </div>
  `,
                host: {
                    class: 'datatable-header'
                }
            },] },
];
/**
 * @nocollapse
 */
DataTableHeaderComponent.ctorParameters = () => [];
DataTableHeaderComponent.propDecorators = {
    'sortAscendingIcon': [{ type: Input },],
    'sortDescendingIcon': [{ type: Input },],
    'scrollbarH': [{ type: Input },],
    'innerWidth': [{ type: Input },],
    'offsetX': [{ type: Input },],
    'sorts': [{ type: Input },],
    'sortType': [{ type: Input },],
    'allRowsSelected': [{ type: Input },],
    'selectionType': [{ type: Input },],
    'reorderable': [{ type: Input },],
    'headerHeight': [{ type: HostBinding, args: ['style.height',] }, { type: Input },],
    'columns': [{ type: Input },],
    'sort': [{ type: Output },],
    'reorder': [{ type: Output },],
    'resize': [{ type: Output },],
    'select': [{ type: Output },],
    'headerWidth': [{ type: HostBinding, args: ['style.width',] },],
};

class DataTableHeaderCellComponent {
    constructor() {
        this.sort = new EventEmitter();
        this.select = new EventEmitter();
        this.sortFn = this.onSort.bind(this);
    }
    /**
     * @param {?} val
     * @return {?}
     */
    set sorts(val) {
        this._sorts = val;
        this.sortDir = this.calcSortDir(val);
        this.sortClass = this.calcSortClass(this.sortDir);
    }
    /**
     * @return {?}
     */
    get sorts() {
        return this._sorts;
    }
    /**
     * @return {?}
     */
    get columnCssClasses() {
        let /** @type {?} */ cls = 'datatable-header-cell';
        if (this.column.sortable)
            cls += ' sortable';
        if (this.column.resizeable)
            cls += ' resizeable';
        if (this.column.headerClass) {
            if (typeof this.column.headerClass === 'string') {
                cls += ' ' + this.column.headerClass;
            }
            else if (typeof this.column.headerClass === 'function') {
                const /** @type {?} */ res = this.column.headerClass({
                    column: this.column
                });
                if (typeof res === 'string') {
                    cls += res;
                }
                else if (typeof res === 'object') {
                    const /** @type {?} */ keys = Object.keys(res);
                    for (const /** @type {?} */ k of keys) {
                        if (res[k] === true)
                            cls += ` ${k}`;
                    }
                }
            }
        }
        const /** @type {?} */ sortDir = this.sortDir;
        if (sortDir) {
            cls += ` sort-active sort-${sortDir}`;
        }
        return cls;
    }
    /**
     * @return {?}
     */
    get name() {
        // guaranteed to have a value by setColumnDefaults() in column-helper.ts
        return this.column.headerTemplate === undefined ? this.column.name : undefined;
    }
    /**
     * @return {?}
     */
    get minWidth() {
        return this.column.minWidth;
    }
    /**
     * @return {?}
     */
    get maxWidth() {
        return this.column.maxWidth;
    }
    /**
     * @return {?}
     */
    get width() {
        return this.column.width;
    }
    /**
     * @return {?}
     */
    get isCheckboxable() {
        return this.column.checkboxable &&
            this.column.headerCheckboxable &&
            this.selectionType === SelectionType.checkbox;
    }
    /**
     * @param {?} sorts
     * @return {?}
     */
    calcSortDir(sorts) {
        if (sorts && this.column) {
            const /** @type {?} */ sort = sorts.find((s) => {
                return s.prop === this.column.prop;
            });
            if (sort)
                return sort.dir;
        }
    }
    /**
     * @return {?}
     */
    onSort() {
        if (!this.column.sortable)
            return;
        const /** @type {?} */ newValue = nextSortDir(this.sortType, this.sortDir);
        this.sort.emit({
            column: this.column,
            prevValue: this.sortDir,
            newValue
        });
    }
    /**
     * @param {?} sortDir
     * @return {?}
     */
    calcSortClass(sortDir) {
        if (sortDir === SortDirection.asc) {
            return `sort-btn sort-asc ${this.sortAscendingIcon}`;
        }
        else if (sortDir === SortDirection.desc) {
            return `sort-btn sort-desc ${this.sortDescendingIcon}`;
        }
        else {
            return `sort-btn`;
        }
    }
}
DataTableHeaderCellComponent.decorators = [
    { type: Component, args: [{
                selector: 'datatable-header-cell',
                template: `
    <div>
      <label
        *ngIf="isCheckboxable" 
        class="datatable-checkbox">
        <input 
          type="checkbox"
          [attr.checked]="allRowsSelected"
          (change)="select.emit(!allRowsSelected)" 
        />
      </label>
      <span 
        *ngIf="!column.headerTemplate"
        class="datatable-header-cell-wrapper">
        <span
          class="datatable-header-cell-label draggable"
          (click)="onSort()"
          [innerHTML]="name">
        </span>
      </span>
      <ng-template
        *ngIf="column.headerTemplate"
        [ngTemplateOutlet]="column.headerTemplate"
        [ngOutletContext]="{ 
          column: column, 
          sortDir: sortDir,
          sortFn: sortFn
        }">
      </ng-template>
      <span
        (click)="onSort()"
        [class]="sortClass">
      </span>
    </div>
  `
            },] },
];
/**
 * @nocollapse
 */
DataTableHeaderCellComponent.ctorParameters = () => [];
DataTableHeaderCellComponent.propDecorators = {
    'sortType': [{ type: Input },],
    'column': [{ type: Input },],
    'sortAscendingIcon': [{ type: Input },],
    'sortDescendingIcon': [{ type: Input },],
    'allRowsSelected': [{ type: Input },],
    'selectionType': [{ type: Input },],
    'headerHeight': [{ type: HostBinding, args: ['style.height.px',] }, { type: Input },],
    'sorts': [{ type: Input },],
    'sort': [{ type: Output },],
    'select': [{ type: Output },],
    'columnCssClasses': [{ type: HostBinding, args: ['class',] },],
    'name': [{ type: HostBinding, args: ['attr.title',] },],
    'minWidth': [{ type: HostBinding, args: ['style.minWidth.px',] },],
    'maxWidth': [{ type: HostBinding, args: ['style.maxWidth.px',] },],
    'width': [{ type: HostBinding, args: ['style.width.px',] },],
};

/**
 * Draggable Directive for Angular2
 *
 * Inspiration:
 *   https://github.com/AngularClass/angular2-examples/blob/master/rx-draggable/directives/draggable.ts
 *   http://stackoverflow.com/questions/35662530/how-to-implement-drag-and-drop-in-angular2
 *
 */
class DraggableDirective {
    /**
     * @param {?} element
     */
    constructor(element) {
        this.dragX = true;
        this.dragY = true;
        this.dragStart = new EventEmitter();
        this.dragging = new EventEmitter();
        this.dragEnd = new EventEmitter();
        this.isDragging = false;
        this.element = element.nativeElement;
    }
    /**
     * @param {?} changes
     * @return {?}
     */
    ngOnChanges(changes) {
        if (changes['dragEventTarget'] && changes['dragEventTarget'].currentValue && this.dragModel.dragging) {
            this.onMousedown(changes['dragEventTarget'].currentValue);
        }
    }
    /**
     * @return {?}
     */
    ngOnDestroy() {
        this._destroySubscription();
    }
    /**
     * @param {?} event
     * @return {?}
     */
    onMouseup(event) {
        if (!this.isDragging)
            return;
        this.isDragging = false;
        this.element.classList.remove('dragging');
        if (this.subscription) {
            this._destroySubscription();
            this.dragEnd.emit({
                event,
                element: this.element,
                model: this.dragModel
            });
        }
    }
    /**
     * @param {?} event
     * @return {?}
     */
    onMousedown(event) {
        // we only want to drag the inner header text
        const /** @type {?} */ isDragElm = ((event.target)).classList.contains('draggable');
        if (isDragElm && (this.dragX || this.dragY)) {
            event.preventDefault();
            this.isDragging = true;
            const /** @type {?} */ mouseDownPos = { x: event.clientX, y: event.clientY };
            const /** @type {?} */ mouseup = Observable$1.fromEvent(document, 'mouseup');
            this.subscription = mouseup
                .subscribe((ev) => this.onMouseup(ev));
            const /** @type {?} */ mouseMoveSub = Observable$1.fromEvent(document, 'mousemove')
                .takeUntil(mouseup)
                .subscribe((ev) => this.move(ev, mouseDownPos));
            this.subscription.add(mouseMoveSub);
            this.dragStart.emit({
                event,
                element: this.element,
                model: this.dragModel
            });
        }
    }
    /**
     * @param {?} event
     * @param {?} mouseDownPos
     * @return {?}
     */
    move(event, mouseDownPos) {
        if (!this.isDragging)
            return;
        const /** @type {?} */ x = event.clientX - mouseDownPos.x;
        const /** @type {?} */ y = event.clientY - mouseDownPos.y;
        if (this.dragX)
            this.element.style.left = `${x}px`;
        if (this.dragY)
            this.element.style.top = `${y}px`;
        this.element.classList.add('dragging');
        this.dragging.emit({
            event,
            element: this.element,
            model: this.dragModel
        });
    }
    /**
     * @return {?}
     */
    _destroySubscription() {
        if (this.subscription) {
            this.subscription.unsubscribe();
            this.subscription = undefined;
        }
    }
}
DraggableDirective.decorators = [
    { type: Directive, args: [{ selector: '[draggable]' },] },
];
/**
 * @nocollapse
 */
DraggableDirective.ctorParameters = () => [
    { type: ElementRef, },
];
DraggableDirective.propDecorators = {
    'dragEventTarget': [{ type: Input },],
    'dragModel': [{ type: Input },],
    'dragX': [{ type: Input },],
    'dragY': [{ type: Input },],
    'dragStart': [{ type: Output },],
    'dragging': [{ type: Output },],
    'dragEnd': [{ type: Output },],
};

class LongPressDirective {
    constructor() {
        this.pressEnabled = true;
        this.duration = 500;
        this.longPressStart = new EventEmitter();
        this.longPressing = new EventEmitter();
        this.longPressEnd = new EventEmitter();
        this.mouseX = 0;
        this.mouseY = 0;
    }
    /**
     * @return {?}
     */
    get press() { return this.pressing; }
    /**
     * @return {?}
     */
    get isLongPress() {
        return this.isLongPressing;
    }
    /**
     * @param {?} event
     * @return {?}
     */
    onMouseDown(event) {
        // don't do right/middle clicks
        if (event.which !== 1 || !this.pressEnabled)
            return;
        // don't start drag if its on resize handle
        const /** @type {?} */ target = ((event.target));
        if (target.classList.contains('resize-handle'))
            return;
        this.mouseX = event.clientX;
        this.mouseY = event.clientY;
        this.pressing = true;
        this.isLongPressing = false;
        const /** @type {?} */ mouseup = Observable$1.fromEvent(document, 'mouseup');
        this.subscription = mouseup.subscribe((ev) => this.onMouseup());
        this.timeout = setTimeout(() => {
            this.isLongPressing = true;
            this.longPressStart.emit({
                event,
                model: this.pressModel
            });
            this.subscription.add(Observable$1.fromEvent(document, 'mousemove')
                .takeUntil(mouseup)
                .subscribe((mouseEvent) => this.onMouseMove(mouseEvent)));
            this.loop(event);
        }, this.duration);
        this.loop(event);
    }
    /**
     * @param {?} event
     * @return {?}
     */
    onMouseMove(event) {
        if (this.pressing && !this.isLongPressing) {
            const /** @type {?} */ xThres = Math.abs(event.clientX - this.mouseX) > 10;
            const /** @type {?} */ yThres = Math.abs(event.clientY - this.mouseY) > 10;
            if (xThres || yThres) {
                this.endPress();
            }
        }
    }
    /**
     * @param {?} event
     * @return {?}
     */
    loop(event) {
        if (this.isLongPressing) {
            this.timeout = setTimeout(() => {
                this.longPressing.emit({
                    event,
                    model: this.pressModel
                });
                this.loop(event);
            }, 50);
        }
    }
    /**
     * @return {?}
     */
    endPress() {
        clearTimeout(this.timeout);
        this.isLongPressing = false;
        this.pressing = false;
        this._destroySubscription();
        this.longPressEnd.emit({
            model: this.pressModel
        });
    }
    /**
     * @return {?}
     */
    onMouseup() {
        this.endPress();
    }
    /**
     * @return {?}
     */
    ngOnDestroy() {
        this._destroySubscription();
    }
    /**
     * @return {?}
     */
    _destroySubscription() {
        if (this.subscription) {
            this.subscription.unsubscribe();
            this.subscription = undefined;
        }
    }
}
LongPressDirective.decorators = [
    { type: Directive, args: [{ selector: '[long-press]' },] },
];
/**
 * @nocollapse
 */
LongPressDirective.ctorParameters = () => [];
LongPressDirective.propDecorators = {
    'pressEnabled': [{ type: Input },],
    'pressModel': [{ type: Input },],
    'duration': [{ type: Input },],
    'longPressStart': [{ type: Output },],
    'longPressing': [{ type: Output },],
    'longPressEnd': [{ type: Output },],
    'press': [{ type: HostBinding, args: ['class.press',] },],
    'isLongPress': [{ type: HostBinding, args: ['class.longpress',] },],
    'onMouseDown': [{ type: HostListener, args: ['mousedown', ['$event'],] },],
};

class OrderableDirective {
    /**
     * @param {?} differs
     * @param {?} document
     */
    constructor(differs, document) {
        this.document = document;
        this.reorder = new EventEmitter();
        this.differ = differs.find({}).create(null);
    }
    /**
     * @return {?}
     */
    ngAfterContentInit() {
        // HACK: Investigate Better Way
        this.updateSubscriptions();
        this.draggables.changes.subscribe(this.updateSubscriptions.bind(this));
    }
    /**
     * @return {?}
     */
    ngOnDestroy() {
        this.draggables.forEach(d => {
            d.dragStart.unsubscribe();
            d.dragEnd.unsubscribe();
        });
    }
    /**
     * @return {?}
     */
    updateSubscriptions() {
        const /** @type {?} */ diffs = this.differ.diff(this.createMapDiffs());
        if (diffs) {
            const /** @type {?} */ subscribe = ({ currentValue, previousValue }) => {
                unsubscribe({ previousValue });
                if (currentValue) {
                    currentValue.dragStart.subscribe(this.onDragStart.bind(this));
                    currentValue.dragEnd.subscribe(this.onDragEnd.bind(this));
                }
            };
            const /** @type {?} */ unsubscribe = ({ previousValue }) => {
                if (previousValue) {
                    previousValue.dragStart.unsubscribe();
                    previousValue.dragEnd.unsubscribe();
                }
            };
            diffs.forEachAddedItem(subscribe.bind(this));
            // diffs.forEachChangedItem(subscribe.bind(this));
            diffs.forEachRemovedItem(unsubscribe.bind(this));
        }
    }
    /**
     * @return {?}
     */
    onDragStart() {
        this.positions = {};
        let /** @type {?} */ i = 0;
        for (const /** @type {?} */ dragger of this.draggables.toArray()) {
            const /** @type {?} */ elm = dragger.element;
            const /** @type {?} */ left = parseInt(elm.offsetLeft.toString(), 0);
            this.positions[dragger.dragModel.prop] = {
                left,
                right: left + parseInt(elm.offsetWidth.toString(), 0),
                index: i++,
                element: elm
            };
        }
    }
    /**
     * @param {?} __0
     * @return {?}
     */
    onDragEnd({ element, model, event }) {
        const /** @type {?} */ prevPos = this.positions[model.prop];
        const /** @type {?} */ target = this.isTarget(model, event);
        if (target) {
            this.reorder.emit({
                prevIndex: prevPos.index,
                newIndex: target.i,
                model
            });
        }
        element.style.left = 'auto';
    }
    /**
     * @param {?} model
     * @param {?} event
     * @return {?}
     */
    isTarget(model, event) {
        let /** @type {?} */ i = 0;
        const /** @type {?} */ x = event.x || event.clientX;
        const /** @type {?} */ y = event.y || event.clientY;
        const /** @type {?} */ targets = this.document.elementsFromPoint(x, y);
        for (const /** @type {?} */ prop in this.positions) {
            // current column position which throws event.
            const /** @type {?} */ pos = this.positions[prop];
            // since we drag the inner span, we need to find it in the elements at the cursor
            if (model.prop !== prop && targets.find((el) => el === pos.element)) {
                return {
                    pos,
                    i
                };
            }
            i++;
        }
    }
    /**
     * @return {?}
     */
    createMapDiffs() {
        return this.draggables.toArray()
            .reduce((acc, curr) => {
            acc[curr.dragModel.$$id] = curr;
            return acc;
        }, {});
    }
}
OrderableDirective.decorators = [
    { type: Directive, args: [{ selector: '[orderable]' },] },
];
/**
 * @nocollapse
 */
OrderableDirective.ctorParameters = () => [
    { type: KeyValueDiffers, },
    { type: undefined, decorators: [{ type: Inject, args: [DOCUMENT,] },] },
];
OrderableDirective.propDecorators = {
    'reorder': [{ type: Output },],
    'draggables': [{ type: ContentChildren, args: [DraggableDirective, { descendants: true },] },],
};

class ResizeableDirective {
    /**
     * @param {?} element
     */
    constructor(element) {
        this.resizeEnabled = true;
        this.resize = new EventEmitter();
        this.resizing = false;
        this.element = element.nativeElement;
    }
    /**
     * @return {?}
     */
    ngAfterViewInit() {
        if (this.resizeEnabled) {
            const /** @type {?} */ node = document.createElement('span');
            node.classList.add('resize-handle');
            this.element.appendChild(node);
        }
    }
    /**
     * @return {?}
     */
    ngOnDestroy() {
        this._destroySubscription();
    }
    /**
     * @return {?}
     */
    onMouseup() {
        this.resizing = false;
        if (this.subscription && !this.subscription.closed) {
            this._destroySubscription();
            this.resize.emit(this.element.clientWidth);
        }
    }
    /**
     * @param {?} event
     * @return {?}
     */
    onMousedown(event) {
        const /** @type {?} */ isHandle = (((event.target))).classList.contains('resize-handle');
        const /** @type {?} */ initialWidth = this.element.clientWidth;
        const /** @type {?} */ mouseDownScreenX = event.screenX;
        if (isHandle) {
            event.stopPropagation();
            this.resizing = true;
            const /** @type {?} */ mouseup = Observable$1.fromEvent(document, 'mouseup');
            this.subscription = mouseup
                .subscribe((ev) => this.onMouseup());
            const /** @type {?} */ mouseMoveSub = Observable$1.fromEvent(document, 'mousemove')
                .takeUntil(mouseup)
                .subscribe((e) => this.move(e, initialWidth, mouseDownScreenX));
            this.subscription.add(mouseMoveSub);
        }
    }
    /**
     * @param {?} event
     * @param {?} initialWidth
     * @param {?} mouseDownScreenX
     * @return {?}
     */
    move(event, initialWidth, mouseDownScreenX) {
        const /** @type {?} */ movementX = event.screenX - mouseDownScreenX;
        const /** @type {?} */ newWidth = initialWidth + movementX;
        const /** @type {?} */ overMinWidth = !this.minWidth || newWidth >= this.minWidth;
        const /** @type {?} */ underMaxWidth = !this.maxWidth || newWidth <= this.maxWidth;
        if (overMinWidth && underMaxWidth) {
            this.element.style.width = `${newWidth}px`;
        }
    }
    /**
     * @return {?}
     */
    _destroySubscription() {
        if (this.subscription) {
            this.subscription.unsubscribe();
            this.subscription = undefined;
        }
    }
}
ResizeableDirective.decorators = [
    { type: Directive, args: [{
                selector: '[resizeable]',
                host: {
                    '[class.resizeable]': 'resizeEnabled'
                }
            },] },
];
/**
 * @nocollapse
 */
ResizeableDirective.ctorParameters = () => [
    { type: ElementRef, },
];
ResizeableDirective.propDecorators = {
    'resizeEnabled': [{ type: Input },],
    'minWidth': [{ type: Input },],
    'maxWidth': [{ type: Input },],
    'resize': [{ type: Output },],
    'onMousedown': [{ type: HostListener, args: ['mousedown', ['$event'],] },],
};

/**
 * Visibility Observer Directive
 *
 * Usage:
 *
 * 		<div
 * 			visibility-observer
 * 			(visible)="onVisible($event)">
 * 		</div>
 *
 */
class VisibilityDirective {
    /**
     * @param {?} element
     * @param {?} zone
     */
    constructor(element, zone) {
        this.element = element;
        this.zone = zone;
        this.isVisible = false;
        this.visible = new EventEmitter();
    }
    /**
     * @return {?}
     */
    ngOnInit() {
        this.runCheck();
    }
    /**
     * @return {?}
     */
    ngOnDestroy() {
        clearTimeout(this.timeout);
    }
    /**
     * @return {?}
     */
    onVisibilityChange() {
        // trigger zone recalc for columns
        this.zone.run(() => {
            this.isVisible = true;
            this.visible.emit(true);
        });
    }
    /**
     * @return {?}
     */
    runCheck() {
        const /** @type {?} */ check = () => {
            // https://davidwalsh.name/offsetheight-visibility
            const { offsetHeight, offsetWidth } = this.element.nativeElement;
            if (offsetHeight && offsetWidth) {
                clearTimeout(this.timeout);
                this.onVisibilityChange();
            }
            else {
                clearTimeout(this.timeout);
                this.zone.runOutsideAngular(() => {
                    this.timeout = setTimeout(() => check(), 50);
                });
            }
        };
        setTimeout(() => check());
    }
}
VisibilityDirective.decorators = [
    { type: Directive, args: [{ selector: '[visibility-observer]' },] },
];
/**
 * @nocollapse
 */
VisibilityDirective.ctorParameters = () => [
    { type: ElementRef, },
    { type: NgZone, },
];
VisibilityDirective.propDecorators = {
    'isVisible': [{ type: HostBinding, args: ['class.visible',] },],
    'visible': [{ type: Output },],
};

class NgxDatatableModule {
}
NgxDatatableModule.decorators = [
    { type: NgModule, args: [{
                imports: [
                    CommonModule
                ],
                providers: [
                    ScrollbarHelper
                ],
                declarations: [
                    DataTableFooterTemplateDirective,
                    VisibilityDirective,
                    DraggableDirective,
                    ResizeableDirective,
                    OrderableDirective,
                    LongPressDirective,
                    ScrollerComponent,
                    DatatableComponent,
                    DataTableColumnDirective,
                    DataTableHeaderComponent,
                    DataTableHeaderCellComponent,
                    DataTableBodyComponent,
                    DataTableFooterComponent,
                    DataTablePagerComponent,
                    ProgressBarComponent,
                    DataTableBodyRowComponent,
                    DataTableRowWrapperComponent,
                    DatatableRowDetailDirective,
                    DatatableRowDetailTemplateDirective,
                    DataTableBodyCellComponent,
                    DataTableSelectionComponent,
                    DataTableColumnHeaderDirective,
                    DataTableColumnCellDirective,
                    DatatableFooterDirective
                ],
                exports: [
                    DatatableComponent,
                    DatatableRowDetailDirective,
                    DatatableRowDetailTemplateDirective,
                    DataTableColumnDirective,
                    DataTableColumnHeaderDirective,
                    DataTableColumnCellDirective,
                    DataTableFooterTemplateDirective,
                    DatatableFooterDirective
                ]
            },] },
];
/**
 * @nocollapse
 */
NgxDatatableModule.ctorParameters = () => [];

/**
 * Generated bundle index. Do not edit.
 */

export { NgxDatatableModule, ColumnMode, SortType, SortDirection, SelectionType, ClickType, DatatableComponent, DataTableHeaderComponent, DataTableHeaderCellComponent, DataTableBodyComponent, DataTableBodyCellComponent, DataTableBodyRowComponent, ProgressBarComponent, ScrollerComponent, DataTableRowWrapperComponent, DataTableSelectionComponent, DataTableFooterComponent, DataTablePagerComponent, DatatableFooterDirective, DataTableFooterTemplateDirective, DataTableColumnDirective, DataTableColumnHeaderDirective, DataTableColumnCellDirective, DatatableRowDetailDirective, DatatableRowDetailTemplateDirective, DataTableBodyCellComponent as ɵu, DataTableBodyComponent as ɵm, DataTableBodyRowComponent as ɵq, DataTableColumnCellDirective as ɵx, DataTableColumnDirective as ɵj, DataTableColumnHeaderDirective as ɵw, DataTableFooterComponent as ɵn, DataTableFooterTemplateDirective as ɵb, DataTableHeaderCellComponent as ɵl, DataTableHeaderComponent as ɵk, DataTablePagerComponent as ɵo, DataTableRowWrapperComponent as ɵr, DataTableSelectionComponent as ɵv, DatatableComponent as ɵi, DatatableFooterDirective as ɵy, DatatableRowDetailDirective as ɵs, DatatableRowDetailTemplateDirective as ɵt, ProgressBarComponent as ɵp, ScrollerComponent as ɵh, DraggableDirective as ɵd, LongPressDirective as ɵg, OrderableDirective as ɵf, ResizeableDirective as ɵe, VisibilityDirective as ɵc, ScrollbarHelper as ɵa };
//# sourceMappingURL=ngx-datatable-ng-packagr.js.map
